      PROGRAM MCLCOM
************************************************************************
*                                                                      *
*         The Merlin Control Language compiler, version 3.0            *
*                                                                      *
*  Implemented by:                                                     *
*         D. G. Papageorgiou                                           *
*         Dept. of Chemistry, University of Ioannina, GREECE.          *
*         Email: jimmy@nrt.cs.uoi.gr                                   *
*                                                                      *
*         I. E. Lagaris                                                *
*         Dept. of Computer Science, University of Ioannina, GREECE.   *
*         Email: lagaris@cs.uoi.gr                                     *
*                                                                      *
*         I. N. Demetropoulos                                          *
*         Dept. of Chemistry, University of Ioannina, GREECE.          *
*         Email: idimitr@cc.uoi.gr                                     *
*                                                                      *
*  All authors can be reached at: merlin@nrt.cs.uoi.gr                 *
*  More information is available at: http://nrt.cs.uoi.gr/merlin       *
*                                                                      *
************************************************************************
      IMPLICIT INTEGER (A-Z)
C
C  Name of the Merlin configuration file.
      CHARACTER*(*) CFILE
      PARAMETER ( CFILE = '$(CONFIG)' )
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
C  Get compilation parameters.
      CALL GETPAR
C
C  Initialize.
      CALL INIT
C
C  Read the Merlin configuration file.
      CALL CONPRO(CFILE,IERRU,IERROR)
      IF (IERROR.EQ.2) THEN
         CALL HEAD
         CALL CONECP(CFILE,IERRU)
         STOP
      END IF
C
C  Initialization that depends on configuration parameters.
      CALL ZINIT
C
C  Read the panel description file.
      CALL PARSP(PDESCF,ERL,.TRUE.,.TRUE.,IERR)
      IF (IERR.NE.0) THEN
         CALL DASH
         STOP
      END IF
C
C  Open the MCL files.
      CALL FOPEN
C
C  Copy configuration errors to the output file.
      IF (IERROR.NE.0) THEN
         CALL HEAD
         CALL CONEOU(CFILE,ERL,IERRU)
         STOP
      END IF
C
C  Compile the MCL program.
      CALL MCL
C
C  Close the MCL files.
      CALL FCLOSE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHARR ( NT, NDIM, ERR, ARR, NN, IOPT, NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will check if an array is defined in a VAR statement.
C    In this case, it will parse the definition and update the
C    information table.
C    The part of the input of the input line that is actually parsed is:
C
C    [ L1:U1, L2:U2, ... Ln:Un ]
C
C    with L1, l2, ...Ln being the low dimension bounds, and
C    U1, U2, ...Un being the upper dimension bounds respectively.
C
C  Input arguments:
C    IOPT       IOPT = 0 -> We are parsing a VAR declaration.
C               IOPT = 1 -> We are parsing the dummy argument list of a
C                           SUB statement. * is allowed as the upper
C                           bound of the last dimension of an array.
C    NAME       The name of the array, we are currently parsing.
C
C  Output arguments:
C    NDIM       Integer variable. This is the number of dimensions for
C               the declared array.
C    ERR        Logical variable. Indicates whether an error occurred
C               (ERR=.TRUE.) in the declaration, or not. (ERR=.FALSE.)
C    ARR        Logical variable. Indicates whether the item in the
C               VAR declaration is an array (ARR=.TRUE.),
C               or not. (ARR=.FALSE.)
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      LOGICAL ERR, ARR
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL ADJ
C  ---------------------------------------------------------------------
C
C  Initialize the output arguments.
      ADJ = .FALSE.
      NDIM = 0
      ERR = .FALSE.
      ARR = .TRUE.
C
C  If a left bracket follows, this is an array definition;
C  a simple variable otherwise.
      IF (TOKEN(NT+1).NE.LBRA) THEN
         ARR = .FALSE.
         RETURN
      END IF
C
      NT = NT+1
C
C  Start a loop over all dimensions in the declaration.
10    NDIM = NDIM+1
      NT = NT+1
C
C  Check low dimension bound.
      ISI = 1
      CUR = TOKEN(NT)
      IF (CUR.EQ.PLUS) THEN
C  A plus sign is defined.
         NT = NT+1
         CUR = TOKEN(NT)
      ELSE IF (CUR.EQ.MINUS) THEN
C  A minus sign is defined.
         ISI = -1
         NT = NT+1
         CUR = TOKEN(NT)
      END IF
C
      IF (CUR.NE.UNSI) THEN
C  Plus, minus signs not defined.
C  This isn't either an unsigned number.
         ERR = .TRUE.
         CALL ERROR(55,NAME)
         RETURN
      END IF
C
C  Determine and push lower dimension bound in the information table.
      NN = NN+1
      NINF = NINF+1
      IF (NINF.GT.MXINFO)
     &    CALL OVER('information','INFO','CHARR',MXINFO,'MXINFO')
      LOW = NINT(ISI*NUM(NN))
      INFO(NINF) = LOW
C
C  A colon should separate lower, upper dimension bounds.
      NT = NT+1
      IF (TOKEN(NT).NE.COLON) THEN
         ERR = .TRUE.
         CALL ERROR(56,NAME)
         RETURN
      END IF
C
C  Check upper dimension bound.
      NT = NT+1
      CUR = TOKEN(NT)
      IF (CUR.NE.STAR) THEN
         ISI = 1
         IF (CUR.EQ.PLUS) THEN
C  A plus sign is defined.
            NT = NT+1
            CUR = TOKEN(NT)
         ELSE IF (CUR.EQ.MINUS) THEN
C  A minus sign is defined.
            ISI = -1
            NT = NT+1
            CUR = TOKEN(NT)
         END IF
C  Plus, minus signs not defined.
C  This isn't either an unsigned number.
         IF (CUR.NE.UNSI) THEN
            ERR = .TRUE.
            CALL ERROR(57,NAME)
            RETURN
         END IF
C  Determine and push upper dimension bound in the information table.
         NN = NN+1
         NINF = NINF+1
         IF (NINF.GT.MXINFO)
     &          CALL OVER('information','INFO','CHARR',MXINFO,'MXINFO')
         UPPER = NINT(ISI*NUM(NN))
         INFO(NINF) = UPPER
C  Lower dimension bound should be less or equal to the
C  upper dimension bound.
         IF (UPPER.LT.LOW) CALL ERROR(59,NAME)
      ELSE IF (IOPT.NE.0) THEN
         IF (TOKEN(NT+1).NE.RBRA) THEN
            CALL ERROR(25,NAME)
         ELSE
            ADJ = .TRUE.
            NINF = NINF+1
            IF (NINF.GT.MXINFO)
     &       CALL OVER('information','INFO','CHARR',MXINFO,'MXINFO')
            UPPER = LOW
            INFO(NINF) = UPPER
         END IF
      ELSE
         CALL ERROR(26,NAME)
      END IF
C
      NT = NT+1
      CUR = TOKEN(NT)
C
C  A comma denotes that another dimension is declared.
      IF (CUR.EQ.COMMA) GOTO 10
C
C  A right bracket ends the array declaration.
      IF (CUR.EQ.RBRA) THEN
         IF (ADJ) NDIM = -NDIM
         RETURN
C
      END IF
C  Neither comma, nor right bracket follows.
      CALL ERROR(58,NAME)
      ERR = .TRUE.
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHI ( ID, NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will parse statements of the form:
C      COMMAND ( <keyword>.<index> )
C    where <keyword> depends on the statement (may be R, L, X  etc).
C    and <index> is an expression. Multiple entries of the
C    <keyword>.<index> pair are allowed:
C      COMMAND ( <keyword>.<index1> ; <keyword>.<index2> ... )
C    The object code generated is:
C
C    code to evaluate <index1>
C    code to evaluate <index2>
C        .........
C    code to evaluate <indexN>
C    $PUSHC
C    <number of key-index pairs>
C    COMMAND
C
C  Input arguments:
C    ID         Integer that identifies the statement being parsed:
C               ID = 1 -> LEFT_DEMARGIN
C               ID = 2 -> RIGHT_DEMARGIN
C               ID = 3 -> FIX
C               ID = 4 -> LOOSE
C               ID = 5 -> NONAME
C               ID = 6 -> CONFIDENCE
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME, STARG(1)
      INTEGER EX(LINLEN)
      LOGICAL CORRE
      CHARACTER KEY(6)
      CHARACTER*14 WCOM(6)
      DATA KEY / 'L', 'R', 'X', 'X', 'X', 'X' /
      DATA WCOM / 'LDEMARGIN', 'RDEMARGIN', 'FIX', 'LOOSE', 'NONAME',
     &            'CONFIDENCE' /
C  ---------------------------------------------------------------------
C
C  Left parenthesis should follow.
      NT = NT+1
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,WCOM(ID))
         RETURN
      END IF
C
C  NC is the number of <keyword>.<index> pairs.
      NC = 0
C
C  Loop over all <keyword>.<index> pairs.
100   CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         RETURN
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,NAME)
      IF (ISCOMP(KEY(ID),NAME).EQ.0) THEN
         CALL ERROR(33,NAME)
         RETURN
      END IF
C
C  A period should follow.
      NT = NT+1
      IF (TOKEN(NT).NE.PERIOD) THEN
         CALL ERROR(34,KEY(ID))
         RETURN
      END IF
C
C  Extract and evaluate the expression.
      CALL EXTRA(NT,EX,NEX,2,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If expression incorrect.
      IF (.NOT.CORRE) RETURN
C
      NC = NC+1
      NT = NT+1
      CUR = TOKEN(NT)
C
C  Another <keyword>.<index> pair ?
      IF (CUR.EQ.SEMI) GOTO 100
C
C  End of statement ?
      IF (CUR.EQ.RIGHTP) THEN
C  If a right parenthesis, then the statement has been completed.
C  Write the final object code lines.
         CALL WPSC(SCR,BLOC,NC)
         BLOC = BLOC+1
         IF (GENE) WRITE (SCR,'(A)') WCOM(ID)
      ELSE
C  Neither semicolon nor a right parenthesis follow.
         CALL ERROR(36,' ')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIS ( ID, NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will parse statements of the form:
C      COMMAND ( <keyword>.<index>=<string> )
C    where <keyword> depends on the statement (may be X, C, G etc.),
C    <index> is an MCL expression, and <string> is a string (enclosed
C    in quotes). Multiple entries of the <keyword>.<index>=<string>
C    specification, are allowed:
C      COMMAND ( <keyword>.<index1>=<string1> ;
C                <keyword>.<index2>=<string2> ... )
C    The object code generated is:
C
C    code to evaluate <index1>
C    code to evaluate <index2>
C            ....
C    code to evaluate <indexN>
C    COMMAND
C    <number of index-value specifications>
C    <length of string1>
C    <string1>
C    <length of string2>
C    <string2>
C      ...
C    <length of stringN>
C    <stringN>
C
C  Input arguments:
C    ID         Integer that identifies the statement being parsed:
C               ID = 1 -> GODFATHER
C               ID = 2 -> CFLAG
C               ID = 3 -> MIXED
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME, STARG(1)
      INTEGER EX(LINLEN)
      LOGICAL CORRE
      CHARACTER*80 LINE
      CHARACTER KEY(3)
      CHARACTER*9 WCOM(3)
      DATA KEY / 'X', 'C', 'G' /
      DATA WCOM / 'GODFATHER', 'CFLAG', 'MIXED' /
C  ---------------------------------------------------------------------
C
C  A left parenthesis should follow.
      NT = NT+1
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,WCOM(ID))
         RETURN
      END IF
C
C  Open the spare file. Strings and string-lengths will be written here.
      OPEN (SPARE,STATUS='SCRATCH',IOSTAT=IERR)
      IF (IERR.NE.0) CALL MABORT('Cannot open the spare file.','CHIS')
C
C  Number of lines in the SPARE file.
      NL = 0
C
C  NC is the number of index-string pairs.
      NC = 0
C
C  Loop over all index-string pairs.
100   CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         GOTO 200
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,NAME)
      IF (ISCOMP(KEY(ID),NAME).EQ.0) THEN
         CALL ERROR(33,NAME)
         GOTO 200
      END IF
C
C  A period should follow the keyword.
      NT = NT+1
      IF (TOKEN(NT).NE.PERIOD) THEN
         CALL ERROR(34,KEY(ID))
         GOTO 200
      END IF
C
C  Extract and evaluate the expression for the <index>.
      CALL EXTRA(NT,EX,NEX,2,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If expression incorrect.
      IF (.NOT.CORRE) RETURN
C
C  The = sign must follow the <index>.
      NT = NT+1
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(35,' ')
         GOTO 200
      END IF
C
C  An quote starts the string.
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL ERROR(169,NAME)
         GOTO 200
      END IF
C
C  Get the string.
      NT = NT+1
      IF (TOKEN(NT).NE.STRING) THEN
         CALL ERROR(28,' ')
         GOTO 200
      END IF
      CALL GETIDS(NA,NN,IS,IE)
C
C  A quote terminates the string.
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL ERROR(169,NAME)
         GOTO 200
      END IF
C
C  Add the string and string-length in the SPARE file.
      NL = NL+2
      IF (GENE) THEN
         WRITE (SPARE,*) IE-IS+1
         WRITE (SPARE,'(A)') ALPHA(IS:IE)
      END IF
C
      NC = NC+1
      NT = NT+1
      CUR = TOKEN(NT)
C
C  Another index-string pair ?
      IF (CUR.EQ.SEMI) GOTO 100
C
C  A right parenthesis terminates the statement.
      IF (CUR.NE.RIGHTP) THEN
         CALL ERROR(36,' ')
         GOTO 200
      END IF
C
C  Add number of index-string pairs.
      BLOC = BLOC+2
      IF (GENE) THEN
         WRITE (SCR,'(A)') WCOM(ID)
         WRITE (SCR,*) NC
      END IF
C
C  Copy SPARE file.
      BLOC = BLOC+NL
      IF (GENE) THEN
         REWIND SPARE
         DO 10,I=1,NL
            READ (SPARE,'(A)') LINE
            WRITE (SCR,'(A)') LINE
10       CONTINUE
      END IF
C
200   CLOSE (SPARE,STATUS='DELETE',IOSTAT=IERR)
      IF (IERR.NE.0)
     &    CALL MABORT('Cannot close the spare file.','CHIS')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CHIV ( NT, NA, NN, ID )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will parse statements of the form:
C      COMMAND ( <keyword>.<index>=<value> )
C    where <keyword> depends on the statement (may be X, S, F etc.)
C    and <index>, <value> are MCL expressions. Multiple entries of the
C    <keyword>.<index>=<value> specification, are allowed:
C      COMMAND ( <keyword>.<index1>=<value1> ;
C                <keyword>.<index2>=<value2> ... )
C    The object code generated is:
C
C    code to evaluate <index1>
C    code to evaluate <value1>
C    code to evaluate <index2>
C    code to evaluate <value2>
C               ...
C    code to evaluate <indexN>
C    code to evaluate <valueN>
C    $PUSHC
C    <number of index-value specifications>
C    COMMAND
C
C  Input arguments:
C    ID         Integer that identifies the statement being parsed:
C               ID = 1 -> POINT
C               ID = 2 -> STEP
C               ID = 3 -> FLAG
C               ID = 4 -> LMARGIN
C               ID = 5 -> RMARGIN
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) STARG(1), K
      CHARACTER KEY(5)
      CHARACTER*7 WCOM(5)
      LOGICAL CORRE
      INTEGER EX(LINLEN)
      DATA WCOM / 'POINT  ',
     &            'STEP   ',
     &            'FLAG   ',
     &            'LMARGIN',
     &            'RMARGIN'/
      DATA KEY  / 'X', 'S', 'F', 'L', 'R' /
C  ---------------------------------------------------------------------
C
C  A left parenthesis is expected after the command name.
      NT = NT+1
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,WCOM(ID))
         RETURN
      END IF
C
C  NC is the number of keywords entered in the statement.
      NC = 0
C
C  Loop over all index-value specifications.
C
1000  CONTINUE
C  An identifier is expected as a keyword.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         RETURN
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,K)
C
C  See if the keyword is valid for the command.
      IF (K.NE.KEY(ID)) THEN
         CALL ERROR(33,K)
         RETURN
      END IF
C
C  A period is expected after the keyword.
      NT = NT+1
      IF (TOKEN(NT).NE.PERIOD) THEN
         CALL ERROR(34,K)
         RETURN
      END IF
C
C  Extract the expression following the period.
      CALL EXTRA(NT,EX,NEX,2,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If expression incorrect.
      IF (.NOT.CORRE) RETURN
C
C  An = sign should follow the expression.
      NT = NT+1
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(35,' ')
         RETURN
      END IF
C
C  Extract the expression following the = sign.
      CALL EXTRA(NT,EX,NEX,2,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If expression incorrect.
      IF (.NOT.CORRE) RETURN
C
C  Get the next token.
      NT = NT+1
      CUR=TOKEN(NT)
      NC = NC+1
C
C  If a semicolon, then another keyword follows.
      IF (CUR.EQ.SEMI) GOTO 1000
      IF (CUR.EQ.RIGHTP) THEN
C
C  If a right parenthesis, then the statement has been completed.
C  Write the final object code lines.
         CALL WPSC(SCR,BLOC,NC)
         BLOC = BLOC+1
         IF (GENE) WRITE (SCR,'(A)') WCOM(ID)
      ELSE
C  Neither semicolon nor a right parenthesis follow.
         CALL ERROR(36,' ')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CMSORT ( COMMAN, NAME, IND, NCOM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CMSORT sorts the character array COMMAN in ascending
C    order, applying the same transformations to the integer arrays
C    NAME and IND.
C    For information on the sorting method (bubble sort) see:
C      N. Wirth, Algorithms & Data Structures,
C      Prentice-Hall, 1986.
C      Knuth D.E., The art of computer programming, vol 3,
C      Addison-Wesley, Reading, Mass., 1973.
C
C  Input arguments:
C    NCOM       Number of entries in array COMMAN.
C
C  Input / Output arguments:
C    COMMAN     Character array to be sorted.
C    NAME       Integer array.
C    IND        Integer array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) COMMAN(NCOM)
      INTEGER NAME(NCOM), IND(NCOM)
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( IDLEN = 30 )
      CHARACTER*(IDLEN) CTEMP
      LOGICAL SORTED
C  ---------------------------------------------------------------------
C
      DO 10,K=1,NCOM-1
         SORTED = .TRUE.
         DO 20,I=NCOM,K+1,-1
            IF (COMMAN(I).LT.COMMAN(I-1)) THEN
               SORTED = .FALSE.
               IM1 = I-1
               CTEMP = COMMAN(I)
               COMMAN(I) = COMMAN(IM1)
               COMMAN(IM1) = CTEMP
               ITEMP = IND(I)
               IND(I) = IND(IM1)
               IND(IM1) = ITEMP
               ITEMP = NAME(I)
               NAME(I) = NAME(IM1)
               NAME(IM1) = ITEMP
            END IF
20       CONTINUE
         IF (SORTED) RETURN
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONPRO ( CFILE, IERRU, IERROR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine CONPRO acts as a front end to subroutine MCONF that
C    parses the Merlin configuration file.
C
C  Input arguments:
C    CFILE      Name of the Merlin configuration file.
C
C  Output arguments:
C    IERRU      Unit number of the temporary file that accumulates
C               error messages during the run of subroutine MCONF.
C               If IERROR .NE. 0, IERRU remains open.
C    IERROR     Error indicator:
C               IERROR = 0 -> No error.
C               IERROR = 1 -> Some error.
C               IERROR = 2 -> Some error related to directives that
C                             alter I/O units/files.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) CFILE
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      COMMON /UALLOC/ MINU, MAXU, LASTU
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
      PARAMETER ( NMCOM = 99, MXPLUG = $(MXPLUG), NES = NMCOM+MXPLUG )
      CHARACTER*14 STAT(NES)
      COMMON /COM1/ STAT
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( NURANG = 10 )
      CHARACTER PROCES, PROLO, EPILO, IN, OUT, HFILE, MFILE
      CHARACTER*(MXFILE) CONFIL
      LOGICAL HEADER, STRO, WEAK, IPRECO, OPRECO, HASAPP, OUTAPP
      $(TYPE) BIGNUM, DWARF
C
      PARAMETER ( NEXCLU = 11 )
      CHARACTER*10 EXCLU(NEXCLU)
      DATA EXCLU / 'WHEN      ',
     &             'IF        ',
     &             'ELSE      ',
     &             'ENDIF     ',
     &             'LOOP      ',
     &             'ENDLOOP   ',
     &             'PROGRAM   ',
     &             'SUB       ',
     &             'END       ',
     &             'VAR       ',
     &             'FUNCTION  '/
C  ---------------------------------------------------------------------
C
      CONFIL = CFILE
      ICONU= NUNIT()
      IFILEU = NUNIT()
      IERRU = NUNIT()
      IF (ICONU.EQ.-1 .OR. IFILEU.EQ.-1 .OR. IERRU.EQ.-1) THEN
         CALL HEAD
         CALL DASHO
         WRITE (*,10)
         CALL DASHO
         STOP
      END IF
C
      IUOUT = -1
      CALL MCONF(CONFIL,ICONU,IFILEU,IERRU,NMCOM,MXPLUG,STAT,EXCLU,
     &           NEXCLU,NURANG,IERROR,HEADER,PROCES,PROLO,EPILO,IN,
     &           OUT,PDESCF,STRO,WEAK,IPRECO,OPRECO,BIGNUM,DWARF,
     &           IUINP,IUOUT,MINU,MAXU,HASAPP,OUTAPP,NSCHA,NSINT,
     &           NSREAL,HFILE,MFILE,EFILE,BFILE,NDIG,NEOFA)
      IF (IUOUT.NE.-1) ERL = IUOUT
      IF (IERROR.EQ.0) LASTU = MINU-1
C
10    FORMAT ($(FC)'Cannot allocate unit numbers.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE COPY ( INIT, FINAL, IUFROM, IUTO )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine COPY is used at the end of passes 2 and 3 of the
C    compiler. It copies part of a file (unit IUFROM) to another file
C    (unit IUTO). The part to be copied is specified by the initial
C    (INIT) and final (FINAL) line counters. These counters have nothing
C    to do with the actual location of the lines in the file.
C    (FINAL-INIT-1) lines are copied starting at the current file
C    position.
C
C  Input arguments:
C    INIT     Initial line counter. Copying starts at INIT+1.
C    FINAL    Final line counter. Copying stops at FINAL-1.
C    IUFROM   Unit to copy from.
C    IUTO     Unit to copy to.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Local variables:
      PARAMETER ( LLSCR = 80 )
      CHARACTER*(LLSCR) LINE
C  ---------------------------------------------------------------------
C
      DO 10,I=INIT+1,FINAL-1
         READ (IUFROM,'(A)') LINE
         LE = LENGTH(LINE)
         IF (LE.GT.0) THEN
            WRITE (IUTO,'(A)') LINE(1:LE)
         ELSE
            WRITE (IUTO,'('' '')')
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DASH
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will print a horizontal line on the output file.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
      WRITE (ERL,10)
C
10    FORMAT ($(FC)79('-'))
      END
C  ---------------------------------------------------------------------
      LOGICAL FUNCTION DEF ( V, VALUE, FILE, NF )
C  ---------------------------------------------------------------------
C
C  Description:
C    This function will test if a given argument (V) is used as a file
C    name in a previously encountered keyword. In such a case, a value
C    of .TRUE. will be returned; .FALSE. otherwise.
C
C  Input arguments:
C    V          Character variable. The function will check if the
C               this file name has been entered as a value to a
C               previously encountered keyword.
C    VALUE      Character array. Contains the values of the keywords
C               entered in the input line.
C    FILE       Integer array. Each entry of this array is a pointer
C               to the value array. It points to the positions of
C               the value array, that have been used as file names
C               in the I, E or B parameters.
C    NF         Integer variable. Number of entries in array FILE.
C               This is actually the number of keywords that used a
C               file name as a value, so far.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) V, VALUE(*)
      INTEGER FILE(*)
C  ---------------------------------------------------------------------
C
      DEF = .TRUE.
      DO 10,I=1,NF
         IF (V.EQ.VALUE(FILE(I))) RETURN
10    CONTINUE
      DEF = .FALSE.
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DISPLA ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine handles the DISPLAY statement.
C    The syntax of this statement is:
C
C    DISPLAY item1; item2; ...itemN              or
C    DISPLAY item1; item2; ...itemN TO <filename>
C
C    where <filename> is the file to receive the output,
C    and item1; item2; ...itemN is a list of items.
C    Each item may be either a valid MCL expression, or a string
C    (delimited by single quotes).
C
C    The routine will first determine the input <filename>. If its
C    omitted, the default Merlin input file will be used.
C    Then it will determine the number of variables to get by counting
C    the semicolons in the input line.
C    A loop over all items in the item list will start.
C    For an expression the code generated is:
C
C    <code to evaluate the expression>
C    $OUT
C    <filename>
C
C    where <code to evaluate the expression> is the code necessary to
C    evaluate the expression. (Obviously more than one line)
C    For a string the code generated is:
C
C    $NOTE
C    <length>
C    <string>
C    <filename>
C
C    where <length> is the length of the string, in characters,
C    <string> is the string of characters to be displayed, and
C    <filename> is the file to receive the output. (".STANDARD" will
C    be used if <filename> is omitted from the DISPLAY statement)
C    When all the items in the list have been processed, the following
C    two more lines of object code are written:
C
C    $FLUSH
C    <filename>
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) FILE
      INTEGER EX(LINLEN)
      LOGICAL CORRE
      PARAMETER ( LLSCR = 80 )
      CHARACTER*(IDLEN) STARG(1), IDF
C  ---------------------------------------------------------------------
C
C  Check to see if <filename> is defined. Use the default otherwise.
      IF (NTOK.GE.7) THEN
         TK1 = TOKEN(NTOK-1)
         TK2 = TOKEN(NTOK-2)
         TK3 = TOKEN(NTOK-3)
         TK4 = TOKEN(NTOK-4)
         IF (TK1.EQ.QUOTE .AND. TK2.EQ.STRING .AND.
     &          TK3.EQ.QUOTE .AND. TK4.EQ.IDENT) THEN
            NID=NA
            DO 99,I=NT+1,NTOK
               TK = TOKEN(I)
               IF (TK.EQ.IDENT .OR. TK.EQ.STRING) NID=NID+1
99          CONTINUE
            CALL GETIAS(NID,IS,IE)
            FILE = ALPHA(IS:IE)
         ELSE
            FILE = '.STANDARD'
         END IF
      ELSE
         FILE='.STANDARD'
      END IF
C
C  Start a loop over all items in the list.
30    CONTINUE
      NT = NT+1
      IF (TOKEN(NT).EQ.QUOTE) THEN
         NT = NT+1
         IF (TOKEN(NT).NE.STRING) THEN
            CALL ERROR(28,' ')
            RETURN
         END IF
C  This item is a string.
         CALL GETIDS(NA,NN,IS,IE)
         NT = NT+1
         IF (TOKEN(NT).EQ.QUOTE) THEN
C  The string terminates normally.
C  Now write the necessary object code.
            IF (GENE) THEN
C  LL is the string length in characters.
               LL=IE-IS+1
               NFULL = LL/LLSCR
               NREM = MOD(LL,LLSCR)
               DO 10,I=1,NFULL
                  IS1 = IS+(I-1)*LLSCR
                  IE1 = IS1+LLSCR-1
                  WRITE (SCR,'(A/I3/A/A)')
     &                      '$NOTE', LLSCR, ALPHA(IS1:IE1), FILE
                  BLOC = BLOC+4
10             CONTINUE
               IF (NREM.NE.0) THEN
                  IS1 = IS+NFULL*LLSCR
                  WRITE (SCR,'(A/I3/A/A)')
     &                      '$NOTE', NREM, ALPHA(IS1:IE), FILE
                  BLOC = BLOC+4
               END IF
            END IF
         ELSE
            CALL ERROR(28,' ')
            RETURN
         END IF
      ELSE
C  This item in the list is expected to be an expression.
         NT = NT-1
C  Extract the expression.
         CALL EXTRA(NT,EX,NEX,1,NA)
C  Evaluate the expression.
         CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
         IF (.NOT.CORRE) RETURN
C  If the expression is correct, generate code.
         BLOC = BLOC+2
         IF (GENE) WRITE (SCR,'(A/A)') '$OUT', FILE
      END IF
C
20    NT=NT+1
      IF (NT.EQ.NTOK) THEN
C  All tokens in the input line exhausted.
C  Write the final instructions and return.
         BLOC = BLOC+2
         IF (GENE) WRITE (SCR,'(A/A)') '$FLUSH', '.STANDARD'
         RETURN
      END IF
C
      CUR = TOKEN(NT)
C  A semicolon denotes another item in the list.
      IF (CUR.EQ.SEMI) GOTO 30
      IF (CUR.EQ.IDENT) THEN
C  If an identifier follows, it should be "TO".
         CALL GETIDF(NA,NN,IDF)
         IF (IDF.NE.'TO') THEN
            CALL ERROR(67,IDF)
            RETURN
         END IF
         NT = NT+1
C  A string (the <filename>) should follow.
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL ERROR(68,' ')
            RETURN
         END IF
         NT = NT+1
C  A string (the <filename>) should follow.
         IF (TOKEN(NT).NE.STRING) THEN
            CALL ERROR(68,' ')
            RETURN
         END IF
         NT = NT+1
C  A string (the <filename>) should follow.
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL ERROR(68,' ')
            RETURN
         END IF
C  Write the final instructions.
         BLOC = BLOC+2
         IF (GENE) WRITE (SCR,'(A/A)') '$FLUSH', FILE
         NT = NT+1
C  If anything else, other than an end of line follows.
         IF (NT.NE.NTOK) CALL ERROR(9,'DISPLAY')
         RETURN
      END IF
C  Neither semicolon, nor an identifier follows.
      CALL ERROR(29,' ')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ELSE
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes the ELSE statement. The syntax of this statement is:
C    ELSE
C    The code generated is:
C
C    $MOVE
C    <blank line>
C
C    The type table will be searched backwards until an entry of 1
C    (indicating the presence of the IF statement, corresponding
C    to this ELSE) is reached.
C    The number of lines to jump so as to reach this ELSE statement
C    (in case the expression involved in the IF statement is not
C    satisfied) will be calculated and pushed in the appropriate
C    position, in the type-link table. The entry in the type table
C    for this IF will be set to -1, indicating that it has been
C    processed. A value of 2 is pushed in the type table, (indicating
C    the presence of the ELSE statement ) while the location of
C    the above <blank line> is pushed in the type-definition table, in
C    order to be used later on.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
C  If more than 2 tokens (ELSE EOL) are in the same line,
C  this is an error.
      IF (NTOK.NE.2) THEN
         CALL ERROR(9,'ELSE')
         RETURN
      END IF
C
C  Generate code, so that when execution of the IF...THEN block
C  completes, the ELSE...ENDIF block will be skipped, and the ENDIF
C  statement will be reached.
C  A blank line will be left after the $MOVE instruction. The correct
C  number of lines to jump will be written in pass 2.
      BLOC = BLOC+2
      IF (GENE) WRITE (SCR,'(A/)') '$MOVE'
C
C  Search the type table backwards until an entry of 1 is located.
      DO 10,I=NTYPE,TSTART,-1
         TP = TYPE(I)
         IF (TP.GT.0) THEN
            IF (TP.EQ.1) THEN
C  An entry of 1 is found in the type table. Calculate the type-link
C  entry for this IF.
               TREL(I) = BLOC-TDEF(I)+1
               TYPE(I) = -TP
C  Put this ELSE in the type table.
               NTYPE = NTYPE+1
               IF (NTYPE.GT.MXTYP)
     &              CALL OVER('type','TYPE','ELSE',MXTYP,'MXTYP')
               TYPE(NTYPE) = 2
               TDEF(NTYPE) = BLOC
            ELSE
C  This is an ELSE with no corresponding IF.
               CALL ERROR(144,' ')
            END IF
            RETURN
         END IF
C  This is an ELSE with no corresponding IF.
10    CONTINUE
      CALL ERROR(144,' ')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ENDIF ( NT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes the ENDIF statement. The syntax of this statement is:
C    ENDIF, or END IF
C    Only one line of object code is generated:
C    $CONTINUE
C
C    The type table will be searched backwards until an entry of 1
C    or 2 (indicating the presence of an IF or ELSE statement)
C    is reached.
C    The number of lines to jump so as to reach this ENDIF statement
C    will be calculated and pushed in the appropriate position, in the
C    type-link table. The entry in the type table for the IF or ELSE
C    located, will be set to -1 or -2 respectively, indicating that they
C    have been processed. A value of -3 is pushed in the type table,
C    (indicating the presence of the ENDIF statement) while the
C    location of the above $CONTINUE instruction is pushed in the
C    type-definition table, in order to be used later on.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C
C  Notes:
C    Note that only two values are possible for NT upon entry: 1 or 2.
C    If the statement is entered as ENDIF (one word), this will
C    cause NT to take on the value 1, since only one token
C    (an identifier) is found.
C    If if the statement is entered as END IF (two distinct words),
C    this will cause NT to take on the value 2, since two tokens
C    (two identifiers) are found. Both of them are used by subroutine
C    PCG in order to recognize the statement.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
C  If anything else follows the ENDIF statement, this is an error.
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL ERROR(9,'END IF')
         RETURN
      END IF
C
C  Generate code.
      BLOC = BLOC+1
      IF (GENE) WRITE (SCR,'(A)') '$CONTINUE'
C
C  Search the type table backwards until an entry of 1 or 2 is reached.
      DO 10,I=NTYPE,TSTART,-1
         TP = TYPE(I)
         IF (TP.GT.0) THEN
            IF (TP.EQ.1 .OR. TP.EQ.2) THEN
C  A value of 1 or 2 has been reached. Calculate the appropriate
C  entry in the type-link table.
               TREL(I) = BLOC-TDEF(I)
               TYPE(I) = -TP
C  Enter this ENDIF in the type table.
               NTYPE = NTYPE+1
               IF (NTYPE.GT.MXTYP)
     &              CALL OVER('type','TYPE','ENDIF',MXTYP,'MXTYP')
               TYPE(NTYPE) = -3
               TDEF(NTYPE) = BLOC
            ELSE
C  This is an ENDIF without a corresponding IF.
               CALL ERROR(145,' ')
            END IF
            RETURN
         END IF
C
10    CONTINUE
C  This is an ENDIF without a corresponding IF.
      CALL ERROR(145,' ')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ENDLOO ( NT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the ENDLOOP statement and generates appropriate object code.
C    The type table is searched backwards until an entry of 4
C    (indicating the corresponding loop) is reached. At this point, the
C    type-link table entry for the loop is calculated. Meanwhile for
C    every entry of 8 in the type table (indicating an EXIT statement)
C    the corresponding type-link table entry is calculated.
C
C    17 lines of object code are generated.
C    Initially the loop control index is incremented by the step value:
C
C    $PUSH     or, for a subprogram argument:    $IPUSH
C    var_mloc                                    N
C    $PUSH
C    step_mloc
C    $+
C    $PUSH    or, for a subprogram argument:    $IPUSH
C    var_mloc                                   N
C
C    where var_mloc is the memory location assigned to the loop control
C    variable. In case the loop control variable is an argument in the
C    current subprogram, N is its serial number in the argument list.
C    step_mloc is the memory location used to store the step for
C    this loop.
C
C    The trip counter is then incremented by 1:
C
C    $PUSHC
C    1
C    $PUSH
C    ctc_mloc
C    $+
C    $POP
C    ctc_mloc
C
C    where ctc_mloc is the memory location used to store the
C    trip counter.
C
C    Then the loop is repeated once more:
C
C    $MOVE
C    <blank line>
C    $CONTINUE
C
C    A value of -5 is pushed in the type table, indicating the ENDLOOP,
C    while the location of the above <blank line> in the scratch file,
C    is pushed in the type-definition table. The entry in the type-link
C    table for this ENDLOOP is calculated.
C
C  Input / Output arguments:
C    NT         Integer variable. This is the number of tokens used
C               so far by the previously called parsing routines.
C
C  Abort conditions that may occur:
C    If the type table overflows, the program will stop and
C    issue an informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
C  If anything else than an EOL follows the statement, this is an error.
C
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL ERROR(9,'END LOOP')
         RETURN
      END IF
C
C  Search the type array backwards.
      DO 10,I=NTYPE,TSTART,-1
         TP = TYPE(I)
         IF (TP.GT.0) THEN
            IF (TP.EQ.4) THEN
C  The corresponding LOOP has been found.
               K = TREL(I)
               LP = BLOC-TDEF(I)
               TREL(I) = LP+17
               TYPE(I) = -4
               GOTO 90
            ELSE IF (TP.EQ.8) THEN
C  An EXIT statement for this loop has been found.
               TREL(I) = BLOC-TDEF(I)+17
               TYPE(I) = -8
            ELSE
C  This is an ENDLOOP without a corresponding LOOP statement.
               CALL ERROR(140,' ')
               RETURN
            END IF
         END IF
10    CONTINUE
C  This is an ENDLOOP without a corresponding LOOP statement.
      CALL ERROR(140,' ')
      RETURN
C
90    CONTINUE
C
C  Get necessary information for the storage used by this loop.
      L     = LLOC(NLOO)
      LM    = LMEM(NLOO)
      ISTEP = LM+2
      ITC   = LM+3
      NLOO  = NLOO-1
C
C  Write object code.
      BLOC = BLOC+5
      IF (GENE) THEN
C  Increment the loop control variable.
         IF (K.EQ.1) THEN
            CALL WPS(SCR,BLOC,ABS(L))
         ELSE
            WRITE (SCR,'(A)') '$IPUSH'
            WRITE (SCR,*) ABS(L)
            BLOC = BLOC+2
         END IF
         CALL WPS(SCR,BLOC,ISTEP)
         WRITE (SCR,'(A)') '$+'
         IF (K.EQ.1) THEN
            CALL WPOP(SCR,BLOC,ABS(L))
         ELSE
            WRITE (SCR,'(A)') '$IPOP'
            WRITE (SCR,*) ABS(L)
            BLOC = BLOC+2
         END IF
C  Increment the trip counter.
         CALL WPSC(SCR,BLOC,1)
         CALL WPS(SCR,BLOC,ITC)
         WRITE (SCR,'(A)') '$+'
         CALL WPOP(SCR,BLOC,ITC)
C  Repeat the loop once more.
         WRITE (SCR,'(A//A)') '$MOVE','$CONTINUE'
      END IF
C
C  Update the type, type-definition and type-link tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','ENDLOO',MXTYP,'MXTYP')
      TYPE(NTYPE) = -5
      TDEF(NTYPE) = BLOC-1
      TREL(NTYPE) = -(LP+21)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ENDSUB
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the END statement.
C    For a main program the object code is:
C      $FINISH
C    For a subprogram:
C      $PUSHC           --->     Deallocate memory.
C      <blank line>
C      $DEALLOC
C      $POPMB           --->     Restore old memory base.
C      $POPINAD         --->     Restore old argument list.
C      $SUBRETURN       --->     Return.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*2 RLAB
      LOGICAL FOUND
C  ---------------------------------------------------------------------
C
      IF (.NOT.GENE) RETURN
C
      IF (INWHAT.EQ.1) THEN
C  Write final code for a main program.
         BLOC = BLOC+1
         IF (GENE) WRITE (SCR,'(A)') '$FINISH'
      ELSE IF (INWHAT.EQ.2) THEN
C
C  Insert fake label "..".
C  All SUBRETURN statements reference this label.
         RLAB = '..'
         CALL LOCAN(LABEL,NLAB,RLAB,FOUND,IND)
C  Check if already in table. This is definitely a malfunction.
         IF (FOUND)
     &          CALL MABORT('Return label ".." is defined.','ENDSUB')
C  Label not in table, so insert it and assign to it
C  the appropriate location.
         NLAB = NLAB+1
         IF (NLAB.GT.MXLAB)
     &          CALL OVER('label','LABEL','ENDSUB',MXLAB,'MXLAB')
         DO 20,I=NLAB,IND+1,-1
            LABEL(I) = LABEL(I-1)
            LABLOC(I) = LABLOC(I-1)
20       CONTINUE
C  Put the label in the label table.
         LABEL(IND) = RLAB
C  Assign the scratch file location to the label.
         LABLOC(IND) = BLOC+1
C
C  Write final code for a subprogram.
         BLOC = BLOC+2
         IF (GENE) THEN
            WRITE (SCR,10) '$PUSHC'
            WRITE (SCR,*)
         END IF
C  Update the type and type-definition tables.
         NTYPE = NTYPE+1
         IF (NTYPE.GT.MXTYP)
     &          CALL OVER('type','TYPE','IF',MXTYP,'MXTYP')
         TYPE(NTYPE)=9
         TDEF(NTYPE)=BLOC
         BLOC = BLOC+4
         IF (GENE) THEN
            WRITE (SCR,10) '$DEALLOC'
            WRITE (SCR,10) '$POPMB'
            WRITE (SCR,10) '$POPINAD'
            WRITE (SCR,10) '$SUBRETURN'
         END IF
      ELSE
         CALL MABORT('Incorrect "inwhat".','ENDSUB')
      END IF
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ERROR ( CODE, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Handles error processing. When a syntactic error is detected during
C    the parsing of an MCL source line, this routine is called to issue
C    the appropriate message.
C    An option (CDIS) is available that controls display of the error
C    code to the output file. This may be useful for debugging
C    purposes, or for further development.
C    This routine may be called directly, to issue error messages and
C    disable code generation, or indirectly (through subroutine WARN)
C    to issue warning messages. Warnings do not disable code generation.
C    The setting of variable ISERR in common block MCDTA determines the
C    behavior of this routine. See also subroutine WARN.
C
C  Input arguments:
C    CODE       Integer variable. This is the error code number to be
C               issued.
C    MESS       Character variable. This is usually the name of a
C               variable, MCL statement, etc. that will be substituted
C               in place of the % sign in the error message.
C
C  Abort conditions that may occur:
C    If the error buffer is full, the program will stop and issue an
C    informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  Maximum number of error messages in one line of the source program.
      PARAMETER ( MAXERR = 20 )
      CHARACTER*90 EBUFF(MAXERR)
      COMMON /MCLERB/ EBUFF
C  ---------------------------------------------------------------------
C  Local variables:
C
C  Total number of error messages issued by the compiler.
      PARAMETER ( MAXMES = 185 )
      CHARACTER*60 MESAGE(MAXMES), MMESS
      CHARACTER*73 PMESS
      CHARACTER ELINE*90, WMESS*13, STR*10
      CHARACTER C
      CHARACTER*7 PREFIX
C
C  Parameter indicating, whether the error code will be displayed
C  on the error list, or not.
      LOGICAL CDIS
      DATA CDIS / .FALSE. /
C
C  The compiler's error messages.
      DATA (MESAGE(I), I=1,15)/
     & 'Label % is already defined.                                 ',
     & 'The definition of label % is not followed by End-of-Line.   ',
     & 'An identifier is expected in the list of variables.         ',
     & 'Variable % is already defined.                              ',
     & 'You must use a semicolon to separate arguments.             ',
     & 'JUST is expected in the WHEN statement.                     ',
     & 'This is not an MCL statement.                               ',
     & 'THEN is expected in the IF statement.                       ',
     & 'The % statement is not followed by End-of-Line.             ',
     & 'An expression is not allowed for the output-only keyword %. ',
     & 'Label % has not been defined.                               ',
     & 'The reserved name % appears in a declaration statement.     ',
     & 'FROM is expected after the loop control variable. Found %.  ',
     & 'TO is expected. Found %.                                    ',
     & 'TO is expected after MOVE. Found % instead.                 '/
      DATA (MESAGE(I), I=16,30)/
     & 'A label name is expected after MOVE TO.                     ',
     & 'TO is expected after MOVE.                                  ',
     & '% is a reserved name and cannot be assigned a value.        ',
     & 'Variable % has not been declared.                           ',
     & 'An assignment statement should be terminated by End-of-Line.',
     & 'JUST is expected in the WHEN statement. Found % instead.    ',
     & 'THEN is expected in the IF statement. Found % instead.      ',
     & 'BY is expected. Found %.                                    ',
     & 'Subprogram % is never used.                                 ',
     & 'Only the last dimension of array % can be adjustable.       ',
     & 'Array % is not a subprogram argument. Cannot be adjustable. ',
     & 'You have more than one main program.                        ',
     & 'Strings must be enclosed in single quotes ('').             ',
     & 'You must use a semicolon to separate arguments.             ',
     & 'Subprogram % occurs more than once.                         '/
      DATA (MESAGE(I),I=31,45)/
     & 'A left parenthesis should follow the % statement.           ',
     & 'A keyword is expected.                                      ',
     & 'The keyword % is not valid for this statement.              ',
     & 'A period (.) is expected after the keyword %.               ',
     & '"=" is expected.                                            ',
     & 'A right parenthesis is expected.                            ',
     & 'Inconsistent number of arguments for subprogram %.          ',
     & 'Keyword % accepts a string as a value and cannot be output. ',
     & 'You attempt to assign a value to the output keyword %.      ',
     & 'An identifier is expected after the output keyword %.       ',
     & 'The % keyword must be specified.                            ',
     & 'The % statement is misplaced.                               ',
     & 'Operand missing, or incorrect operator in expression.       ',
     & 'Misssing operand or missing operator in expression.         ',
     & 'Expression contains an illegal sequence.                    '/
      DATA (MESAGE(I),I= 46, 60)/
     & 'Empty expression into parenthesis or brackets.              ',
     & 'There are too many continuation lines. Maximum is %.        ',
     & 'Unmatched ")" or "]" or comma in wrong place.               ',
     & 'Unmatched left parenthesis or left bracket.                 ',
     & 'Identifier not found before left bracket.                   ',
     & 'A continuation mark is present at the last source line.     ',
     & 'An identifier is expected as the name of the subprogram.    ',
     & 'Only left parenthesis is allowed after the subprogram name. ',
     & 'Neither semicolon, nor right parenthesis was found.         ',
     & 'A low bound of array % is not a number.                     ',
     & 'A colon (:) is expected after a low bound of array %.       ',
     & 'An upper bound of array % is not a number.                  ',
     & 'Right bracket not found in the declaration of array %.      ',
     & 'Low bound greater than the upper bound in array %.          ',
     & 'An identifier is expected as a function name.               '/
      DATA (MESAGE(I),I= 61, 75)/
     & 'A left bracket is missing in the declaration of function %. ',
     & 'Only identifiers are allowed as dummy arguments.            ',
     & 'A right bracket must end the dummy argument list.           ',
     & '"=" is expected after the function name and arguments.      ',
     & 'Function % has already been declared.                       ',
     & 'Argument % is entered twice.                                ',
     & 'TO is expected. Found % instead.                            ',
     & 'A string is expected as file name after TO.                 ',
     & 'Right parenthesis missing at the end of the argument list.  ',
     & 'Actual argument #% is empty.                                ',
     & 'Nothing should follow the % command.                        ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'/
      DATA (MESAGE(I),I= 76, 90)/
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'Incorrect number of subscripts or arguments, for %.         ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'/
      DATA (MESAGE(I),I= 91,105)/
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'/
      DATA (MESAGE(I),I=106,120)/
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & '"=" is expected after an array element.                     ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'/
      DATA (MESAGE(I),I=121,135)/
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'Arithmetic expression expected into statement.              ',
     & 'A simple variable is expected as a loop control variable.   ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'Nested loops have the same loop control variable: %.        '/
      DATA (MESAGE(I),I=136,150)/
     & 'FROM is expected after the loop control variable.           ',
     & 'TO is expected.                                             ',
     & 'BY is expected.                                             ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ENDLOOP found, with no corresponding LOOP statement.        ',
     & 'The loop control variable % should not be assigned a value. ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'EXIT is illegal outside a LOOP-ENDLOOP structure.           ',
     & 'ELSE found without a corresponding IF.                      ',
     & 'ENDIF found without a corresponding IF.                     ',
     & 'One or more LOOPs do not terminate (END LOOP is missing).   ',
     & 'One or more block IFs do not terminate (END IF is missing). ',
     & 'Illegal jump into a block IF or LOOP.                       ',
     & 'FROM is expected after the list of variables.               ',
     & 'A string is expected as a file name, after FROM.            '/
      DATA (MESAGE(I),I=151,165)/
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & '% is not a simple variable, or array element.               ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'Simple variable followed by "(" or "[", into an expression. ',
     & 'Intrinsic variable followed by "(" or "[".                  ',
     & 'Reference to a function with no "[".                        ',
     & 'Simple array called with no "[" after its name.             ',
     & 'Intrinsic array called with no "[" after its name.          ',
     & 'More than one subscript in intrinsic array.                 ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'/
      DATA (MESAGE(I),I=166,180)/
     & 'The keyword % has been entered more than once.              ',
     & 'The loop control variable % should not be input a value.    ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'The keyword % requires a string value.                      ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ',
     & 'A PROGRAM or SUB statement must be present.                 ',
     & 'An expression is not allowed in a GET statement.            ',
     & 'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'/
      DATA (MESAGE(I),I=181,MAXMES)/
     & 'SUBRETURN can be used only in a subprogram.                 ',
     & 'A main program has not been found.                          ',
     & 'The subprogram % is missing.                                ',
     & 'An END statement is missing.                                ',
     & 'The value assigned to the % keyword is incorrect.           '/
C  ---------------------------------------------------------------------
C
C  Check if the message, input to the routine is longer than 10
C  characters. If so, truncate it.
      L = LENGTH(MESS)
      IF (L.LE.10) THEN
         WMESS = MESS
      ELSE
         L = L+3
         WMESS = MESS(1:10)//'...'
      END IF
C
C  Disable code generation.
      IF (ISERR.NE.0) GENE = .FALSE.
C
C  Update warnings counter.
      IF (ISERR.EQ.0) NWARN = NWARN+1
C
C  Substitute %
      MMESS = MESAGE(CODE)
      LEM = LENGTH(MMESS)
      IF (L.EQ.0) THEN
C  Nothing to substitute.
         PMESS = MMESS
         LEP = LEM
      ELSE
C  Substitute.
         LEP = 0
         DO 10,I=1,LEM
            C = MMESS(I:I)
            IF (C.EQ.'%') THEN
               PMESS(LEP+1:LEP+L) = WMESS(1:L)
               LEP = LEP+L
            ELSE
               LEP = LEP+1
               PMESS(LEP:LEP) = C
            END IF
10       CONTINUE
      END IF
C
C  Check if the error code should be displayed.
      IF (ISERR.EQ.0) THEN
         PREFIX = 'Warning'
         LEIX = 7
      ELSE
         PREFIX = 'Error'
         LEIX = 5
      END IF
C
C  Construct the the error line.
      IF (CDIS) THEN
         CALL I2STR(CODE,STR,LES)
         WRITE (ELINE,20) PREFIX(1:LEIX), ' code ',
     &                       STR(1:LES), PMESS(1:LEP)
      ELSE
         WRITE (ELINE,30) PREFIX(1:LEIX), PMESS(1:LEP)
      END IF
C
C  Put the error line in the error buffer.
      NERR = NERR+1
C
C  Add the error.
      IF (NERR.GT.MAXERR)
     &    CALL MABORT('Error buffer overflow.','ERROR')
      EBUFF(NERR) = ELINE
C
20    FORMAT (A,A,a,': ',A)
30    FORMAT (A,': ',a)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EXE ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the main routine that drives other program units according
C    to the statement to be processed. It handles any statement that may
C    appear next to a WHEN statement. Upon entry it assumes that the
C    first token encountered is an identifier. (A command, or the name
C    of a variable, in case of an assignment.)
C
C    EXE will first check if the next token in the sequence is an =
C    or a [. In such a case it will assume that an assignment takes
C    place. Otherwise it will check whether the first identifier upon
C    entry is one of the valid commands (defined in the array STAT),
C    and if so it will perform a computed goto and call the appropriate
C    routine to process the statement.
C    Object code is generated by this routine only in case of an
C    assignment:
C
C    <code to calculate memory address of the element to get the value>
C    <code to calculate the expression on the right side of the = sign>
C    $POPCONT.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  Notes:
C    If an array is being assigned a value, the whole left side of the
C    = sign is considered as an expression and subroutine XPR is called
C    to process it. XPR will generate code to calculate the array
C    element address and push it on the top of the stack. Then a
C    $GETCONT instruction will be generated by the same routine, so as
C    when executed, to replace the top of the stack with the actual
C    value of the array element.
C    The object code just described will be written in the spare file.
C    As soon as XPR returns, all of the spare file will be copied to
C    the scratch file, excluding the final $GETCONT instruction.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
      PARAMETER ( NMCOM = 99, MXPLUG = $(MXPLUG), NES = NMCOM+MXPLUG )
      CHARACTER*14 STAT(NES)
      COMMON /COM1/ STAT
      INTEGER NAME(NES), INDE(NES)
      COMMON /COM2/ NAME, INDE
C  ---------------------------------------------------------------------
C  Local variables:
C
C  Line-length of the scratch file.
      PARAMETER ( LLSCR = 80 )
C
      CHARACTER*(IDLEN) VARID, STARG(1), IDF
      CHARACTER*(LLSCR) LINE
      INTEGER EX(LINLEN)
      LOGICAL FOUND, CORRE
C  ---------------------------------------------------------------------
C
      NT = NT+1
C
C  Get the statement, and check if its an assignment.
      CALL GETIDF(NA,NN,VARID)
      CUR = TOKEN(NT+1)
      IF (CUR.EQ.EQ .OR. CUR.EQ.LBRA) THEN
C  This is an assignment.
         NT = NT+1
C  Check if variable present in the var table.
         CALL LOCAN(VAR,NVAR,VARID,FOUND,IND)
         IF (.NOT.FOUND) THEN
C  Variable not found in the var table.
            CALL ERROR(19,VARID)
C  Put variable in the var table. Update related tables also.
            CALL INSERT(VARID,0,0,0,-1,IND)
         END IF
C
C  Calculate address of the element that accepts the value.
         IF (CUR.EQ.EQ) THEN
            K = KIND(IND)
C  If an = sign was found, only simple variables or variables that
C  have not been declared are allowed.
            IF (K.NE.1 .AND. K.NE.-1 .AND. K.NE.101)
     &                CALL ERROR(153,VARID)
C  Check whether the variable is used as a loop control index.
            IF (ISLCI(IND).NE.0) CALL ERROR(141,VARID)
            L = LOC(IND)
C  Generate necessary code to push the address of the variable on
C  the top of the stack.
            IF (K.EQ.1) THEN
               BLOC = BLOC+2
               IF (GENE) THEN
                  WRITE (SCR,'(A)') '$PUSHA'
                  WRITE (SCR,*) L
               END IF
            ELSE IF (K.EQ.101) THEN
               CALL WPSC(SCR,BLOC,0)
               CALL WPSC(SCR,BLOC,L)
               BLOC = BLOC+1
               IF (GENE) WRITE (SCR,'(A)') '$IOFFSET'
            END IF
C  CORRE indicates whether calculation of the address of the element
C  on the left side of the = sign was successful.
            CORRE = .TRUE.
         ELSE
C  This is the case of an array element being assigned a value.
            NT = NT-2
            NA = NA-1
            NN = NN-2
C  Extract the array element from the sequence of tokens.
            CALL EXTRA(NT,EX,NEX,3,NA)
C  NLS counts the object code lines necessary for the evaluation of
C  the memory address of the array element.
            NLS = 0
            OPEN (SPARE,STATUS='SCRATCH',IOSTAT=IERR)
            IF (IERR.NE.0)
     &            CALL MABORT('Cannot not open the spare file.','EXE')
            REWIND SPARE
C  The array element is processed as an expression. Object code
C  will be written on the spare file, while the number of lines
C  written on this file is counted by NLS.
            CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,NLS,SPARE,STARG)
            IF (GENE) THEN
C  Copy all of the spare file to the scratch file excluding the
C  last line. (A $GETCONT instruction )
               REWIND SPARE
               DO 88,I=1,NLS-1
                  READ (SPARE,'(A)') LINE
                  WRITE (SCR,'(A)') LINE
88             CONTINUE
C  Update the scratch file counter.
               BLOC = BLOC+NLS-1
            END IF
            CLOSE (SPARE,IOSTAT=IERR)
            IF (IERR.NE.0)
     &            CALL MABORT('Cannot close the spare file.','EXE')
C  Array element location is now on the top of the stack.
            IF (CORRE) THEN
C  Confirm that an = sign follows.
               NT = NT+1
               IF (TOKEN(NT).NE.EQ) THEN
                  CALL ERROR(114,' ')
                  RETURN
               END IF
            END IF
         END IF
C
C  Process expression on the left side of the "=" sign.
         IF (CORRE) THEN
C  Extract the expression from the sequence of tokens.
            CALL EXTRA(NT,EX,NEX,0,NA)
C  Evaluate the expression.
            CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C  If the expression is incorrect.
            IF (.NOT.CORRE) RETURN
C  Write the final $POPCONT instruction.
            BLOC = BLOC+1
            IF (GENE) WRITE (SCR,'(A)') '$POPCONT'
C  Check if anything else but an EOL follows the expression.
            NT = NT+1
            IF (NT.NE.NTOK) CALL ERROR(20,' ')
         END IF
C
C  Check if the variable that accepts the expression's value is a
C  Merlin variable, Merlin array or intrinsic function.
         K = KIND(IND)
         IF (K.EQ.4 .OR. K.EQ.5 .OR. K.EQ.6) THEN
            CALL ERROR(18,VARID)
            RETURN
         ELSE IF (K.EQ.3) THEN
            CALL ERROR(153,VARID)
         END IF
      ELSE
C  Neither = nor [ follows the first token.
C  This is not the case of an assignment.
C  Check if command present in array STAT.
         CALL LOCAN(STAT,NES,VARID,FOUND,IND)
         IF (FOUND) THEN
C
C  Jump to the appropriate label, and process the statement.
C
            IX = INDE(IND)
            IF (IX.GE.1 .AND. IX.LE.NMCOM) THEN
               GOTO (  1,  1,  1, 25, 24,  7, 24, 24, 24, 14,
     &                         24, 24,  1, 24, 20,  1, 10,  5,  1,  7,
     &                          1,  1,  1, 15,  8,  23, 1,  1,  1, 24,
     &                         24,  1,  1,  1,  5,  7, 24,  1, 13, 12,
     &                          1,  1,  1,  1,  1, 24, 24, 10, 24,  7,
     &                          1, 26,  1, 24,  1, 24,  1,  7,  1,  1,
     &                         24,  1,  1,  1,  1, 21, 22,  5,  5,  8,
     &                          8,  5, 23, 23, 23, 23, 23, 23, 23, 23,
     &                         23, 23, 23, 23, 23, 23, 23, 23, 24, 24,
     &                         24, 24, 24,  1, 24,  5, 24,  1,  1), IX
C
               CALL MABORT('Too few gotos.','EXE')
            ELSE IF (IX.GT.NMCOM .AND. IX.LE.NES) THEN
               CALL MULTI(NT,NA,NN,VARID)
            ELSE
               CALL MABORT('Invalid "ind".','EXE')
            END IF
         ELSE
C  The statement was not found in array STAT.
            CALL ERROR(7,' ')
         END IF
      END IF
      RETURN
C  .........................................................  NOPAR
C
C  No parameter commands, already in the Merlin operating system.
1     CONTINUE
      CALL NOPAR(NT,VARID,0)
      RETURN
C  .........................................................  CHI
C
C  COMMAND ( KEY.INDEX )
5     CONTINUE
      ID = NAME(IND)
      CALL CHI(ID,NT,NA,NN)
      RETURN
C  .........................................................  CHIV
C
C  COMMAND ( KEY.INDEX = VALUE )      VALUE is a number.
7     CONTINUE
      ID = NAME(IND)
      CALL CHIV(NT,NA,NN,ID)
      RETURN
C  .........................................................  CHIS
C
C  COMMAND ( KEY.INDEX = 'VALUE' )    VALUE is a string.
8     CONTINUE
      ID=NAME(IND)
      CALL CHIS(ID,NT,NA,NN)
      RETURN
C  .........................................................  $NOPAR
C
C  No parameter commands added to Merlin for MCL use only.
10    CONTINUE
      CALL NOPAR(NT,VARID,1)
      RETURN
C  .........................................................  MOVETO
C
C  MOVE TO (Entered as "MOVETO").
12    CONTINUE
      CALL MOVE(NT,NA,NN)
      RETURN
C  .........................................................  MOVE TO
C
C  MOVE TO (Entered as "MOVE TO").
13    CONTINUE
C  After MOVE an identifier is expected (TO).
      NT = NT+1
      IF (TOKEN(NT).EQ.IDENT) THEN
C  Check if the identifier is TO.
         CALL GETIDF(NA,NN,IDF)
         IF (IDF.EQ.'TO') THEN
            CALL MOVE(NT,NA,NN)
         ELSE
C  TO does not follow MOVE.
            CALL ERROR(15,IDF)
         END IF
      ELSE
C  An identifier does not follow MOVE.
         CALL ERROR(17,' ')
      END IF
      RETURN
C  .........................................................  DISPLAY
C
14    CONTINUE
      CALL DISPLA(NT,NA,NN)
      RETURN
C  .........................................................  GET
C
15    CONTINUE
      CALL GET(NT,NA,NN)
      RETURN
C  .........................................................  EXIT
C
20    CONTINUE
      CALL EXITL(NT)
      RETURN
C  .........................................................  CALL
C
21    CONTINUE
      CALL SUBCAL(NT,NA,NN)
      RETURN
C  .........................................................  SUBRETURN
C
22    CONTINUE
      CALL SUBRET(NT)
      RETURN
C  .........................................................  GKS
C
C  COMMAND ( KEY = 'VALUE' )       VALUE is a string.
23    CONTINUE
      ID = NAME(IND)
      CALL GKS(ID,NT,NA,NN)
      RETURN
C  .........................................................  MULTI
C
C  Panel commands.
24    CONTINUE
      CALL MULTI(NT,NA,NN,VARID)
      RETURN
C  .........................................................  TARGET
C
25    CONTINUE
      CALL TARGET(NT,NA,NN)
      RETURN
C  .........................................................  QUIT
C
26    CONTINUE
      CALL QUIT(NT,NA,NN)
      RETURN
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EXITL ( NT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Handles the EXIT statement. An MCL error will be issued if the
C    statement is not used inside a loop structure.
C    The object code generated is:
C
C    $MOVE
C    <blank line>
C
C    The target of the $MOVE instruction, is the $CONTINUE instruction
C    generated by the ENDLOOP statement of the loop.
C    A value of 8 is pushed in the type table (to indicate the EXIT
C    statement), while the location in the scratch file of the above
C    <blank line> is pushed in the type-definition table.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL ERR
C  ---------------------------------------------------------------------
C
C  ERR will be set to .TRUE. If EXIT is followed by anything else
C  but an EOL, or if the statement is used outside a loop structure.
      ERR = .FALSE.
C
C  If anything else than an EOL follows the EXIT statement,
C  this is an error.
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL ERROR(9,'EXIT')
         ERR = .TRUE.
      END IF
C
C  EXIT is valid only inside a loop.
      IF (NLOO.EQ.0) THEN
         CALL ERROR(143,' ')
         ERR = .TRUE.
      END IF
C
      IF (ERR) RETURN
C
C  Generate code.
      BLOC = BLOC+2
      IF (GENE) WRITE (SCR,'(A/)') '$MOVE'
C
C  Update the type and type-definition tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','EXITL',MXTYP,'MXTYP')
      TYPE(NTYPE) = 8
      TDEF(NTYPE) = BLOC
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE EXTRA ( NT, EX, NEX, IOPT, NA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Extracts from a line, all tokens, valid in an MCL expression and
C    stores them in an array that is returned as output.
C    The line is given in the form of tokens. Expression scanning always
C    stops when an invalid token is reached. Additionally according to
C    an input option, expression scanning stops when different
C    situations are met.
C
C    A token is accepted in an expression sequence, if it's one
C    of the following:
C
C    Right parenthesis       )           Left parenthesis         (
C    Multiplication          *           Division                 /
C    Addition                +           Subtraction              -
C    Identifier                          Unsigned number
C    Raise to a power       **           Comma                    ,
C    Right bracket           ]           Left bracket             [
C    Greater than            >           Less than sign           <
C    Greater or equal than  >=           Less or equal than      <=
C    Equality operator      ==           Not equal                #
C    Exclusive or operator               Not operator
C    And operator                        Or operator
C
C  Input arguments:
C    IOPT       Integer variable. This is an option.
C               It may take on the values:
C               0  Expression scanning stops at the first invalid
C                  token reached.
C               1  Expression scanning stops also when one of the
C                  identifiers: THEN, JUST, FROM, BY, TO is encountered.
C               2  Expression scanning stops also when a right
C                  parenthesis is reached before the end of the line.
C               3  Expression scanning stops also at the equality
C                  sign ( = ) appearing after a right bracket.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  Output arguments:
C    EX      Integer array. Upon return, contains the valid tokens
C            that have been found in the token array.
C    NEX     Integer variable. Number of entries in array EX.
C
C  Input / Output arguments:
C    NT      When the routine is entered, this is the number of tokens
C            used by the previously called parsing modules. When this
C            routine returns, this is the total number of tokens
C            (including the ones extracted as an expression) used
C            from the current MCL source line.
C
C  Abort conditions that may occur:
C    If an end of line code is not present in the token array
C    and all previous codes are accepted in an expression, the
C    program will stop and issue an informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER EX(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL RESE, OK
      CHARACTER*(IDLEN) IDF
C
C  The following statement function will test if a token is accepted
C  in an MCL expression.
      OK(M) = M.EQ.RIGHTP .OR. M.EQ.LEFTP .OR.
     &        M.EQ.STAR   .OR. M.EQ.SLASH .OR.
     &        M.EQ.PLUS   .OR. M.EQ.MINUS .OR.
     &        M.EQ.IDENT  .OR. M.EQ.UNSI  .OR.
     &        M.EQ.BISTAR .OR. M.EQ.COMMA .OR.
     &        M.EQ.RBRA   .OR. M.EQ.LBRA  .OR.
     &        M.EQ.GT     .OR. M.EQ.LT    .OR.
     &        M.EQ.GE     .OR. M.EQ.LE    .OR.
     &        M.EQ.EQUAL  .OR. M.EQ.NE    .OR.
     &        M.EQ.XOR    .OR. M.EQ.NOT   .OR.
     &        M.EQ.AND    .OR. M.EQ.OR
C
C  INIT is the next available token to be searched.
      INIT = NT+1
      NNA = NA
C
C  NEX is the number of tokens in the expression stream.
      NEX = 0
C
C  The following do-loop will check all tokens in the token array.
      DO 10,I=INIT,NTOK
         CUR = TOKEN(I)
         IF (OK(CUR)) THEN
C  The token is accepted in the expression.
            IF (IOPT.EQ.1 .AND. CUR.EQ.IDENT) THEN
C  Check if one of the reserved identifiers.
               NNA = NNA+1
               CALL GETIDA(NNA,IDF)
               IF (RESE(IDF)) THEN
                  NT = I-1
                  RETURN
               END IF
            ELSE IF (IOPT.EQ.3 .AND. I.NE.INIT) THEN
               IF (CUR.EQ.EQ .AND. TOKEN(I-1).EQ.RBRA) THEN
                  NT = I-1
                  RETURN
               END IF
            END IF
            NEX = NEX+1
            EX(NEX) = CUR
         ELSE
C  An invalid token has been found.
            IF (IOPT.EQ.2  .AND.
     &         CUR.EQ.EOL .AND. TOKEN(I-1).EQ.RIGHTP) THEN
               NT = I-2
               NEX = NEX-1
            ELSE
               NT = I-1
            END IF
            RETURN
         END IF
 10   CONTINUE
C
C  Tokens in the current line have been exhausted, yet no
C  invalid token has been found. (At least one invalid token
C  should be present: the end of line.)
      CALL MABORT('EOL not found.','EXTRA')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FCLOSE
C  ---------------------------------------------------------------------
C
C  Description:
C    Closes the MCL files.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      CLOSE (SCR,STATUS='DELETE',IOSTAT=IERR1)
      CLOSE (BIN,STATUS='DELETE',IOSTAT=IERR2)
      CLOSE (STF,STATUS='DELETE',IOSTAT=IERR3)
      CLOSE (FHELP,STATUS='DELETE',IOSTAT=IERR4)
      IERR = 0
      IF (IERR1.NE.0) IERR= IERR1
      IF (IERR2.NE.0) IERR= IERR2
      IF (IERR3.NE.0) IERR= IERR3
      IF (IERR4.NE.0) IERR= IERR4
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LES)
         CALL HEAD
         CALL DASH
         WRITE (ERL,20) STR(1:LES)
         CALL DASH
         STOP
      END IF
C
      IF (GENE) THEN
         CLOSE (ABSFI,STATUS='KEEP',IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR,LES)
            CALL HEAD
            CALL DASH
            WRITE (ERL,10) 'object code', BFILE(1:LENGTH(BFILE)),
     &                   STR(1:LES)
            CALL DASH
            STOP
         END IF
         IF (EOPEN) CLOSE (ERL,STATUS='DELETE')
         WRITE (*,30) BFILE(1:LENGTH(BFILE))
         STOP
      ELSE
         CLOSE (ABSFI,STATUS='DELETE',IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR,LES)
            CALL HEAD
            CALL DASH
            WRITE (ERL,10) 'object code', BFILE(1:LENGTH(BFILE)),
     &                   STR(1:LES)
            CALL DASH
            STOP
         END IF
         IF (EOPEN) THEN
            CLOSE (ERL,STATUS='KEEP')
            WRITE (*,40) EFILE(1:LENGTH(EFILE))
            STOP
         ELSE
            STOP
         END IF
      END IF
C
10    FORMAT ($(FC)'Cannot close the ',A,' file "',A,
     &        '". IOSTAT = ',A,'.')
20    FORMAT ($(FC)'Cannot close one or more scratch files. ',
     &        'IOSTAT = ',A,'.')
30    FORMAT ($(FC)'Complilation succesful, output is in file "',
     &        A,'".')
40    FORMAT ($(FC)'Compilation failed, see the error file "',
     &        A,'".')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FEBUFF ( LINE, TOTAL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Flushes the error buffer on the error list file.
C    Assumes that upon entry the buffer contains at least one error.
C
C  Input arguments:
C    LINE       The MCL source line that is currently under processing
C               and is responsible for the error. If LINE = ' ' the
C               error occurred outside a source line (eg. when linking
C               subprograms) and will not be printed on the error
C               list file.
C    TOTAL      The total number of errors encountered in the program,
C               so far.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) LINE
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      PARAMETER ( MAXERR = 20 )
      CHARACTER*90 EBUFF(MAXERR)
      COMMON /MCLERB/ EBUFF
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C, TAB, CNL*10
C  ---------------------------------------------------------------------
C
      CALL HEAD
C
C  Write the MCL source line and its line number as well.
      LL = LENGTH(LINE)
      IF (LL.NE.0) THEN
         TAB = CHAR(9)
         DO 20,IS=1,LL
            C = LINE(IS:IS)
            IF (C.NE.' ' .AND. C.NE.TAB) GOTO 21
20       CONTINUE
         CALL MABORT('Empty line.','FEBUFF')
21       CONTINUE
         CALL I2STR(NL,CNL,LEC)
         LACT = LL-IS+1
         IF (LACT.GT.80) LL = IS+79
         WRITE (ERL,30) CNL(1:LEC), LINE(IS:LL)
      ELSE
         WRITE (ERL,*)
      END IF
C
C  Flush the error buffer.
      DO 10,I=1,NERR
         LL = LENGTH(EBUFF(I))
         WRITE (ERL,40) EBUFF(I)(1:LL)
10    CONTINUE
C
C  Update the total error counter.
      TOTAL = TOTAL+NERR
C
C  No more errors present in the error buffer.
      NERR = 0
C
30    FORMAT (/$(FC)'Line ',A,': ',A)
40    FORMAT ($(FC)A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE FOPEN
C  ---------------------------------------------------------------------
C
C  Description:
C    Opens the MCL files. Aborts on errors.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
      LOGICAL EXI
C  ---------------------------------------------------------------------
C
C  Open the error list file.
      IF (EFILE.NE.' ') THEN
         ERL = NUNIT()
         IF (ERL.EQ.-1) THEN
            CALL HEAD
            CALL DASHO
            WRITE (*,10)
            CALL DASHO
            STOP
         END IF
         OPEN (ERL,FILE=EFILE,IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR,LES)
            CALL HEAD
            CALL DASHO
            WRITE (*,30) 'error list', EFILE(1:LENGTH(EFILE)),
     &                           STR(1:LES)
            CALL DASHO
            STOP
         END IF
         EOPEN = .TRUE.
      END IF
C
C  Dynamically allocate units for all MCL files.
      INP = NUNIT()
      BIN = NUNIT()
      SCR = NUNIT()
      STF = NUNIT()
      SPARE = NUNIT()
      FHELP = NUNIT()
      ABSFI = NUNIT()
      IF (INP.EQ.-1 .OR. BIN.EQ.-1 .OR. SCR.EQ.-1  .OR. STF.EQ.-1 .OR.
     &    SPARE.EQ.-1. OR. FHELP.EQ.-1 .OR. ABSFI.EQ.-1) THEN
         CALL HEAD
         CALL DASH
         WRITE (ERL,20)
         CALL DASH
         STOP
      END IF
C
C  Open the input file.
      INQUIRE (FILE=IFILE,EXIST=EXI,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LES)
         CALL HEAD
         CALL DASH
         WRITE (ERL,50) IFILE(1:LENGTH(IFILE)), STR(1:LES)
         CALL DASH
         STOP
      END IF
      IF (.NOT.EXI) THEN
         CALL HEAD
         CALL DASH
         WRITE (ERL,60) IFILE(1:LENGTH(IFILE))
         CALL DASH
         STOP
      END IF
      OPEN (INP,FILE=IFILE,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LES)
         CALL HEAD
         CALL DASH
         WRITE (ERL,30) 'input', IFILE(1:LENGTH(IFILE)), STR(1:LES)
         CALL DASH
         STOP
      END IF
C
C  Open the final MOC file.
      OPEN (ABSFI,FILE=BFILE,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LES)
         CALL HEAD
         CALL DASH
         WRITE (ERL,30) 'object code', BFILE(1:LENGTH(BFILE)),
     &                     STR(1:LES)
         CALL DASH
         STOP
      END IF
C
C  Open the scratch files.
      OPEN (SCR,STATUS='SCRATCH',IOSTAT=IERR1)
      OPEN (BIN,STATUS='SCRATCH',IOSTAT=IERR2)
      OPEN (STF,STATUS='SCRATCH',IOSTAT=IERR3)
      OPEN (FHELP,STATUS='SCRATCH',IOSTAT=IERR4)
      IERR = 0
      IF (IERR1.NE.0) IERR= IERR1
      IF (IERR2.NE.0) IERR= IERR2
      IF (IERR3.NE.0) IERR= IERR3
      IF (IERR4.NE.0) IERR= IERR4
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LES)
         CALL HEAD
         CALL DASH
         WRITE (ERL,40) STR(1:LES)
         CALL DASH
         STOP
      END IF
C
10    FORMAT ($(FC)'Cannot allocate the output unit.')
20    FORMAT ($(FC)'Cannot allocate unit numbers.')
30    FORMAT ($(FC)'Cannot open the ',A,' file "',A,
     &        '". IOSTAT = ',A,'.')
40    FORMAT ($(FC)'Cannot open one or more scratch files. ',
     &        'IOSTAT = ',A,'.')
50    FORMAT ($(FC)'An error occurred while inquiring the input ',
     &        'file "',A,'". IOSTAT = ',A,'.')
60    FORMAT ($(FC)'The input file "',A,'" does not exist.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GET ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine parses the GET statement and generates appropriate
C    object code. The syntax of GET statement is:
C
C    GET var1; var2; ...varN       or
C    GET var1; var2; ...varN FROM <filename>
C
C    where var1, var2, ...varN is a list of variables, and
C    <Filename> is a filename (a string enclosed in quotes.)
C
C    The routine will first determine the number of variables to get
C    by counting the semicolons in the input line. Then it will
C    determine the input <filename>. If its omitted the default Merlin
C    input file will be used.
C    Only simple variables (not used as loop control variables) or
C    array elements (declared by the user) may appear in the list of
C    variables. If any of the variables is not declared, or is not of
C    the type mentioned above, an MCL error will be issued.
C    The object code generated is:
C
C    $IN
C    <number of variables>
C    <filename>
C    <code to calculate memory address for var1>
C    $POPLV
C    <code to calculate memory address for var2>
C    $POPLV
C    ...
C    <Code to calculate memory address for varN>
C    $POPLV
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  Abort conditions that may occur:
C    This routine uses the spare file, so when its opening or closing
C    fails, the program will stop and issue an informative message.
C
C  Notes:
C    If an array element appears in the list of variables, it is
C    considered as an expression and subroutine XPR is called
C    to process it. XPR will generate code to calculate the array
C    element address and push it on the top of the stack. Then a
C    $GETCONT instruction will be generated by the same routine, so as
C    when executed, to replace the top of the stack with the actual
C    value of the array element.
C    The object code just described will be written in the spare file.
C    As soon as XPR returns, all of the spare file will be copied to
C    the scratch file, excluding the final $GETCONT instruction.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LLSCR = 80 )
      CHARACTER LINE*(LLSCR)
      CHARACTER*(IDLEN) NAME, STARG(1), IDF
      LOGICAL FOUND, CORRE
      INTEGER EX(LINLEN)
C  ---------------------------------------------------------------------
C
C  Calculate NG, (that is the number of variables to get) by
C  counting the number of semicolons in the input line.
      NG = 1
      DO 10,I=NT+1,NTOK
10       IF (TOKEN(I).EQ.SEMI) NG = NG+1
C
C  Determine the input file name.
      IF (NTOK.GE.7) THEN
         TK1 = TOKEN(NTOK-1)
         TK2 = TOKEN(NTOK-2)
         TK3 = TOKEN(NTOK-3)
         TK4 = TOKEN(NTOK-4)
         IF (TK1.EQ.QUOTE .AND. TK2.EQ.STRING .AND.
     &          TK3.EQ.QUOTE .AND. TK4.EQ.IDENT) THEN
            NID =NA
            DO 99,I=NT+1,NTOK
               TK = TOKEN(I)
               IF (TK.EQ.IDENT .OR. TK.EQ.STRING) NID=NID+1
99          CONTINUE
            CALL GETIAS(NID,IS,IE)
            NAME = ALPHA(IS:IE)
         ELSE
            NAME = '.STANDARD'
         END IF
      ELSE
         NAME = '.STANDARD'
      END IF
C
C  Write initial lines of object code.
      BLOC = BLOC+3
      IF (GENE) WRITE (SCR,'(A/I4/A)') '$IN', NG, NAME
C
C  Start a loop for all variables in the variable list.
40    CONTINUE
C
C  Check to see if the next token is an identifier.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(3,' ')
         RETURN
      END IF
C
C  Get the name of the variable.
      CALL GETIDF(NA,NN,NAME)
C
C  Check if present in the var table.
      CALL LOCAN(VAR,NVAR,NAME,FOUND,IND)
      IF (.NOT.FOUND) THEN
C  Variable not present in the var table.
         CALL ERROR(19,NAME)
C  Put variable in the var table. Update related tables as well.
         CALL INSERT(NAME,0,0,0,-1,IND)
      END IF
C
C  Calculate memory address for this variable.
      K = KIND(IND)
      IF (TOKEN(NT+1).NE.LBRA) THEN
C  The next token is not a left bracket. Thus this is a simple
C  variable to get.
         IF (K.NE.1 .AND. K.NE.-1 .AND. K.NE.101) CALL ERROR(153,NAME)
C  Check whether variable is used as a loop control variable.
         IF (ISLCI(IND).NE.0) CALL ERROR(167,NAME)
         L = LOC(IND)
C  Generate code necessary for its address evaluation.
         IF (K.EQ.1) THEN
            BLOC =BLOC+2
            IF (GENE) THEN
               WRITE (SCR,'(A)') '$PUSHA'
               WRITE (SCR,*) L
            END IF
         ELSE
            CALL WPSC(SCR,BLOC,0)
            CALL WPSC(SCR,BLOC,L)
            BLOC = BLOC+1
            IF (GENE) WRITE (SCR,'(A)') '$IOFFSET'
         END IF
      ELSE
         IF (K.EQ.3) CALL ERROR(153,NAME)
C  This is an array element.
         NT = NT-1
         NA = NA-1
         NN = NN-2
C  Extract the array element as an expression.
         CALL EXTRA(NT,EX,NEX,1,NA)
C
C  Check whether an expression is involved. (Only an array element
C  is allowed.) Expressions for the array subscript are ok.
         NBR = 1
         DO 30,IEX=3,NEX
            IF (EX(IEX).EQ.LBRA) NBR = NBR+1
            IF (EX(IEX).EQ.RBRA) NBR = NBR-1
            IF (NBR.EQ.0) THEN
               IF (IEX.NE.NEX) CALL ERROR(179,' ')
               GOTO 31
            END IF
30       CONTINUE
         CALL ERROR(179,' ')
31       CONTINUE
C
C  Evaluate this expression. The object code will be written
C  in the spare file. The number of object code lines written
C  there is counted by NLS.
         NLS = 0
         OPEN (SPARE,STATUS='SCRATCH',IOSTAT=IERR)
         IF (IERR.NE.0)
     &          CALL MABORT('Cannot open the spare file.','GET')
         REWIND SPARE
         CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,NLS,SPARE,STARG)
         IF (GENE) THEN
C
C  Copy the spare file to the scratch file. The final $GETCONT
C  instruction will not be written in the scratch file.
            REWIND SPARE
            DO 88,I=1,NLS-1
               READ (SPARE,'(A)') LINE
               WRITE (SCR,'(A)') LINE
88          CONTINUE
            BLOC = BLOC+NLS-1
         END IF
         CLOSE (SPARE,STATUS='DELETE',IOSTAT=IERR)
         IF (IERR.NE.0)
     &          CALL MABORT('Cannot open the spare file.','GET')
C  Array element location is now on the top of the stack.
      END IF
C
C  Check if the name is one of the merlin variables,
C  merlin arrays or intrinsic functions.
      IF (K.EQ.4 .OR. K.EQ.5 .OR. K.EQ.6) CALL ERROR(18,NAME)
C
C  Write the $POPLV instruction for this variable.
      BLOC = BLOC+1
      IF (GENE) WRITE (SCR,'(A)') '$POPLV'
C
      NT = NT+1
C
C  If all tokens exhausted.
      IF (NT.EQ.NTOK) RETURN
      CUR = TOKEN(NT)
C
C  If there's a semicolon, another variable should follow in the list.
      IF (CUR.EQ.SEMI) GOTO 40
C
C  If an identifier follows, this should be FROM.
      IF (CUR.EQ.IDENT) THEN
         CALL GETIDF(NA,NN,IDF)
         IF (IDF.EQ.'FROM') THEN
            NT = NT+1
            IF (TOKEN(NT).EQ.QUOTE) THEN
C  A string (the <filename>) should follow after FROM.
               NT = NT+1
               IF (TOKEN(NT).EQ.STRING) THEN
                  NT = NT+1
                  IF (TOKEN(NT).EQ.QUOTE) THEN
                     NT = NT+1
                     IF (NT.NE.NTOK) CALL ERROR(9,'GET')
                  ELSE
                     CALL ERROR(150,' ')
                  END IF
               ELSE
                  CALL ERROR(150,' ')
               END IF
            ELSE
               CALL ERROR(150,' ')
            END IF
         ELSE
C  FROM does not follow the list of variables.
            CALL ERROR(149,' ')
         END IF
         RETURN
      END IF
C  Neither semicolon, nor an identifier was found after
C  the list of variables.
      CALL ERROR(149,' ')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETIAS ( NID, IS, IE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETIAS returns the NID-th entry of the ALPHA table. This
C    entry is assumed to be an identifier and is returned as a substring
C    of the ALPHA table, using two pointers: ALPHA(IS:IE)
C
C  Input arguments:
C    NID        The ALPHA-table entry to be returned.
C
C  Output arguments:
C    IS         Starting character position of the string, in the ALPHA
C               table.
C    IE         Ending character position of the string, in the ALPHA
C               table.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C
      NNN = 0
      NNA = 0
      DO 10,I=1,NTOK
         TK = TOKEN(I)
         IF (TK.EQ.UNSI) THEN
            NNN = NNN+1
         ELSE IF (TK.EQ.STRING) THEN
            NNA = NNA+1
            IF (NID.EQ.NNA) THEN
               IS = NINT(NUM(NNN+1))
               IE = NINT(NUM(NNN+2))
               IF (IE-IS+1.LT.0)
     &                      CALL MABORT('Negative length.','GETIAS')
               RETURN
            ELSE
               NNN = NNN+2
            END IF
         ELSE IF (TK.EQ.IDENT) THEN
            NNA = NNA+1
            NNN = NNN+2
         END IF
10    CONTINUE
      CALL MABORT('Tokens exausted.','GETIAS')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETIDA ( NID, ID )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GETIDA returns the NID-th entry of the ALPHA table. This
C    entry is assumed to be an identifier.
C
C  Input arguments:
C    NID        The ALPHA-table entry to be returned.
C
C  Output arguments:
C    ID         The identifier.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) ID
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C
      NNN = 0
      NNA = 0
      DO 10,I=1,NTOK
         TK = TOKEN(I)
         IF (TK.EQ.UNSI) THEN
            NNN = NNN+1
         ELSE IF (TK.EQ.STRING) THEN
            NNA = NNA+1
            NNN = NNN+2
         ELSE IF (TK.EQ.IDENT) THEN
            NNA = NNA+1
            IF (NID.EQ.NNA) THEN
               IS = NINT(NUM(NNN+1))
               IE = NINT(NUM(NNN+2))
               LES = IE-IS+1
               IF (LES.EQ.0) THEN
                  ID = ' '
               ELSE IF (LES.GT.0) THEN
                  ID = ALPHA(IS:IE)
               ELSE
                  CALL MABORT('Negative length.','GETIDA')
               END IF
               RETURN
            ELSE
               NNN = NNN+2
            END IF
         END IF
10    CONTINUE
      CALL MABORT('Tokens exausted.','GETIDA')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETIDF ( NA, NN, NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the next, ( (NA+1)-th ) entry of the ALPHA table, updating
C    the NA and NN pointers. This entry is assumed to be an identifier,
C    and is returned in a single character variable (NAME).
C
C  Output arguments:
C    NAME       The identifier.
C
C  Input / Output arguments:
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  ---------------------------------------------------------------------
C
      NN = NN+1
      IS = NINT(NUM(NN))
      NN = NN+1
      IE = NINT(NUM(NN))
      NA = NA+1
      LES = IE-IS+1
C
      IF (LES.EQ.0) THEN
         NAME = ' '
      ELSE IF (LES.GT.0) THEN
         NAME = ALPHA(IS:IE)
      ELSE
C  This should never happen !!!!!
         CALL MABORT('Negative length.','GETIDF')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETIDS ( NA, NN, IS, IE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the next, ( (NA+1)-th ) entry of the ALPHA table, updating
C    the NA and NN pointers. This entry is assumed to be a string,
C    and is returned as a substring of the ALPHA table, using two
C    pointers: ALPHA(IS:IE)
C
C  Output arguments:
C    IS         Starting character position of the string, in the ALPHA
C               table.
C    IE         Ending character position of the string, in the ALPHA
C               table.
C
C  Input / Output arguments:
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  ---------------------------------------------------------------------
C
      NN = NN+1
      IS = NINT(NUM(NN))
      NN = NN+1
      IE = NINT(NUM(NN))
      NA = NA+1
C
      IF (IE-IS+1.LE.0) CALL MABORT('Length not positive.','GETIDS')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GETPAR
C  ---------------------------------------------------------------------
C
C  Description:
C    Displays the compilers initial message and prompts the user
C    to enter the compilation parameters.
C    These are entered in one line, up to 80 characters long.
C    The input line must have the form:
C      K1 = V1, K2 = V2, ..., Kn = Vn
C    where K1, K2, ..., Kn are keywords and V1, V2, ..., Vn is the
C    value assigned to them.
C    The keywords, their significance and the values that should
C    be assigned to them are as follows:
C
C    Keyword     Significance             Allowed values
C    -------     ------------             ---------------
C    I           the MCL source           Any file name
C    E           the error list file      Any file name
C    B           the object code file     Any file name
C    DEBUG       the DEBUG parameter      true, false, t, f
C    BOUNDS      the BOUNDS parameter     true, false, t, f
C
C    The file names specified as input to the compiler may
C    contain special characters (such as . / + - * etc) and may
C    be up to MXFILE (80) characters at most. This is particularly
C    useful, since most operating systems allow special characters
C    in a file name. Note that the comma may not be used in a file
C    name since it serves separation purposes in the input line.
C
C    The input file (the I parameter) must be always specified.
C    The other parameters when not explicitly declared
C    assume default values as:
C
C    Parameter          Default value
C    ---------          -------------
C    E                  ERRORS (may be changed in the Merlin
C                               configuration file)
C    B                  MOC    (may be changed in the Merlin
C                               configuration file)
C    DEBUG              FALSE
C    BOUNDS             FALSE
C
C    The above parameters obtain their default values in the
C    block data subprogram MCLDEF.
C    Blanks are allowed in the input line at any point.
C    Duplicate entries of keywords are not allowed. Thus the
C    following line is incorrect:
C      i=inp, b=bin, e=err, b=bin
C    If a file name is specified as a value to a keyword, it may
C    not be specified as a value to another keyword.
C    For this reason the following line is incorrect.
C      i=file1, e=xfile, bounds=t, b=xfile
C    if any of the above rules is violated, or the input line is
C    incorrectly formatted, the program will stop and issue
C    an informative message.
C
C    Upon return of this routine, the values of IFILE, EFILE and BFILE
C    in the /FILES/ common block will be updated. The values of
C    BOUNDS and DEBUG in the /MCLDTA/ common block will also be updated.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(MXFILE) VALUE(20), V
      CHARACTER*10 KEY(20),K
      CHARACTER*82 LINE
      INTEGER FILE(3)
      LOGICAL DEF, DEFINP, TRUE, FALSE
C
C  The following statement functions will test if a value
C  of TRUE/FALSE is supplied as their argument.
      TRUE(V) = V.EQ.'T' .OR. V.EQ.'TRUE'
      FALSE(V)= V.EQ.'F' .OR. V.EQ.'FALSE'
C  ---------------------------------------------------------------------
C
C  Formats for the initial printout.
500   FORMAT ($(FC)79('-')
     &        /$(FC)15X,'The Merlin Control Language compiler v. 3.0'
     &        /$(FC)11X,'D.G. Papageorgiou, I.E. Lagaris, ',
     &        'I.N. Demetropoulos'
     &        /$(FC)21X,'University of Ioannina, GREECE'
     &        /$(FC)23X,'http://nrt.cs.uoi.gr/merlin'
     &        /$(FC)79('-'))
600   FORMAT ($(FC)'Enter compilation parameters   - I, B, E, DEBUG, ',
     &         'BOUNDS -' )
C
C  Output initial message.
      WRITE (*,500)
      WRITE (*,600)
C
C  This variable indicates whether the input file (the I parameter)
C  has been defined.
      DEFINP = .FALSE.
C
C  Get line containing compilation parameters.
      READ (*,'(A)',END=400) LINE
C
C  Pack the line just read, so as to eliminate any embedded blanks.
      LONG = LENGTH(LINE)
15    IL = INDEX(LINE,' ')
      IF (IL.LT.LONG .AND. IL.NE.0) THEN
         LINE(IL:) = LINE(IL+1:)
         LONG = LONG-1
         GOTO 15
      END IF
C
C  Add a comma and a blank to the end of the packed line.
C  Keyword searching will stop as soon as these two characters
C  are reached.
      LINE(81:82) = ', '
C
C  NK is the number of keywords that have been read.
C  L is the character position in LINE, where the next keyword
C  is expected to be found.
      NK = 0
      L = 1
C
C  IEQ is the character position in LINE, where the "=" for
C  the next keyword appears.
10    IEQ = INDEX(LINE(L:),'=')+L-1
C
C  If "=" was not found, or "=" was found just after the comma.
      IF (IEQ.EQ.L-1 .OR. IEQ.EQ.L) GOTO 100
C
C  Get the keyword.
      NK = NK+1
      KEY(NK) = LINE(L:IEQ-1)
C
C  IC is the index of the comma, following the current keyword.
      IC = INDEX(LINE(IEQ+1:),',')
      IC = IC+IEQ
C
C  If the comma follows the "=".
      IF (IEQ+1.EQ.IC) GOTO 100
C
C  Get the value assigned to this keyword.
C
      VALUE(NK) = LINE(IEQ+1:IC-1)
C  The character position, where the next keyword should
C  start, is the one, just after the comma.
      L = IC+1
C
C  If anything but blanks follow, try to get another keyword.
      IF (LINE(L:).NE.' ') GOTO 10
C
C  NF counts how many parameters that accept a file name as
C  a value, have been processed.
      NF = 0
C
C  The following loop will process all keywords and their values,
C  checking the validity of each entry.
      DO 20,M=1,NK
         CALL UPPER(KEY(M),LEN(KEY(M)))
         K = KEY(M)
         V = VALUE(M)
C  Check if V is empty.
         IF (V.EQ.' ') THEN
            WRITE (*,*) '-- Empty value for keyword: ',K
            STOP
         END IF
C  Check if the keyword has already been entered.
         DO 30,J=1,M-1
            IF (K.EQ.KEY(J)) THEN
               WRITE (*,*) '-- Keyword already declared: ', K
               STOP
            END IF
30       CONTINUE
C
C  Check if keyword belongs to the valid ones, and if so,
C  take appropriate action.
         IF (K.EQ.'I') THEN
C  Check whether the value of this keyword has been entered
C  as a file name in another keyword.
            IF (DEF(V,VALUE,FILE,NF)) GOTO 300
C  Indicate that this keyword, accepts a file name as a value.
            NF = NF+1
            FILE(NF) = M
C  Indicate that the input file (the I parameter) has been entered.
            DEFINP = .TRUE.
C  Assign the value to be returned by the routine.
            IFILE = V
         ELSE IF (K.EQ.'B') THEN
C  Check whether the value of this keyword has been entered
C  as a file name in another keyword.
            IF (DEF(V,VALUE,FILE,NF)) GOTO 300
C  Indicate that this keyword accepts a file name as a value.
            NF = NF+1
            FILE(NF) = M
C  Assign the value to be returned by the routine.
            BFILE = V
         ELSE IF (K.EQ.'E') THEN
C  Check whether the value of this keyword has been entered
C  as a file name in another keyword.
            IF (DEF(V,VALUE,FILE,NF)) GOTO 300
C  Indicate that this keyword accepts a file name as a value.
            NF = NF+1
            FILE(NF) = M
C  Assign the value to be returned by the routine.
            EFILE = V
         ELSE IF (K.EQ.'DEBUG') THEN
C  Check if a value that indicates debug selection has been entered.
            CALL UPPER(V,LEN(V))
            IF (TRUE(V)) THEN
               DEBUG = .TRUE.
            ELSE IF (FALSE(V)) THEN
               DEBUG = .FALSE.
            ELSE
               WRITE (*,*)
     &                   '-- Incorrect value for the -DEBUG- parameter.'
               STOP
            END IF
         ELSE IF (K.EQ.'BOUNDS') THEN
C  Check if a value that indicates bounds selection has been entered.
            CALL UPPER(V,LEN(V))
            IF (TRUE(V)) THEN
               BOUNDS = .TRUE.
            ELSE IF (FALSE(V)) THEN
               BOUNDS = .FALSE.
            ELSE
               WRITE (*,*)
     &                 '-- Incorrect value for the -BOUNDS- parameter.'
               STOP
            END IF
         ELSE
C  This is not a correct keyword.
            WRITE (*,*) '-- Incorrect keyword: ', K
            STOP
         END IF
20    CONTINUE
C
C  If an input file has been defined, return normally.
400   IF (DEFINP) RETURN
      WRITE (*,*) '-- Input file not specified.'
      STOP
C
C  If any error occurred in keyword/value separation.
100   WRITE (*,*) '-- Incorrect entry.'
      STOP
C
C  If the same file is specified for different keywords.
300   WRITE (*,*) '-- File already specified: ', V(1:LENGTH(V))
      STOP
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE GKS ( ID, NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine GKS parses a number of statements of the form:
C      COMMAND ( KEY = VALUE )
C    VALUE is a string enclosed in single quotes.
C    Multiple entries of KEY=VALUE pairs are allowed. Each one
C    of the statements being parsed has a unique set of allowed
C    keywords. Some statements may be entered without any keywords,
C    while for others, a keyword is required.
C
C  Input arguments:
C    ID         Integer that identifies the statement being parsed:
C               ID =  1 -> DELETE
C               ID =  2 -> DISCARD
C               ID =  3 -> REWIND
C               ID =  4 -> GOEOF
C               ID =  5 -> HIDEOUT
C               ID =  6 -> STOP
C               ID =  7 -> RETURN
C               ID =  8 -> ALIAS
C               ID =  9 -> UNALIAS
C               ID = 10 -> EPILOG
C               ID = 11 -> TITLE
C               ID = 12 -> PDUMP
C               ID = 13 -> PANELON
C               ID = 14 -> PANELOFF
C               ID = 15 -> PSTATUS
C               ID = 16 -> GRADCHECK
C               ID = 17 -> CLOSE
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME
      CHARACTER*80 LINE
      PARAMETER ( MXCO = 17, MXK = 2 )
      CHARACTER*10 COM(MXCO), KEY(MXK,MXCO)
      INTEGER NOARG(MXCO), IACTS(MXK), IACTE(MXK), IKEY(MXK)
C
C  COM = command to be written to the MOC file.
C  NOARG = Number of arguments that must be present.
      DATA (COM(I), NOARG(I), (KEY(J,I),J=1,MXK),I=1,10) /
     &      'DELETE    ', 1, 'FILE      ', '          ',
     &      'DISCARD   ', 1, 'FILE      ', 'TYPE      ',
     &      'REWIND    ', 1, 'FILE      ', '          ',
     &      'GOEOF     ', 1, 'FILE      ', '          ',
     &      'HIDEOUT   ', 1, 'FILE      ', 'APPEND    ',
     &      'STOP      ', 0, 'EPILOG    ', '          ',
     &      'RETURN    ', 0, 'EPILOG    ', '          ',
     &      'ALIAS     ', 2, 'NAME      ', 'COMMAND   ',
     &      'UNALIAS   ', 1, 'NAME      ', '          ',
     &      'EPILOG    ', 1, 'COMMAND   ', '          ' /
      DATA (COM(I), NOARG(I), (KEY(J,I),J=1,MXK),I=11,MXCO) /
     &      'TITLE     ', 1, 'TITLE     ', '          ',
     &      'PDUMP     ', 1, 'FILE      ', 'COMMAND   ',
     &      'PANELON   ', 0, 'COMMAND   ', '          ',
     &      'PANELOFF  ', 0, 'COMMAND   ', '          ',
     &      'PSTATUS   ', 0, 'COMMAND   ', '          ',
     &      'GRADCHECK ', 1, 'MODE      ', 'MODE2     ',
     &      'CLOSE     ', 1, 'FILE      ', '          '/
C  ---------------------------------------------------------------------
C
      NT = NT+1
      IF (NT.EQ.NTOK) THEN
         IF (NOARG(ID).EQ.0) THEN
C  Permitted without arguments. Write the command.
            BLOC = BLOC+1
            IF (GENE) WRITE (SCR,'(A)') COM(ID)
         ELSE
            CALL ERROR(31,COM(ID))
         END IF
         RETURN
      END IF
C
C  Expect a left parenthesis.
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,COM(ID))
         RETURN
      END IF
C
      DO 20,I=1,MXK
         IKEY(I) = 0
         IACTS(I) = 0
         IACTE(I) = 0
20    CONTINUE
      NC = 0
C
C  Loop over all keyword-value pairs.
100   CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         RETURN
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,NAME)
      DO 10,IK=1,MXK
         IF (ISCOMP(NAME,KEY(IK,ID)).NE.0) GOTO 11
10    CONTINUE
C
C  Keyword not allowed for this statement.
      CALL ERROR(33,NAME)
      RETURN
C
C  Duplicate entry of the keyword ?
11    CONTINUE
      IF (IKEY(IK).NE.0) THEN
         CALL ERROR(166,NAME)
         RETURN
      END IF
      IKEY(IK) = 1
C
C  Expect the = sign.
      NT=NT+1
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(35,' ')
         RETURN
      END IF
C
C  Get the value.
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL ERROR(169,NAME)
         RETURN
      END IF
C
      NT = NT+1
      IF (TOKEN(NT).NE.STRING) THEN
         CALL ERROR(28,' ')
         RETURN
      END IF
      CALL GETIDS(NA,NN,IS,IE)
      IACTS(IK) = IS
      IACTE(IK) = IE
C
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL ERROR(28,NAME)
         RETURN
      END IF
C
      NC = NC+1
      NT = NT+1
      CUR = TOKEN(NT)
C
C  Another keyword ?
      IF (CUR.EQ.SEMI) GOTO 100
      IF (CUR.NE.RIGHTP) THEN
         CALL ERROR(36,' ')
         RETURN
      END IF
C
C Check for keywords that must be present.
      IERR = 0
      DO 40,I=1,NOARG(ID)
         IF (IKEY(I).EQ.0) THEN
C  You must supply this keyword.
            CALL ERROR(41,KEY(I,ID))
            IERR = 1
         END IF
40    CONTINUE
      IF (IERR.NE.0) RETURN
C
C  Produce object code.
      LEC = LENGTH(COM(ID))
      LINE(1:LEC) = COM(ID)(1:LEC)
      NL = LEC+1
      LINE(NL:NL) = ' '
      IF (ID.EQ.17) THEN
C  Close.
         BLOC = BLOC+2
         IF (GENE) WRITE (SCR,'(A/A)')
     &                '$CLOSE',ALPHA(IACTS(1):IACTE(1))
      ELSE
         DO 30,I=1,MXK
            IS = IACTS(I)
            IE = IACTE(I)
            IF (IS.NE.0) THEN
C
C  Command/keyword specific part.
               IF (ID.EQ.2 .AND. I.EQ.2) THEN
C  Discard.
                  IF (ISCOMP(ALPHA(IS:IE),'TEXT').EQ.0 .AND.
     &                            ISCOMP(ALPHA(IS:IE),'BIN').EQ.0) THEN
                     CALL ERROR(185,KEY(I,ID))
                     RETURN
                  END IF
               END IF
               IF (ID.EQ.5 .AND. I.EQ.2) THEN
C  Hideout.
                  IF (ISCOMP(ALPHA(IS:IE),'YES').NE.0) THEN
                     LINE(NL+1:NL+6) = 'APPEND'
                     NL = NL+6
                  ELSE IF (ISCOMP(ALPHA(IS:IE),'NO').EQ.0) THEN
                     CALL ERROR(185,KEY(I,ID))
                     RETURN
                  END IF
               ELSE IF ((ID.EQ.6 .OR. ID.EQ.7) .AND. I.EQ.1) THEN
C  Stop, return.
                  IF (ISCOMP(ALPHA(IS:IE),'NO').NE.0) THEN
                     LINE(NL+1:NL+8) = 'NOEPILOG'
                     NL = NL+8
                  ELSE IF (ISCOMP(ALPHA(IS:IE),'YES').EQ.0) THEN
                     CALL ERROR(185,KEY(I,ID))
                     RETURN
                  END IF
               ELSE
C  Delimit/escape special characters.
                  CALL SDELIM(ALPHA(IS:IE),IE-IS+1,LINE(NL+1:),LL)
                  NL = NL+LL+1
                  LINE(NL:NL) = ' '
               END IF
            END IF
30       CONTINUE
         BLOC = BLOC+1
         IF (GENE) WRITE (SCR,'(A)') LINE(1:NL)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE HEAD
C  ---------------------------------------------------------------------
C
C  Description:
C    Issues the standard header (program version and compilation
C    options).
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL FIRST
      SAVE FIRST
      DATA FIRST / .TRUE. /
C  ---------------------------------------------------------------------
C
      IF (FIRST) THEN
         IF (EOPEN) THEN
            WRITE (ERL,20)
     &            IFILE(1:LENGTH(IFILE)), EFILE(1:LENGTH(EFILE)),
     &            BFILE(1:LENGTH(BFILE)), DEBUG, BOUNDS
         ELSE
            WRITE (*,20)
     &            IFILE(1:LENGTH(IFILE)), '*',
     &            BFILE(1:LENGTH(BFILE)), DEBUG, BOUNDS
         END IF
C  Subsequent calls to this routine shouldn't issue the above message.
         FIRST = .FALSE.
      END IF
C
20    FORMAT (/$(FC)'M C L - 3.0'/
     &         $(FC)'Compilation parameters:'/
     &         $(FC)'- I=',A,', E=',A,', B=',A,
     &         ', DEBUG=',L1,', BOUNDS=',L1,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IF
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes the IF statement. The syntax of this statement is:
C    IF <MCL expression> THEN
C
C    The object code generated is:
C
C    Code to evaluate the <MCL expression>
C    $NOT
C    $TESTMOVE
C    <blank line>
C
C    If the statement is correct, a value of 1 is pushed in the type
C    table, while the location of the above <blank line> is pushed in
C    the type-definition table.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
C
      LOGICAL CORRE
      INTEGER EX(LINLEN)
      CHARACTER*(IDLEN) STARG(1), IDF
C  ---------------------------------------------------------------------
C
C  Initialize pointers to the TOKEN, ALPHA and NUM arrays respectively.
      NT = 1
      NA = 1
      NN = 2
C
C  Extract and evaluate the <MCL expression>.
      CALL EXTRA(NT,EX,NEX,1,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If expression incorrect.
      IF (.NOT.CORRE) RETURN
C
C  Write code to skip the IF-part of the block if, in case the
C  expression yields a zero value. (Which means that the given
C  condition is not satisfied)
      BLOC = BLOC+3
      IF (GENE) WRITE (SCR,'(A/A/)') '$NOT','$TESTMOVE'
C
C  THEN is expected after the <MCL expression> in the IF statement.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(8,' ')
         RETURN
      ELSE
         CALL GETIDF(NA,NN,IDF)
         IF (IDF.NE.'THEN') THEN
            CALL ERROR(22,IDF)
            RETURN
         END IF
      END IF
C
C  EOL should follow.
      NT=NT+1
      IF (NT.NE.NTOK) THEN
         CALL ERROR(9,'IF')
         RETURN
      END IF
C
C  The IF statement, is found to be correct. Update the type
C  and type-definition tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP) CALL OVER('type','TYPE','IF',MXTYP,'MXTYP')
      TYPE(NTYPE) = 1
      TDEF(NTYPE) = BLOC
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INIT
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine INIT performs several initializations, before the
C    Merlin configuration file is read.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 99, MXPLUG = $(MXPLUG), NES = NMCOM+MXPLUG )
      CHARACTER*14 STAT(NES)
      COMMON /COM1/ STAT
      INTEGER NAME(NES), INDE(NES)
      COMMON /COM2/ NAME, INDE
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      DO 10,I=1,MXPLUG
         CALL I2STR(I,STR,LES)
         IND = NMCOM+I
         STAT(IND) = 'PLUG'//STR(1:LES)
         NAME(IND) = 0
10    CONTINUE
C
C  The tab and escape (backslash) characters are specific for the
C  ASCII character set.
      TAB = CHAR(9)
      ESC = CHAR(92)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ZINIT
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine ZINIT performs initializations, after the Merlin
C    configuration file is read.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( NMCOM = 99, MXPLUG = $(MXPLUG), NES = NMCOM+MXPLUG )
      CHARACTER*14 STAT(NES)
      COMMON /COM1/ STAT
      INTEGER NAME(NES), INDE(NES)
      COMMON /COM2/ NAME, INDE
C  ---------------------------------------------------------------------
C
C  Sort array STAT.
      DO 20,I=1,NES
20       INDE(I) = I
      CALL CMSORT(STAT,NAME,INDE,NES)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE INSERT ( ID, ND, ML, IP, KI, IND )
C  ---------------------------------------------------------------------
C
C  Description:
C    Inserts information in the tables related to variables, arrays and
C    statement functions. The tables affected are:
C
C      1) The var table,
C      2) the dimensionality table,
C      3) the memory location table,
C      4) the position table and
C      5) the kind table.
C
C  The position where information should be inserted is
C  input to the routine.
C
C  Input arguments:
C    ID         Character variable. This value will be inserted in
C               the var table.
C    ND         Integer variable. This value will be inserted in the
C               dimensionality table.
C    ML         Integer variable. This value will be inserted in the
C               memory location table.
C    IP         Integer variable. This value will be inserted in the
C               position table.
C    KI         Integer variable. This value will be inserted in the
C               kind table.
C    IND        Integer variable. This is the position where the input
C               values should be inserted in the tables stated above.
C
C  Abort conditions that may occur:
C    In case the tables stated, overflow, the program will stop and
C    issue an informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) ID
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C
      NVAR = NVAR+1
      IF (NVAR.GT.MXVAR) CALL OVER('var','VAR','INSERT',MXVAR,'MXVAR')
C
C  Move up all table entries by one position.
      DO 80,K=NVAR,IND+1,-1
         KM1 = K-1
         VAR(K)  = VAR(KM1)
         DIM(K)  = DIM(KM1)
         LOC(K)  = LOC(KM1)
         IPOS(K) = IPOS(KM1)
         KIND(K) = KIND(KM1)
80    CONTINUE
C
C  Insert the input values at the specified position.
      VAR(IND) = ID
      DIM(IND) = ND
      LOC(IND) = ML
      IPOS(IND)= IP
      KIND(IND)= KI
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WPSC ( NU, NCOUNT, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Writes the $PUSHC instruction with an integer argument.
C
C  Input arguments:
C    NU         Unit to write to.
C    N          Argument to the $PUSHC instruction.
C
C  Input / Output arguments:
C    NCOUNT     Line counter for unit NU.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      NCOUNT = NCOUNT+2
      IF (GENE) THEN
         CALL I2STR(N,STR,LES)
         WRITE (NU,10) STR(1:LES)
      END IF
C
10    FORMAT ('$PUSHC'/A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WPS ( NU, NCOUNT, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Writes the $PUSH instruction.
C
C  Input arguments:
C    NU         Unit to write to.
C    N          Argument to the $PUSH instruction.
C
C  Input / Output arguments:
C    NCOUNT     Line counter for unit NU.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      NCOUNT = NCOUNT+2
      IF (GENE) THEN
         CALL I2STR(N,STR,LES)
         WRITE (NU,10) STR(1:LES)
      END IF
C
10    FORMAT ('$PUSH'/A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WPOP ( NU, NCOUNT, N )
C  ---------------------------------------------------------------------
C
C  Description:
C    Writes the $POP instruction.
C
C  Input arguments:
C    NU         Unit to write to.
C    N          Argument to the $POP instruction.
C
C  Input / Output arguments:
C    NCOUNT     Line counter for unit NU.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      NCOUNT = NCOUNT+2
      IF (GENE) THEN
         CALL I2STR(N,STR,LES)
         WRITE (NU,10) STR(1:LES)
      END IF
C
10    FORMAT ('$POP'/A)
      END
C  ---------------------------------------------------------------------
      FUNCTION ISLCI ( IND )
C  ---------------------------------------------------------------------
C
C  Description:
C    Function ISLCI returns 1 if the input variable (as indicated by
C    its position (IND) in the var table) is a loop control variable.
C    Otherwise it returns 0.
C
C  Input arguments:
C    IND        Index of the variable, in the var table.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C
C  Only simple variables (local or arguments) can be used as loop
C  control variables.
      K = KIND(IND)
      IF (K.NE.1 .AND. K.NE.101) THEN
         ISLCI = 0
         RETURN
      END IF
C
C  For a local variable, use its address. For a subprogram argument,
C  its (negative) serial number in the argument list.
      L = LOC(IND)
      IF (K.EQ.101) L = -L
C
C  Search the loop table.
      DO 10,I=1,NLOO
         IF (LLOC(I).EQ.L) THEN
            ISLCI = 1
            RETURN
         END IF
10    CONTINUE
      ISLCI = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE JUMPIN ( F1, F2, T1, T2 )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine checks whether any jumps are performed from a block
C    of statements (specified by two pointers in the type table) to
C    another block of statements (specified also by two pointers in
C    the type table). Upon exit, all MOVE TO statements that have been
C    processed will assume a value of 7.
C
C  Input arguments:
C    F1         Integer variable. This is a pointer to the type table,
C               and is the lower limit from where the search for
C               any MOVE TO statements will start.
C    F2         Integer variable. This is a pointer to the type table,
C               and is the upper limit to where the search for
C               any MOVE TO statements will terminate.
C    T1         Integer variable. This is a pointer to the type table,
C               and is the lower limit of the block that shouldn't
C               be the target of a MOVE TO statement residing
C               in the first block.
C    T2         Integer variable. This is a pointer to the type table,
C               and is the upper limit of the block that shouldn't
C               be the target of a MOVE TO statement residing
C               in the first block.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  ---------------------------------------------------------------------
C
C  Calculate definition locations for the second block.
      LOC1 = TDEF(T1)
      LOC2 = TDEF(T2)
C
C  Start searching the type table.
      DO 10,I=F1,F2
         IF (TYPE(I).EQ.-7) THEN
C  A MOVE TO statement has been found.
C  Since the type-link table holds only relative numbers, calculate
C  the absolute location of the target label.
            IR = TREL(I)
            IF (IR.LT.0) IR = IR-1
            JT = TDEF(I)+IR
C  If the target location lies inside the second block,
C  this is an error.
            IF (JT.GE.LOC1 .AND. JT.LE.LOC2) THEN
               CALL ERROR(148,' ')
               TYPE(I) = 7
               TREL(I) = -TDEF(I)+1
            END IF
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LBDEF ( NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes a label definition. The syntax of a label definition is:
C    LABEL :
C    If the label has already been defined an error occurs. Only one
C    line of object code is generated:
C
C    $CONTINUE
C
C    The label is assigned the location of the above line in the scratch
C    file (unit SCR).
C
C  Input arguments:
C    NAME       Name of the label.
C
C  Abort conditions that may occur:
C    If the label table overflows, the program will stop and issue an
C    informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL FOUND
C  ---------------------------------------------------------------------
C
C  Locate label in the associated table.
      CALL LOCAN(LABEL,NLAB,NAME,FOUND,IND)
      IF (FOUND) THEN
C  Label already in table. Issue error message.
         CALL ERROR(1,NAME)
      ELSE
C  Label not in table, so insert it and assign to it
C  the appropriate location.
         NLAB = NLAB+1
         IF (NLAB.GT.MXLAB)
     &          CALL OVER('label','LABEL','LBDEF',MXLAB,'MXLAB')
         DO 20,I=NLAB,IND+1,-1
            LABEL(I) = LABEL(I-1)
            LABLOC(I) = LABLOC(I-1)
20       CONTINUE
         BLOC = BLOC+1
C  Put the label in the label table.
         LABEL(IND) = NAME
C  Assign the scratch file location to the label.
         LABLOC(IND) = BLOC
         IF (GENE) WRITE (SCR,'(A)') '$CONTINUE'
      END IF
C
C  If more than 3 tokens (LABEL COLON EOL) are in the same line,
C  this is an error.
      IF (NTOK.NE.3) CALL ERROR(2,NAME)
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOCAN ( TABLE, ELEM, NAME, FOUND, POSIT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Locates a name in a character type array, sorted in ascending
C    order. The method used, is a binary search.
C
C  Input arguments:
C    TABLE   The character array that will be searched. It should
C            be sorted in ascending order. It is assumed to be
C            dimensioned as: TABLE(1:ELEM)
C    ELEM    Number of elements in the above array.
C    NAME    The name whose position in the array we are looking for.
C
C  Output arguments:
C    FOUND   Logical variable indicating, whether the requested name
C            was (FOUND=.TRUE.), or was not found (FOUND=.FALSE.) in
C            the given array.
C    POSIT   If the name was found in the array, this is its position.
C            Otherwise, this is the position that the given name
C            should occupy.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) TABLE(*), NAME
      LOGICAL FOUND
      INTEGER ELEM, POSIT
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER FIRST
C  ---------------------------------------------------------------------
C
      FIRST = 1
      LAST = ELEM
C
10    CONTINUE
      IF (FIRST.GT.LAST) GOTO 20
      MIDDLE = (FIRST+LAST)/2
      IF (NAME.GT.TABLE(MIDDLE)) THEN
         FIRST = MIDDLE+1
      ELSE
         LAST = MIDDLE-1
      END IF
      GOTO 10
C
20    POSIT = FIRST
      IF (POSIT.GT.ELEM) THEN
         FOUND = .FALSE.
      ELSE
         FOUND = NAME.EQ.TABLE(POSIT)
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOCIF ( IF, ELSE, END )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will locate an IF and an ELSE in the type table,
C    given the position of the ENDIF.
C    It is used while checking the validity of the MOVE TO statements
C    and assumes that entries in the type table that have not been
C    processed are negative; positive otherwise.
C
C  Input arguments:
C    END        Integer variable. This is the position of the ENDIF
C               in the type table.
C
C  Output arguments:
C    IF         Integer variable. This is the position of the IF in
C               the type table, corresponding to the given ENDIF.
C    ELSE       Integer variable. This is the position of the ELSE in
C               the type table, corresponding to the given ENDIF.
C               A value of 0 is returned if this is an IF
C               without ELSE.
C
C  Abort conditions that may occur:
C    If an IF can't be located in the type table, the program will
C    stop and issue an informative message. Normally this should never
C    happen.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C
      ELSE = 0
C
C  Search the type array backwards, starting at the ENDIF position.
      DO 10,I=END,TSTART,-1
         IF (TYPE(I).EQ.-1) THEN
C  An IF has been found.
            IF = I
            RETURN
         ELSE IF (TYPE(I).EQ.-2) THEN
C  An ELSE has been found. Don't return, but locate the
C  corresponding IF as well.
            ELSE = I
         END IF
10    CONTINUE
C
C  An IF could not be located in the type table.
C  This program is malfunctioning.
      CALL MABORT('Cannot locate "IF" in the type table.','LOCIF')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOCLOO ( LOO, ENDLOO )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will locate a LOOP in the type table, given the
C    position of the ENDLOOP.
C    It is used while checking the validity of the MOVE TO statements
C    and assumes that entries in the type table that have not been
C    processed are negative; positive otherwise.
C
C  Input arguments:
C    ENDLOO     Integer variable. This is the position of the ENDLOOP
C               in the type table.
C
C  Output arguments:
C    LOO        Integer variable. This is the position of the LOOP
C               in the type table, corresponding to the given ENDLOOP.
C
C  Abort conditions that may occur:
C    If a LOOP can't be located in the type table, the program will
C    stop and issue an informative message. Normally this should never
C    happen.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C
C  Start searching the type table.
      DO 10,LOO=ENDLOO,TSTART,-1
         IF (TYPE(LOO).EQ.-4) RETURN
10    CONTINUE
C
C  A LOOP could not be located in the type table.
C  This program is malfunctioning.
      CALL MABORT('Cannot locate "LOOP" in the type table.','LOCLOO')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LOOP
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the loop statement an generates appropriate object code.
C    The statement syntax is:
C
C    LOOP <var> FROM <init> TO <final> by <step>         or
C    LOOP <var> FROM <init> TO <final>
C
C    where <var> is a simple variable (referred to as the loop control
C    variable) and
C    <init>, <final> and <step> are expressions specifying the initial,
C    final and step values for the loop.
C    For each loop 5 consecutive storage locations are needed. Their
C    contents are listed along with the symbolic name, they will be
C    be referred to later in this description:
C
C    1) Storage for the value of <init>. ( init_mloc )
C    2) Storage for the value of <final>. ( final_mloc)
C    3) Storage for the value of <step>. ( step_mloc)
C    4) Storage for the current trip count of the loop. ( ctc_mloc )
C    5) Storage for the final trip count of the loop. ( ftc_mloc )
C
C    Since these memory locations are used only while the loop is
C    executing, the storage required for all loops in the program is:
C    5*MAXL, where MAXL is the maximum number of nested loops
C    encountered in the source. Due however to the memory allocation
C    strategy, each loop is assigned its own 5 memory locations.
C
C    Initially the <init>, <final> and <step> expressions are
C    calculated:
C
C    <code to evaluate init>
C    $POP
C    init_mloc
C    <code to evaluate final>
C    $POP
C    final_mloc
C    <code to evaluate step>
C    $POP
C    step_mloc
C
C    Then the final trip count is calculated by the formula:
C    ftc = max( int ( (<final>-<init>+<step>)/<step> ) , 0 )
C
C    $PUSH
C    final_mloc
C    $PUSH
C    init_mloc
C    $-
C    $PUSH
C    step_mloc
C    $+
C    $PUSH
C    step_mloc
C    $/
C    $TRUNC
C    $PUSHC
C    0
C    $MAX
C    2
C    $POP
C    ftc_mloc
C
C    <var> is then initialized to the value of <init>:
C
C    $PUSH
C    init_mloc
C    $POP            or, for a subprogram argument:   $IPOP
C    var_mloc                                         N
C
C    where var_mloc is the memory location assigned to <var> and,
C    in case <var> is an argument to the current subprogram, N is
C    its serial number in the argument list.
C    The trip counter is then initialized to 1:
C
C    $PUSHC
C    1
C    $POP
C    ctc_mloc
C
C    A check will be made if the trip counter has reached the final
C    trip count:
C
C    $PUSH
C    ctc_mloc
C    $PUSH
C    ftc_mloc
C    $>
C    $TESTMOVE
C    <blank line>
C
C    The target of the $TESTMOVE instruction is the ENDLOOP instruction,
C    to be found later on.
C    A value of 4 is pushed in the type table, indicating that a loop
C    statement has been reached. The location of the above <blank line>
C    is also pushed in the type-definition table.
C
C  Abort conditions that may occur:
C    If either the type, or the loop table overflows, the program
C    will stop and issue an informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER EX(LINLEN)
      LOGICAL FOUND,CORRE
      CHARACTER*(IDLEN) NAME, STARG(1), IDF
C  ---------------------------------------------------------------------
C
C  Initialize pointers to the TOKEN, ALPHA and NUM arrays respectively.
      NT = 1
      NA = 1
      NN = 2
C
C  An identifier is expected as the loop control variable.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(133,' ')
         RETURN
      END IF
C
C  NLT is the number of nested loops currently in effect.
      NLT = NLOO+1
C
C  If the loop table overflows.
      IF (NLT.GT.MXLOO)
     &    CALL OVER('loop','LLOC','LOOP',MXLOO,'MXLOO')
C
C  Get the name of the loop control variable.
      CALL GETIDF(NA,NN,NAME)
C
C  Check if in the var table.
      CALL LOCAN(VAR,NVAR,NAME,FOUND,IND)
C
      IF (FOUND) THEN
C
C  Its kind should be 1 (simple variable), -1 (undeclared variable),
C  or 101 (simple variable-argument).
         K = KIND(IND)
         IF (K.NE.1 .AND. K.NE.-1 .AND. K.NE.101) THEN
            CALL ERROR(133,' ')
C  L is the memory location of the loop control variable. Since the
C  kind is incorrect, assign a false location to it, different
C  however from all loops nested outside or possibly inside this one.
            L = MEMLOC+NLT
         ELSE
C  L will be set to a negative number if the loop control variable
C  is an argument to the current subprogram. In fact, if it is the
C  N-th argument, L is set to -N.
            L = LOC(IND)
            IF (K.EQ.101) L = -L
         END IF
      ELSE
C
C  The loop control variable was not found in the var table.
         CALL ERROR(19,NAME)
         L = MEMLOC+NLT
C  Insert it in the table as a simple variable. (KIND=1, not -1)
         CALL INSERT(NAME,0,L,0,1,IND)
      END IF
C
C  Put the location of the loop control index in the loop table.
      LLOC(NLT) = L
C
C  Find out whether this loop control index has been used as control
C  index in a loop nested outside this one.
      DO 99,I=1,NLT-1
         IF (LLOC(I).EQ.L) THEN
C  This loop control variable has been used as control variable in
C  a loop nested outside this one.
            CALL ERROR(135,NAME)
            GOTO 40
         END IF
99    CONTINUE
C
40    CONTINUE
C
C  LMEM in the memory location where the 5 storage positions for
C  this loop start.
C  Assign the appropriate location to LMEM. Also determine locations
C  for the necessary loop information.
      LM = MEMLOC+1
      LMEM(NLT) = LM
      MEMLOC = MEMLOC+5
      INIT  = LM
      IFIN  = LM+1
      ISTEP = LM+2
      ITC   = LM+3
      ITCF  = LM+4
C
C  An identifier (FROM) is expected after the loop control variable.
      NT=NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(136,' ')
         RETURN
      END IF
C  FROM is expected after the loop control index.
      CALL GETIDF(NA,NN,IDF)
      IF (IDF.NE.'FROM') THEN
         CALL ERROR(13,IDF)
         RETURN
      END IF
C
C  Calculate initial loop value.
      CALL EXTRA(NT,EX,NEX,1,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
      IF (.NOT.CORRE) RETURN
C
C  Pop the value from the stack in the appropriate memory location.
      CALL WPOP(SCR,BLOC,INIT)
C
C  An identifier (TO) is expected.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(137,' ')
         RETURN
      END IF
C
C  TO is expected after the expression for the initial loop value.
      CALL GETIDF(NA,NN,IDF)
      IF (IDF.NE.'TO') THEN
         CALL ERROR(14,IDF)
         RETURN
      END IF
C
C  Calculate the final loop value.
      CALL EXTRA(NT,EX,NEX,1,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
      IF (.NOT.CORRE) RETURN
C
C  Pop the final loop value to the appropriate memory location.
      CALL WPOP(SCR,BLOC,IFIN)
C
C  Check if a step is defined.
      NT = NT+1
      IF (NT.NE.NTOK) THEN
C  The statement has not been completed yet.
C  So a step value may be defined.
         IF (TOKEN(NT).NE.IDENT) THEN
C  An identifier (BY) does not follow.
            CALL ERROR(138,' ')
            RETURN
         END IF
         CALL GETIDF(NA,NN,IDF)
         IF (IDF.NE.'BY') THEN
C  An identifier follows but is not BY.
            CALL ERROR(23,IDF)
            RETURN
         END IF
C
C  Everything in order. Now calculate the step value.
         CALL EXTRA(NT,EX,NEX,1,NA)
         CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
         IF (.NOT.CORRE) RETURN
C  Pop the step value to the appropriate memory location.
         CALL WPOP(SCR,BLOC,ISTEP)
      ELSE
C
C  A step is not defined. Use a default of 1.
         CALL WPSC(SCR,BLOC,1)
         CALL WPOP(SCR,BLOC,ISTEP)
      END IF
C
C  Loop statement correct. Proceed with code generation.
      BLOC = BLOC+9
      IF (GENE) THEN
C  Calculate final trip count.
         CALL WPS(SCR,BLOC,IFIN)
         CALL WPS(SCR,BLOC,INIT)
         WRITE (SCR,'(A)') '$-'
         CALL WPS(SCR,BLOC,ISTEP)
         WRITE (SCR,'(A)') '$+'
         CALL WPS(SCR,BLOC,ISTEP)
         WRITE (SCR,'(A/A)')'$/','$TRUNC'
         CALL WPSC(SCR,BLOC,0)
         WRITE (SCR,'(A/I2)') '$MAX',2
         CALL WPOP(SCR,BLOC,ITCF)
C  Initialize the loop control variable.
         CALL WPS(SCR,BLOC,INIT)
         IF (K.EQ.1) THEN
            CALL WPOP(SCR,BLOC,ABS(L))
         ELSE
            WRITE (SCR,'(A)') '$IPOP'
            WRITE (SCR,*) ABS(L)
            BLOC = BLOC+2
         END IF
C  Initialize the trip counter.
         CALL WPSC(SCR,BLOC,1)
         CALL WPOP(SCR,BLOC,ITC)
C  Check if the trip counter has reached its target value.
         CALL WPS(SCR,BLOC,ITC)
         CALL WPS(SCR,BLOC,ITCF)
         WRITE (SCR,'(A/A/)') '$>','$TESTMOVE'
      END IF
C
      NLOO = NLT
C  Push a value of 4 in the type table to indicate the loop statement.
C  Also push the location of the blank line just written to the scratch
C  file (following the $TESTMOVE instruction) in the
C  type-definition table.
C  TREL keeps temporarily the variable type (1 or 101), until the
C  corresponding ENDLOOP is reached.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','LOOP',MXTYP,'MXTYP')
      TYPE(NTYPE) = 4
      TDEF(NTYPE) = BLOC
      TREL(NTYPE) = K
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MABORT ( MESS, SUB )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will be called to issue informative messages and stop
C    the program when a situation that normally should never happen,
C    is encountered.
C
C  Input arguments:
C    MESS       Character variable. This is an informative message to
C               be displayed.
C    SUB        Character variable. This is the routine name in which
C               the abnormal situation happened.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) MESS, SUB
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
      WRITE (*,10) SUB
      WRITE (*,20) MESS
C
C  Write the same message on the error list file.
      IF (EOPEN) THEN
         WRITE (ERL,10) SUB
         WRITE (ERL,20) MESS
      END IF
      STOP 'ABRT'
C
10    FORMAT (/$(FC)'**** Routine ',A,' aborted.')
20    FORMAT ($(FC)'**** ',A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MAINPR
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the definition of a main program and generates appropriate
C    object code:
C      $PUSHC           --->  Allocate memory
C      <blank line>
C      $ALLOC
C      $POPMB           --->  Set current memory base.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL FOUND
C  ---------------------------------------------------------------------
C
C  Nothing else should follow the PROGRAM statement.
      IF (NTOK.NE.2) THEN
         CALL ERROR(9,'PROGRAM')
         RETURN
      END IF
C
C  Is it the first occurrence of  a main program ?
      CALL SERLOC(EXT,NEXT,'..',FOUND,POSIT)
      IF (FOUND) THEN
         CALL ERROR(27,' ')
         RETURN
      END IF
C
C  Update the externals table. A main program is designated by "..".
      NEXT = NEXT+1
      IF (NEXT.GT.MXEXT)
     &    CALL OVER('externals','EXT','MAINPR',MXEXT,'MXEXT')
      EXT(NEXT) = '..'
      EXTDEF(NEXT) = BLOC+1
      NARDEF(NEXT) = 0
      BLOC = BLOC+2
      IF (GENE) THEN
         WRITE (SCR,'(A)') '$PUSHC'
         WRITE (SCR,*)
      END IF
C
C  Update the type and type-definition tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','MAINPR',MXTYP,'MXTYP')
      TYPE(NTYPE) = 9
      TDEF(NTYPE) = BLOC
      BLOC = BLOC+2
      IF (GENE) THEN
         WRITE (SCR,'(A)') '$ALLOC'
         WRITE (SCR,'(A)') '$POPMB'
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCALC
C  ---------------------------------------------------------------------
C
C  Description:
C    This subroutine is called when all statements in an MCL subprogram
C    have been processed. It searches the type table for entries of
C    type 9 (memory requirement) and fills in the corresponding
C    type-link table entry.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      IMPLICIT INTEGER (A-Z)
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL FOUND
C  ---------------------------------------------------------------------
C
      IF (.NOT.GENE) RETURN
C
      FOUND = .FALSE.
      DO 10,I=TSTART,NTYPE
         IF (TYPE(I).EQ.9) THEN
            TYPE(I) = -9
            TREL(I) = MEMLOC
            FOUND = .TRUE.
         END IF
10    CONTINUE
      IF (.NOT.FOUND) CALL MABORT('Type 9 not found.','MCALC')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCL
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine MCL controls the three stages of compilation. It
C    repeatedly calls subroutine ONESUB (pass 1), then links all
C    jumps due to MOVE TO, LOOP etc. (pass 2), and finally links
C    all subprograms together (pass 3), adding code for the predefined
C    ones.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
C
C  Total number of intrinsic functions and Merlin variables and arrays.
      PARAMETER ( NPRE = 56 )
      CHARACTER*(IDLEN) PRVAR(NPRE)
      INTEGER PRDIM(NPRE), PRKIND(NPRE)
C
C  Number of predefined subprograms.
      PARAMETER ( MXPSU = 3 )
      CHARACTER*(IDLEN) PSU(MXPSU)
      INTEGER NARPSU(MXPSU)
      LOGICAL FIRST, FOUND
      CHARACTER TMP
      CHARACTER*10 STR
      SAVE FIRST
      DATA FIRST / .TRUE. /
C
      DATA (PRVAR(I),PRDIM(I),PRKIND(I),I=1,10) /
     &     'ABS                           ', 1, 6,
     &     'ACOS                          ', 1, 6,
     &     'ACOSH                         ', 1, 6,
     &     'ASIN                          ', 1, 6,
     &     'ASINH                         ', 1, 6,
     &     'ATAN                          ', 1, 6,
     &     'ATANH                         ', 1, 6,
     &     'BACKUP                        ', 0, 4,
     &     'X                             ', 1, 5,
     &     'COS                           ', 1, 6 /
      DATA (PRVAR(I),PRDIM(I),PRKIND(I),I=11,20) /
     &     'COSH                          ', 1, 6,
     &     'DERIVA                        ', 0, 4,
     &     'DIM                           ', 0, 4,
     &     'EXP                           ', 1, 6,
     &     'FACT                          ', 1, 6,
     &     'FIX                           ', 1, 5,
     &     'FLAG                          ', 1, 5,
     &     'GRADNORM                      ', 1, 6,
     &     'GRAD                          ', 1, 5,
     &     'L                             ', 1, 5 /
      DATA (PRVAR(I),PRDIM(I),PRKIND(I),I=21,30) /
     &     'LOG                           ', 1, 6,
     &     'LOG10                         ', 1, 6,
     &     'MARG                          ', 1, 5,
     &     'MAX                           ', 0, 6,
     &     'MEAN                          ', 0, 6,
     &     'MIN                           ', 0, 6,
     &     'MOD                           ', 2, 6,
     &     'HTCOUNT                       ', 0, 4,
     &     'PCOUNT                        ', 0, 4,
     &     'PRECISION                     ', 0, 4 /
      DATA (PRVAR(I),PRDIM(I),PRKIND(I),I=31,40) /
     &     'PRINTOUT                      ', 0, 4,
     &     'PROCESS                       ', 0, 4,
     &     'R                             ', 1, 5,
     &     'RAN                           ', 1, 6,
     &     'ROUND                         ', 1, 6,
     &     'SIN                           ', 1, 6,
     &     'SINH                          ', 1, 6,
     &     'SQRT                          ', 1, 6,
     &     'STEP                          ', 1, 5,
     &     'TAN                           ', 1, 6 /
      DATA (PRVAR(I),PRDIM(I),PRKIND(I),I=41,50) /
     &     'TANH                          ', 1, 6,
     &     'TCOUNT                        ', 0, 4,
     &     'TRUNC                         ', 1, 6,
     &     'HPCOUNT                       ', 0, 4,
     &     'VALUE                         ', 0, 4,
     &     'GTCOUNT                       ', 0, 4,
     &     'GPCOUNT                       ', 0, 4,
     &     'GRMS                          ', 1, 6,
     &     'TERMS                         ', 0, 4,
     &     'TERM                          ', 1, 5 /
      DATA (PRVAR(I),PRDIM(I),PRKIND(I),I=51,NPRE) /
     &     'FUNMODE                       ', 0, 4,
     &     'NFLAGS                        ', 0, 4,
     &     'NCFLAGS                       ', 0, 4,
     &     'JTCOUNT                       ', 0, 4,
     &     'JPCOUNT                       ', 0, 4,
     &     'JACOMO                        ', 0, 4 /
C
C  MCL predefined subprograms.
      DATA (PSU(I), NARPSU(I), I=1,MXPSU) /
     &     'NORM                          ', 4,
     &     'GETSEED                       ', 1,
     &     'SETSEED                       ', 1 /
C  ---------------------------------------------------------------------
C
      IF (FIRST) THEN
         FIRST = .FALSE.
C  Sort predefined names.
         CALL CMSORT(PRVAR,PRDIM,PRKIND,NPRE)
      END IF
C
C  The total number of errors encountered in the MCL source program,
C  including warnings.
      TOTAL = 0
C
C  Rewind necessary files.
      REWIND INP
      IF (EOPEN) REWIND ERL
      REWIND SCR
      REWIND STF
C
C  Initialize per-file variables.
      NTYPE = 0
      NEXT = 0
      NCAL = 0
      DO 80,I=1,MXEXT
80       SUBUSE(I) = 0
C
C  First pass.
10    CONTINUE
      CALL ONESUB(PRVAR,PRDIM,PRKIND,NPRE,TOTAL,IEOF)
      IF (IEOF.EQ.0) GOTO 10
C
      IF (GENE) THEN
C
C  Add predefined externals.
         DO 72,I=1,NCAL
            CALL SERLOC(EXT,NEXT,CAL(I),FOUND,POSIT)
            IF (.NOT.FOUND) THEN
               CALL SERLOC(PSU,MXPSU,CAL(I),FOUND,POSIT)
               IF (FOUND) THEN
                  NEXT = NEXT+1
                  IF (NEXT.GT.MXEXT)
     &          CALL OVER('externals','EXT','MCL',MXEXT,'MXEXT')
                  EXT(NEXT) = CAL(I)
                  EXTDEF(NEXT) = BLOC+1
                  NARDEF(NEXT) = NARPSU(POSIT)
                  CALL WPRESU(CAL(I))
               END IF
            END IF
72       CONTINUE
C
C  Second pass: Write $TESTMOVE arguments, etc.
         REWIND BIN
         REWIND SCR
         IC = 0
         DO 100,I=1,NTYPE
            IF (TYPE(I).EQ.3) GOTO 100
            M = TDEF(I)
            CALL COPY(IC,M,SCR,BIN)
            WRITE (BIN,*) TREL(I)
            IC = M
            READ (SCR,'(A)') TMP
100      CONTINUE
         CALL COPY(IC,BLOC+1,SCR,BIN)
C
C  Third pass: Link external subprograms.
         REWIND BIN
         REWIND ABSFI
         WRITE (ABSFI,'(''MCL-3.0'')')
C
C  Search for the program.
         CALL SERLOC(EXT,NEXT,'..',FOUND,POSIT)
         IF (.NOT.FOUND) THEN
            CALL ERROR(182,' ')
            GOTO 500
         END IF
         SUBUSE(POSIT) = 1
         IF (EXTDEF(POSIT).NE.1) THEN
            WRITE (ABSFI,'(A)') '$MOVE'
            CALL I2STR(EXTDEF(POSIT),STR,LES)
            WRITE (ABSFI,'(A)') STR(1:LES)
         END IF
         IC = 0
         DO 70,I=1,NCAL
            M = CALDEF(I)
            CALL COPY(IC,M,BIN,ABSFI)
C  Search for the subprogram.
            CALL SERLOC(EXT,NEXT,CAL(I),FOUND,POSIT)
            IF (.NOT.FOUND) THEN
               CALL ERROR(183,CAL(I))
               GOTO 500
            END IF
            SUBUSE(POSIT) = 1
C  Check number of arguments.
            IF (NARDEF(POSIT).NE.NARCAL(I)) THEN
               CALL ERROR(37,CAL(I))
               GOTO 500
            END IF
            IR = EXTDEF(POSIT)-M
            IF (IR.LT.0) IR = IR+1
            CALL I2STR(IR,STR,LES)
            WRITE (ABSFI,'(A)') STR(1:LES)
            IC = M
            READ (BIN,'(A)') TMP
70       CONTINUE
         CALL COPY(IC,BLOC+1,BIN,ABSFI)
C
C  Warn for unused subroutines.
         DO 85,I=1,NEXT
            IF (SUBUSE(I).EQ.0) CALL WARN(24,EXT(I))
85       CONTINUE
      END IF
C
C  Issue messages if any errors or warnings. Note that TOTAL is
C  the total number of fatal+warning errors.
500   CONTINUE
      IF (NERR.NE.0) CALL FEBUFF(' ',TOTAL)
      TERR = TOTAL-NWARN
      IF (NWARN.EQ.1) THEN
         CALL I2STR(NWARN,STR,LET)
         WRITE (ERL,71) STR(1:LET), 'warning'
      ELSE IF (NWARN.GT.1) THEN
         CALL I2STR(NWARN,STR,LET)
         WRITE (ERL,71) STR(1:LET), 'warnings'
      END IF
      IF (TERR.EQ.1) THEN
         CALL I2STR(TERR,STR,LET)
         WRITE (ERL,71) STR(1:LET), 'error'
      ELSE IF (TERR.GT.1) THEN
         CALL I2STR(TERR,STR,LET)
         WRITE (ERL,71) STR(1:LET), 'errors'
      END IF
C
71    FORMAT (//$(FC)'------> ',A,' MCL ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WPRESU ( NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine WPRESU writes object code for a predefined subprogram.
C    The object code is:
C      <object code for the CALL statement>
C      $NAME
C      <object code for the SUBRETURN statement>
C
C  Input arguments:
C    NAME       Name of the predefined subprogram.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
      BLOC = BLOC+16
      WRITE (SCR,10) '$STACKSWAP'
      WRITE (SCR,10) '$PUSHINAD'
      WRITE (SCR,10) '$STACKSWAP'
      WRITE (SCR,10) '$POPINAD'
      WRITE (SCR,10) '$PUSHMB'
      WRITE (SCR,10) '$PUSHC'
      WRITE (SCR,10) '0'
      WRITE (SCR,10) '$ALLOC'
      WRITE (SCR,10) '$POPMB'
      WRITE (SCR,'(''$'',A)') NAME
      WRITE (SCR,10) '$PUSHC'
      WRITE (SCR,10) '0'
      WRITE (SCR,10) '$DEALLOC'
      WRITE (SCR,10) '$POPMB'
      WRITE (SCR,10) '$POPINAD'
      WRITE (SCR,10) '$SUBRETURN'
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      BLOCK DATA MCLDEF
C  ---------------------------------------------------------------------
C
C  Description:
C    Initializes some of the common blocks used by the MCL compiler.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
      PARAMETER ( MXFILE = 80 )
      CHARACTER*(MXFILE) IFILE, EFILE, BFILE, PDESCF
      COMMON /FILES/ IFILE, EFILE, BFILE, PDESCF
      COMMON /UALLOC/ MINU, MAXU, LASTU
      PARAMETER ( NMCOM = 99, MXPLUG = $(MXPLUG), NES = NMCOM+MXPLUG )
      CHARACTER*14 STAT(NES)
      COMMON /COM1/ STAT
      INTEGER NAME(NES), INDE(NES)
      COMMON /COM2/ NAME, INDE
C  ---------------------------------------------------------------------
      DATA INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI /
     &      -1,   6,  -1,  -1,  -1,    -1,    -1,    -1 /
C
      DATA NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, EOPEN, ISERR /
     &      0,.TRUE.,   0,      0,.FALSE.,.FALSE.,.FALSE.,   1 /
C
      DATA EFILE / '                    ' /
     &     BFILE / 'MOC                 ' /
     &     PDESCF/ 'PDESC               ' /
C
      DATA MINU, MAXU, LASTU /
     &       50,   80,    49 /
C
C  STAT holds the executable commands.
C  NAME is an identification code for each one of the STAT array
C  commands. It is used when a single subroutine processes more than
C  one MCL statement.
      DATA (STAT(I), NAME(I), I=1,10) /
     &           'ADJUST        ', 0,
     &           'ANAL          ', 0,
     &           'GNORM         ', 0,
     &           'TARGET        ', 0,
     &           'HESSIAN       ', 0,
     &           'LMARGIN       ', 4,
     &           'AUTO          ', 0,
     &           'HISTORY       ', 0,
     &           'DUMP          ', 0,
     &           'DISPLAY       ', 0 /
      DATA (STAT(I), NAME(I), I=11,20) /
     &           'INIT          ', 0,
     &           'BACKUP        ', 0,
     &           'EVALUATE      ', 0,
     &           'CONGRA        ', 0,
     &           'EXIT          ', 0,
     &           'FAST          ', 0,
     &           'FINISH        ', 0,
     &           'FIX           ', 3,
     &           'FIXALL        ', 0,
     &           'FLAG          ', 3 /
      DATA (STAT(I), NAME(I), I=21,30) /
     &           'FLAGDIS       ', 0,
     &           'FULLBACK      ', 0,
     &           'FULLPRINT     ', 0,
     &           'GET           ', 0,
     &           'GODFATHER     ', 1,
     &           'CLOSE         ', 17,
     &           'GRADDIS       ', 0,
     &           'CFLAGDIS      ', 0,
     &           'HALFPRINT     ', 0,
     &           'GRAPH         ', 0 /
      DATA (STAT(I), NAME(I), I=31,40) /
     &           'ACCUM         ', 0,
     &           'LASTBACK      ', 0,
     &           'LIMITS        ', 0,
     &           'LOOSALL       ', 0,
     &           'LOOSE         ', 4,
     &           'RMARGIN       ', 5,
     &           'TOLMIN        ', 0,
     &           'MODEDIS       ', 0,
     &           'MOVE          ', 0,
     &           'MOVETO        ', 0 /
      DATA (STAT(I), NAME(I), I=41,50) /
     &           'NOBACK        ', 0,
     &           'NOEVAL        ', 0,
     &           'NOPRINT       ', 0,
     &           'NOTARGET      ', 0,
     &           'NUMER         ', 0,
     &           'LEVE          ', 0,
     &           'ROLL          ', 0,
     &           'PAUSE         ', 0,
     &           'SIMPLEX       ', 0,
     &           'POINT         ', 1 /
      DATA (STAT(I), NAME(I), I=51,60) /
     &           'QUAD          ', 0,
     &           'QUIT          ', 3,
     &           'RESET         ', 0,
     &           'DFP           ', 0,
     &           'REVEAL        ', 0,
     &           'BFGS          ', 0,
     &           'SHORTDIS      ', 0,
     &           'STEP          ', 2,
     &           'STEPALL       ', 0,
     &           'STEPDIS       ', 0 /
      DATA (STAT(I), NAME(I), I=61,70) /
     &           'TRUST         ', 0,
     &           'VALDIS        ', 0,
     &           'GENERAL       ', 0,
     &           'SOS           ', 0,
     &           'ALIASDIS      ', 0,
     &           'CALL          ', 0,
     &           'SUBRETURN     ', 0,
     &           'LDEMARGIN     ', 1,
     &           'RDEMARGIN     ', 2,
     &           'CFLAG         ', 2 /
      DATA (STAT(I), NAME(I), I=71,80) /
     &           'MIXED         ', 3,
     &           'NONAME        ', 5,
     &           'DELETE        ', 1,
     &           'DISCARD       ', 2,
     &           'REWIND        ', 3,
     &           'GOEOF         ', 4,
     &           'HIDEOUT       ', 5,
     &           'STOP          ', 6,
     &           'RETURN        ', 7,
     &           'ALIAS         ', 8 /
      DATA (STAT(I), NAME(I), I=81,90) /
     &           'UNALIAS       ', 9,
     &           'EPILOG        ', 10,
     &           'TITLE         ', 11,
     &           'PDUMP         ', 12,
     &           'PANELON       ', 13,
     &           'PANELOFF      ', 14,
     &           'PSTATUS       ', 15,
     &           'GRADCHECK     ', 16,
     &           'PSGRAF        ',  0,
     &           'CONTROL       ',  0 /
      DATA (STAT(I), NAME(I), I=91,NMCOM) /
     &           'MEMO          ',  0,
     &           'INSPECT       ',  0,
     &           'PICK          ',  0,
     &           'TERMDIS       ',  0,
     &           'MAD           ',  0,
     &           'CONFIDENCE    ',  6,
     &           'COVARIANCE    ',  0,
     &           'JANAL         ',  0,
     &           'JNUMER        ',  0 /
C
      END
C  ---------------------------------------------------------------------
      FUNCTION MEMSPC ( IP, NDIM )
C  ---------------------------------------------------------------------
C
C  Description:
C    Calculates total memory space occupied by an MCL array.
C    This is calculated as a product of the space required for each one
C    of the array dimensions.
C
C  Input arguments:
C    IP         Integer variable. Information for the array starts
C               at position IP of the information table.
C    NDIM       Dimensionality of the array.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER HIGH
C  ---------------------------------------------------------------------
C
C  IND is a pointer to the information table.
      IND = IP-1
      MEMSPC = 1
C
C  This loop will process all array dimensions.
      DO 10,I=1,NDIM
C  Advance pointer.
         IND = IND+1
C  Get lower array bound.
         LOW = INFO(IND)
C  Advance pointer.
         IND = IND+1
C  Get upper dimension bound.
         HIGH = INFO(IND)
C  HIGH-LOW+1 is the space occupied by this dimension of the array
         MEMSPC = MEMSPC*(HIGH-LOW+1)
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MOVE ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine parses the MOVE TO statement and generates appropriate
C    object code. The syntax of this statement is:
C
C    MOVE TO <label name>       or
C    MOVETO <label name>
C
C    where <label name> is the destination label.
C    If the syntax is correct, the following two lines of object code
C    are generated:
C
C    $MOVE
C    <blank line>
C
C    The <label name> is pushed in the move table. A value of -7 is
C    pushed in the type table, to indicate the MOVE TO statement,
C    while the location of the above <blank line> is also pushed in
C    the type-definition table. The actual argument for the $MOVE
C    instruction will be calculated in pass 2, after all label
C    definitions have been read.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  Abort conditions that may occur:
C    If the move table overflows, this will cause the program to stop
C    and issue an informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
C  An identifier (the <label name>) is expected after the
C  MOVE TO statement.
      NT = NT+1
      IF (TOKEN(NT).EQ.IDENT) THEN
C  Push the destination label in the move table.
         NMOVE = NMOVE+1
         IF (NMOVE.GT.MXMOV)
     &          CALL OVER('move','MVELAB','MOVE',MXMOV,'MXMOV')
         CALL GETIDF(NA,NN,MVELAB(NMOVE))
C  Generate code.
         BLOC = BLOC+2
         IF (GENE) WRITE (SCR,'(A/)')'$MOVE'
C  Update the type and type-definition tables.
         NTYPE = NTYPE+1
         IF (NTYPE.GT.MXTYP)
     &          CALL OVER('type','TYPE','MOVE',MXTYP,'MXTYP')
         TYPE(NTYPE) = -7
         TDEF(NTYPE) = BLOC
         NT = NT+1
C  If anything else than an end of line follows the statement.
         IF (NT.NE.NTOK) CALL ERROR(9,'MOVE TO')
      ELSE
C  An identifier does not follow the MOVE TO statement.
         CALL ERROR(16,' ')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MULTI ( NT, NA, NN, NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses MCL statements that use a panel. User commands (with
C    and without a panel) declared in the Merlin configuration file
C    are parsed as well. The Merlin panel description file is
C    consulted for allowed keywords, types etc.
C    The syntax is:
C      COMMAND (KEY1 = VAL1; KEY2  = VAL2 ; ... )
C    VAL1, VAL2, etc are either arithmetic expressions or strings
C    according to the type of the corresponding keyword.
C    For a keyword designated as "output" in the panel description
C    file, the syntax is:
C      COMMAND ( KEY1? = VAR ; ... )
C    VAR is either a simple variable, or array element.
C
C    The object code is:
C
C    <code to evaluate VAL1>  -> Only if KEY1 is input and Real/Integer
C    <code to evaluate VAL2>  -> Only if KEY1 is input and Real/Integer
C     ...
C    NAME
C    NC  NCR    -> Number of changes, number of real/integer changes.
C    KEY1
C    type of KEY1
C    KEY2
C    type of KEY2
C     ...
C     ...
C
C    If any output keywords are present, then for each one:
C
C    $GETPP
C    <name of output keyword>
C    <type of output keyword (R or I)>
C    <code to push on the stack the address of the variable>
C    $POPCONT
C
C  Input arguments:
C    NAME       Name of the MCL statement.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
C
C  Maximum number of keywords for any command.
      PARAMETER ( MAXKEY = LINLEN/4, MXOKEY = LINLEN/5 )
      LOGICAL CORRE, FOUND
      LOGICAL DEFKEY(MAXKEY)
      INTEGER EX(LINLEN), KOUT(MXOKEY)
      CHARACTER*(IDLEN) KEY, STARG(1), VNAME
      CHARACTER*80 LINE
      CHARACTER CKTYPE, KOUTT(MXOKEY)
C  ---------------------------------------------------------------------
C
C  Is this a valid command ? (PLUGxx is not valid)
      IF (ISCOMP(NAME(1:4),'PLUG').NE.0) THEN
         CALL ERROR(7,' ')
         RETURN
      END IF
C
C  Is it a panel command ?
      LEV = LENGTH(NAME)
      CALL PLOCP(NAME(1:LEV),IPA)
      IF (IPA.EQ.0) THEN
C  Statement has no panel.
         NT = NT+1
         IF (NT.NE.NTOK) THEN
            CALL ERROR(71,NAME(1:LEV))
         ELSE
C  Generate object code.
            BLOC = BLOC+1
            IF (GENE) WRITE (SCR,'(A)') NAME(1:LEV)
         END IF
         RETURN
      END IF
C
C  The statement has a panel.
C  If just the command is entered without any parameters.
      NT = NT+1
      IF (NT.EQ.NTOK) THEN
         BLOC=BLOC+2
         IF (GENE) WRITE (SCR,'(A/''0 0'')') NAME(1:LEV)
         RETURN
      END IF
C
C  A left parenthesis should follow the command.
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,NAME)
         RETURN
      END IF
C
C  Initialize keyword indicators. (DEFKEY(I) = .TRUE. if keyword I
C   has been entered.)
      DO 101,I=1,MAXKEY
101      DEFKEY(I) = .FALSE.
C
C  NC is the total number of changes to the panel parameters.
      NC = 0
C
C  NCR is the number of changes to panel parameters that accept
C  integer or real keywords (expressions).
      NCR = 0
C
C  Number of lines written to unit SPARE.
      NSP = 0
C
C  Number of lines written to unit NU.
      NSP2 =0
C
C  NKOUT is the number of output keywords: KEY?=...
      NKOUT = 0
C
C  Keywords for this command start at position KP.
      KP = KPOINT(IPA)
C
C  Open the scratch files.
C
C  Unit SPARE stores all keyword/type pairs.
      OPEN (SPARE,STATUS='SCRATCH',IOSTAT=IERR)
      IF (IERR.NE.0) CALL MABORT('Cannot open the spare file.','MULTI')
      REWIND SPARE
C
C  Unit NU stores object code for the addresses of variables/array
C  elements used in output keywords.
      NU = NUNIT()
      OPEN (NU,STATUS='SCRATCH',IOSTAT=IERR)
      IF (IERR.NE.0) CALL MABORT('Cannot open a scratch file.','MULTI')
      REWIND NU
C
C  Loop over all keywords.
1000  CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         GOTO 500
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,KEY)
C
C  Locate the keyword in the panel description file.
      LEK = LENGTH(KEY)
      CALL PLOCK(IPA,KEY,I)
C
      IF (I.EQ.0) THEN
C  Invalid keyword.
         CALL ERROR(33,KEY)
         GOTO 500
      END IF
C
C  The keyword has been found.
C  Check for an output keyword: KEY?=...
C  IOUT = 1 if this is an output keyword; 0 otherwise.
      NT = NT+1
      IOUT = 0
      IF (TOKEN(NT).EQ.QMARK) THEN
         IOUT = 1
         NT = NT+1
      END IF
C
C  The = sign is expected.
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(35,' ')
         GOTO 500
      END IF
C
C  CKTYPE is the type of the keyword (I, R, S or C).
      CKTYPE = KTYPE(I)
C
      IF (IOUT.EQ.0) THEN
C
C  An input keyword.
         IF (KIO(I).NE.' ') CALL WARN(39,KEY)
C
C  Only one occurrence of the keyword is allowed.
         IKP = I-KP+1
         IF (DEFKEY(IKP)) THEN
            CALL ERROR(166,KEY)
            GOTO 500
         END IF
         DEFKEY(IKP) = .TRUE.
C
         NC = NC+1
         WRITE (SPARE,'(A/A)') KEY(1:LEK), CKTYPE
         NSP = NSP+2
C
C  Proceed according to keyword type.
         IF (ISCOMP(CKTYPE,'S').NE.0 .OR.
     &          ISCOMP(CKTYPE,'C').NE.0) THEN
C  String.
            IF (TOKEN(NT+1).NE.QUOTE) THEN
               CALL ERROR(169,KEY)
               GOTO 500
            END IF
            NT = NT+2
            IF (TOKEN(NT).NE.STRING) THEN
               CALL ERROR(28,' ')
               GOTO 500
            END IF
C  Get the string.
            CALL GETIDS(NA,NN,IS1,IE1)
C  LL is the string length in characters.
            LL = IE1-IS1+1
            NT = NT+1
            IF (TOKEN(NT).NE.QUOTE) THEN
               CALL ERROR(28,' ')
               GOTO 500
            END IF
            WRITE (SPARE,*) LL
            WRITE (SPARE,'(A)') ALPHA(IS1:IE1)
            NSP = NSP+2
         ELSE
C  Extract the expression following the = sign.
C  Stop at the first invalid token, or the last right parenthesis
C  before the end of line.
            CALL EXTRA(NT,EX,NEX,2,NA)
C  Evaluate the expression.
            CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C  If expression incorrect.
            IF (.NOT.CORRE) THEN
               CLOSE (NU2,STATUS='DELETE')
               GOTO 500
            END IF
            NCR = NCR+1
         END IF
      ELSE
C
C  Output keyword.
C
C  A string cannot be returned.
         IF (ISCOMP(CKTYPE,'S').NE.0 .OR.
     &          ISCOMP(CKTYPE,'C').NE.0 ) THEN
            CALL ERROR(38,KEY)
            GOTO 500
         END IF
         NKOUT = NKOUT+1
         KOUT(NKOUT) = NA
         KOUTT(NKOUT) = CKTYPE
         NT = NT+1
C
C  An identifier (simple variable or array element) is expected.
         IF (TOKEN(NT).NE.IDENT) THEN
            CALL ERROR(40,KEY)
            GOTO 500
         END IF
C
C  Get the name of the variable.
         CALL GETIDF(NA,NN,VNAME)
C
C  Check if present in the var table.
         CALL LOCAN(VAR,NVAR,VNAME,FOUND,IND)
         IF (.NOT.FOUND) THEN
C  Variable not present in the var table.
            CALL ERROR(19,VNAME)
C  Put variable in the var table. Update related tables as well.
            CALL INSERT(VNAME,0,0,0,-1,IND)
         END IF
C
C  Calculate memory address for this variable.
         K = KIND(IND)
         IF (TOKEN(NT+1).NE.LBRA) THEN
C  The next token is not a left bracket. Thus this is a simple
C  variable.
            IF (K.NE.1 .AND. K.NE.-1 .AND. K.NE.101)
     &                CALL ERROR(153,VNAME)
C  Check whether the variable is used as a loop control index.
            IF (ISLCI(IND).NE.0) CALL ERROR(141,VNAME)
            L = LOC(IND)
C  Generate code necessary for its address evaluation.
            IF (K.EQ.1) THEN
               NSP2 = NSP2+2
               IF (GENE) THEN
                  WRITE (NU,'(A)') '$PUSHA'
                  WRITE (NU,*) L
               END IF
            ELSE
               CALL WPSC(NU,NSP2,0)
               CALL WPSC(NU,NSP2,L)
               NSP2 = NSP2+1
               IF (GENE) WRITE (NU,'(A)') '$IOFFSET'
            END IF
         ELSE
C  This is an array element.
            NT = NT-1
            NA = NA-1
            NN = NN-2
C  Extract the array element as an expression.
            CALL EXTRA(NT,EX,NEX,2,NA)
C  Check whether an expression is involved. (Only an array element
C  is allowed.) Expressions for the array subscript are ok.
            NBR = 1
            DO 30,IEX=3,NEX
               IF (EX(IEX).EQ.LBRA) NBR = NBR+1
               IF (EX(IEX).EQ.RBRA) NBR = NBR-1
               IF (NBR.EQ.0) THEN
                  IF (IEX.NE.NEX) CALL ERROR(10,KEY)
                  GOTO 31
               END IF
30          CONTINUE
            CALL ERROR(10,KEY)
31          CONTINUE
C  Evaluate this expression. The object code will be written
C  in unit NU2. The number of object code lines written
C  there is counted by NLS.
            NLS = 0
            NU2 = NUNIT()
            IF (NU2.EQ.-1)
     &       CALL MABORT('Cannot open a scratch file.','MULTI')
            OPEN (NU2,STATUS='SCRATCH',IOSTAT=IERR)
            IF (IERR.NE.0)
     &       CALL MABORT('Cannot open a scratch file.','MULTI')
            REWIND NU2
            CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,NLS,NU2,STARG)
            IF (GENE) THEN
C  Copy unit NU2 to unit NU. The final $GETCONT
C  instruction will not be written in the scratch file.
               REWIND NU2
               DO 88,I=1,NLS-1
                  READ (NU2,'(A)') LINE
                  WRITE (NU,'(A)') LINE
88             CONTINUE
               NSP2 = NSP2+NLS-1
            END IF
            CLOSE (NU2,STATUS='DELETE',IOSTAT=IERR)
            IF (IERR.NE.0)
     &     CALL MABORT('Cannot close a scratch file.','MULTI')
C  Array element location is now on the top of the stack.
         END IF
C
C  Check if the name is one of the Merlin variables,
C  Merlin arrays or intrinsic functions.
         IF (K.EQ.4 .OR. K.EQ.5 .OR. K.EQ.6) THEN
            CALL ERROR(18,VNAME)
         ELSE IF (K.EQ.3) THEN
            CALL ERROR(153,VNAME)
         END IF
C
      END IF
C
      NT = NT+1
      CUR = TOKEN(NT)
C
C  A semicolon denotes another keyword-value pair is present.
      IF (CUR.EQ.SEMI) GOTO 1000
C
C  A right parenthesis terminates the statement, so write the
C  final object code.
      IF (CUR.EQ.RIGHTP) THEN
         BLOC = BLOC+(2+NSP)
         IF (GENE) THEN
C  Write the command and number of changes.
            WRITE (SCR,'(A/2I3)') NAME(1:LEV), NC, NCR
C  Write keywords and types.
            REWIND SPARE
            DO 89,I=1,NSP
               READ (SPARE,'(A)') LINE
               WRITE (SCR,'(A)') LINE
89          CONTINUE
C  Write $GETPP instructions for all output keywords.
            IF (NKOUT.NE.0) THEN
               BLOC = BLOC+NSP2
               REWIND NU
               DO 79,I=1,NSP2
                  READ (NU,'(A)') LINE
                  WRITE (SCR,'(A)') LINE
79             CONTINUE
C
               DO 71,I=NKOUT,1,-1
                  CALL GETIDA(KOUT(I),KEY)
                  WRITE (SCR,'(A/A/A/A/A)')
     &                 '$GETPP', NAME(1:LEV), KEY, KOUTT(I), '$POPCONT'
                  BLOC = BLOC+5
71             CONTINUE
               GOTO 500
            END IF
         END IF
      ELSE
C  Neither semicolon nor a right parenthesis was found.
         CALL ERROR(54,' ')
      END IF
      RETURN
C
C  Close SPRARE, NU and return.
500   CONTINUE
      CLOSE (SPARE,STATUS='DELETE',IOSTAT=IERR)
      IF (IERR.NE.0)
     &    CALL MABORT('Cannot close the spare file.','MULTI')
      CLOSE (NU,STATUS='DELETE',IOSTAT=IERR)
      IF (IERR.NE.0)
     &    CALL MABORT('Cannot close a scratch file.','MULTI')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE NOPAR ( NT, NAME, OPTION )
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes commands with no parameters. Only one line of object code
C    is generated:
C
C    <NAME>
C
C    if the command resides already in the Merlin operating system, or:
C
C    $<NAME>
C
C    if the command is implemented for MCL use only.
C    <NAME> is the command to be processed and is input to the routine.
C
C  Input arguments:
C    NAME       The command to be processed.
C    OPTION     This is an option. If OPTION=0 the command resides in
C               the Merlin operating system. If OPTION=1 the command is
C               implemented for MCL use only and will be prefixed by a
C               dollar sign when written to the scratch file.
C
C  Input / Output arguments:
C    NT         Integer variable. This is the number of tokens in the
C               input line used so far by the previously called
C               parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
      NT = NT+1
C
C  If nothing else follows the command, generate code.
      IF (NT.EQ.NTOK) THEN
         BLOC = BLOC+1
         IF (GENE) THEN
C  Code is generated, according to the input option.
            IF (OPTION.EQ.0) WRITE (SCR,'(A)') NAME
            IF (OPTION.EQ.1) WRITE (SCR,'(''$'',A)') NAME
         END IF
      ELSE
C  The command is not followed by an end of line.
         CALL ERROR(9,NAME)
      END  IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ONESUB ( PRVAR, PRDIM, PRKIND, NPRE, TOTAL, IEOF )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine ONESUB compiles (pass 1) an MCL subprogram (or main
C    program).
C
C  Input arguments:
C    PRVAR      The table of predefined names (variables, functions
C               Merlin arrays etc).
C    PRDIM      The dimensionality table for the predefined names.
C    PRKIND     The kind table for the predefined names.
C    NPRE       Number of predefined names.
C
C  Output arguments:
C    IEOF       IEOF = 1 -> An End-of-File was encountered in the MCL
C                           source file.
C               IEOF = 0 -> No End-of-File in the MCL source file.
C
C  Input / Output arguments:
C    TOTAL      Total number of errors in the MCL source file.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) PRVAR(*)
      INTEGER PRDIM(*), PRKIND(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) LABID
      LOGICAL FOUND
      CHARACTER*(LENONE) ONELIN
      CHARACTER*(LINLEN) LINE
      CHARACTER*(IDLEN) IDF
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
C  Copy predefined names to the var table.
      DO 65,I=1,NPRE
         VAR(I) = PRVAR(I)
         DIM(I) = PRDIM(I)
         KIND(I) = PRKIND(I)
         LOC(I) = 0
         IPOS(I) = 0
         INFO(I) = 0
65    CONTINUE
C
C  Initialize counters in the /MCLCNT/ common block.
C
C  Number of labels.
      NLAB = 0
C
C  Entries in the var table.
      NVAR = NPRE
C
C  MOVE TO statements.
      NMOVE = 0
C
C  Entries in the information table.
      NINF = NPRE
C
C  Number of lines in the statement function file.
      NSTL = 0
C
C  Number of nested loops (nesting level).
      NLOO = 0
C
C  Number of errors in a source line (including warnings).
      NERR = 0
C
C  Number of warnings in a source line.
      NWARN = 0
C
C  Needed memory.
      MEMLOC = 0
C
C  What are we compiling (PROGRAM or SUB) ?
      INWHAT = 0
C
C  Starting point in the type table, for this subprogram.
      TSTART = NTYPE+1
C
C  Start compiling the MCL program.
C
C  Try to read lines containing a PROGRAM or SUB statement.
11    CONTINUE
      CALL OLREAD(INP,MXCONT,ONELIN,LINE,ILL,NLCO,IERR)
      IF (IERR.EQ.1) THEN
         CALL I2STR(MXCONT,STR,LES)
         CALL ERROR(47,STR(1:LES))
      ELSE IF (IERR.EQ.2) THEN
         IF (ILL.NE.0) CALL ERROR(51,' ')
         GOTO 70
      END IF
      NL = NL+NLCO
C
C  Analyze the input line in a sequence of tokens.
      CALL LEX(LINE,ILL,TOKEN,NTOK,NUM,ALPHA)
C
C  Check if this is a comment line, or a blank line.
      INIT = TOKEN(1)
      IF (INIT.EQ.GT .OR. INIT.EQ.EOL) GOTO 11
      IF (INIT.NE.IDENT) THEN
         CALL ERROR(178,' ')
         GOTO 17
      END IF
      CALL GETIDA(1,IDF)
      IF (IDF.EQ.'PROGRAM') THEN
         CALL MAINPR
         INWHAT = 1
      ELSE IF (IDF.EQ.'SUB') THEN
         CALL SUBR
         INWHAT = 2
      ELSE
         CALL ERROR(178,' ')
         GOTO 17
      END IF
C
C  If any error occurred in this line, flush the error buffer.
      IF (NERR.NE.0) CALL FEBUFF(LINE,TOTAL)
C
C  Try to read lines containing a VAR statement.
10    CONTINUE
      CALL OLREAD(INP,MXCONT,ONELIN,LINE,ILL,NLCO,IERR)
      IF (IERR.EQ.1) THEN
         CALL I2STR(MXCONT,STR,LES)
         CALL ERROR(47,STR(1:LES))
      ELSE IF (IERR.EQ.2) THEN
         IF (ILL.NE.0) CALL ERROR(51,' ')
         GOTO 70
      END IF
      NL = NL+NLCO
C
C  Analyze the input line in a sequence of tokens.
      CALL LEX(LINE,ILL,TOKEN,NTOK,NUM,ALPHA)
C
C  Check if this is a comment line, or a blank line.
      INIT = TOKEN(1)
      IF (INIT.EQ.GT .OR. INIT.EQ.EOL) GOTO 10
C
C  If this is not a VAR declaration.
17    CONTINUE
      IF (INIT.NE.IDENT) GOTO 30
      CALL GETIDA(1,IDF)
      IF (IDF.NE.'VAR') GOTO 47
      NT = 1
      NA = 1
      NN = 2
      CALL VARIA(0,NT,NA,NN,NVDECL)
C
C  If any error occurred in this line, flush the error buffer.
      IF (NERR.NE.0) CALL FEBUFF(LINE,TOTAL)
      GOTO 10
C
C  Try to read lines containing a FUNCTION declaration.
37    CONTINUE
      CALL OLREAD(INP,MXCONT,ONELIN,LINE,ILL,NLCO,IERR)
      IF (IERR.EQ.1) THEN
         CALL I2STR(MXCONT,STR,LES)
         CALL ERROR(47,STR(1:LES))
      ELSE IF (IERR.EQ.2) THEN
         IF (ILL.NE.0) CALL ERROR(51,' ')
         GOTO 70
      END IF
      NL = NL+NLCO
C
C  Analyze the input line in a sequence of tokens.
      CALL LEX(LINE,ILL,TOKEN,NTOK,NUM,ALPHA)
C
C  Check if this is a comment line, or a blank line.
      INIT = TOKEN(1)
      IF (INIT.EQ.GT .OR. INIT.EQ.EOL) GOTO 37
C
C  If this is not a FUNCTION declaration.
      IF (INIT.NE.IDENT) GOTO 30
47    CONTINUE
      CALL GETIDA(1,IDF)
      IF (IDF.NE.'FUNCTION') GOTO 30
      CALL STAFU
C
C  If any error occurred in this line, flush the error buffer.
      IF (NERR.NE.0) CALL FEBUFF(LINE,TOTAL)
      GOTO 37
C
C  VAR and FUNCTION statements processed. Now compile
C  the rest of the program.
60    CONTINUE
      CALL OLREAD(INP,MXCONT,ONELIN,LINE,ILL,NLCO,IERR)
      IF (IERR.EQ.1) THEN
         CALL I2STR(MXCONT,STR,LES)
         CALL ERROR(47,STR(1:LES))
      ELSE IF (IERR.EQ.2) THEN
         IF (ILL.NE.0) CALL ERROR(51,' ')
         GOTO 70
      END IF
      NL = NL+NLCO
C
C  Analyze the input line in a sequence of tokens.
      CALL LEX(LINE,ILL,TOKEN,NTOK,NUM,ALPHA)
C
C  Check if this is a comment line, or a blank line.
      INIT = TOKEN(1)
      IF (INIT.EQ.GT .OR. INIT.EQ.EOL) GOTO 60
C
30    CONTINUE
C
C  If the debug option has been selected, write debug information.
      IF (DEBUG .AND. GENE) THEN
         BLOC = BLOC+3
         CALL I2STR(NL,STR,LES)
         WRITE (SCR,'(A/A)') '$LINE', STR(1:LES)
         WRITE (SCR,'(''$>> '',A)') LINE(1:MIN(80,LEN(LINE)))
      END IF
C
C  PCG is the main routine that will determine what kind of statement
C  is it about, and will call appropriate program units to compile it.
      CALL PCG(IEND)
C
C  If any error occurred in this line, flush the error buffer.
      IF (NERR.NE.0) CALL FEBUFF(LINE,TOTAL)
C
C  If an end statement has been found terminate compilation of this
C  subprogram.
      IF (IEND.EQ.1) THEN
         IEOF = 0
         GOTO 77
      END IF
      GOTO 60
C
70    CONTINUE
      IEOF = 1
77    CONTINUE
C
C  Did we compile comments and empty lines only ?
      IF (INWHAT.EQ.0) RETURN
      IF (IEOF.NE.0 .AND. INWHAT.NE.0) THEN
         CALL ERROR(184,' ')
         CALL FEBUFF(' ',TOTAL)
      ELSE
         CALL ENDSUB
      END IF
C
C  If any error, flush the error buffer.
      IF (NERR.NE.0) CALL FEBUFF(LINE,TOTAL)
C
C  The first pass over the MCL source has been completed.
C
C  Link jumps due to MOVE TO etc. statements.
C
      K = TSTART
      DO 700,I=1,NMOVE
C  Try to locate a type of -7 (a MOVE TO) in the type table
600      K = K+1
         IF (TYPE(K).NE.-7) GOTO 600
C  A type of -7 was found in position K of the type table.
C  Get the destination label of this MOVE TO and check
C  whether it is defined.
         LABID = MVELAB(I)
         CALL LOCAN(LABEL,NLAB,LABID,FOUND,IND)
         IF (FOUND) THEN
C  Calculate the type-link table entry for this MOVE TO.
            IR = LABLOC(IND)-TDEF(K)
            IF (IR.LT.0) IR = IR+1
            TREL(K) = IR
         ELSE
C  The label is not defined.
            CALL ERROR(11,LABID)
C  Put the label in the label table to avoid further error messages.
C  The location assigned to this label is: 0
            NLAB = NLAB+1
            IF (NLAB.GT.MXLAB)
     &              CALL OVER('label','LABEL','MCL',MXLAB,'MXLAB')
            DO 21,J=NLAB,IND+1,-1
               LABEL(J) = LABEL(J-1)
               LABLOC(J) = LABLOC(J-1)
21          CONTINUE
            LABEL(IND) = LABID
            LABLOC(IND) = 0
         END IF
700   CONTINUE
C
C  Check for block IFs and LOOPs that do not terminate.
      CALL TERMIN
C
C  Check the validity of the MOVE TO statements in relation with
C  block IFs and LOOPs.
      CALL VALIJ
C
C  If any error, flush the error buffer.
      IF (NERR.NE.0) CALL FEBUFF(' ',TOTAL)
C
C  Fill in memory requirement (alloc/dealloc) for this subprogram.
      CALL MCALC
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OVER ( TABLE, ARRAY, SUB, MAXIM, PARAM )
C  ---------------------------------------------------------------------
C
C  Description:
C    When any of the MCLCOM tables overflows, this routine will be
c    called to issue informative messages and stop execution of the
C    program.
C
C  Input arguments:
C    TABLE      Character variable. This the name of the table that
C               overflowed.
C    ARRAY      Character variable. This is the name of the array
C               implementing the table.
C    SUB        Character variable. This is the name of the routine
C               in which the overflow occurred.
C    MAXIM      Integer variable. This is the maximum number of entries
C               allowed in the table.
C    PARAM      Character variable. This is the name of the parameter
C               that dimensions the table.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) TABLE, ARRAY, SUB, PARAM
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      CALL I2STR(MAXIM,STR,LES)
      WRITE (*,10) TABLE, SUB
      WRITE (*,20) ARRAY, STR(1:LES)
      WRITE (*,30) PARAM
C
C  Write the same message on the error list file.
      IF (EOPEN) THEN
         WRITE (ERL,10) TABLE, SUB
         WRITE (ERL,20) ARRAY, MAXIM
         WRITE (ERL,30) PARAM
      END IF
C
      STOP 'ABRT'
C
10    FORMAT (/$(FC)'**** The ',A,' table has overflown in ',
     &        'routine ',A,'.')
20    FORMAT ($(FC)'**** Array ',A,' was assigned only ',A,
     &        ' storage locations.')
30    FORMAT ($(FC)'**** This array is dimensioned by the ',A,
     &        ' parameter.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PCG ( IEND )
C  ---------------------------------------------------------------------
C
C  Description:
C    This is the main routine that drives the others according to the
C    input line. This routine is called for executable statements (ie.
C    statements that produce object code). At this point all declaration
C    statements have been processed.
C    The first token encountered in the input line, should be an
C    identifier, otherwise an MCL-error is issued.
C    If a colon follows, then the routine will assume that this is a
C    label definition line.
C    If an equality sign, or a left bracket follows, the routine will
C    assume, that this is an assignment.
C    If its neither a label, nor an assignment, the routine will test
C    for one of the following statements:
C    WHEN, IF, ELSE, ENDIF, LOOP, ENDLOOP
C    if these tests fail also, the subroutine EXE will be called to
C    test for the rest of the MCL statements.
C
C  Output arguments:
C    IEND       IEND = 1 -> The current MCL source line contains an
C                           END statement that terminates the
C                           subprogram under processing.
C               IEND = 1 -> An END statement has not been found.
C
C  Notes:
C    Note that this routine processes directly all MCL-statements that
C    may appear next to a WHEN statement. Subroutine EXE will take
C    care of the rest of the statements.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME, IDF
C  ---------------------------------------------------------------------
C
      IEND = 0
C
C  Get the first token from the input line.
      INIT = TOKEN(1)
C
C  The first token should be an identifier, since all MCL-statements
C  start this way.
      IF (INIT.EQ.IDENT) THEN
         CALL GETIDA(1,NAME)
         NXT = TOKEN(2)
         IF (NXT.EQ.COLON) THEN
C  Possibly a label.
            CALL LBDEF(NAME)
         ELSE IF (NXT.EQ.EQ .OR. NXT.EQ.LBRA) THEN
C  Possibly an assignment.
            NT = 0
            NA = 0
            NN = 0
            CALL EXE(NT,NA,NN)
         ELSE IF (NAME.EQ.'WHEN') THEN
C  This is a WHEN statement.
            CALL WHEN
         ELSE IF (NAME.EQ.'IF') THEN
C  This is an IF statement.
            CALL IF
         ELSE IF (NAME.EQ.'ELSE') THEN
C  This is an ELSE statement.
            CALL ELSE
C  This is an ENDIF statement.
         ELSE IF (NAME.EQ.'ENDIF') THEN
            NT = 1
            CALL ENDIF(NT)
C  This is a LOOP statement.
         ELSE IF (NAME.EQ.'LOOP') THEN
            CALL LOOP
C  This is an ENDLOOP statement.
         ELSE IF (NAME.EQ.'ENDLOOP') THEN
            NT = 1
            CALL ENDLOO(NT)
         ELSE IF (NAME.EQ.'END') THEN
            IF (NXT.EQ.IDENT) THEN
C  This may be the case of an END IF or an END LOOP.
               CALL GETIDA(2,IDF)
               IF (IDF.EQ.'IF') THEN
                  NT = 2
                  CALL ENDIF(NT)
               ELSE IF (IDF.EQ.'LOOP') THEN
                  NT = 2
                  CALL ENDLOO(NT)
               ELSE
                  NT = 0
                  NA = 0
                  NN = 0
C  Neither END IF nor END LOOP was found.
                  CALL EXE(NT,NA,NN)
               END IF
            ELSE IF (NXT.EQ.EOL) THEN
               IEND = 1
            ELSE
               NT = 0
               NA = 0
               NN = 0
C  Neither END IF nor END LOOP was found.
               CALL EXE(NT,NA,NN)
            END IF
         ELSE IF (NAME.EQ.'VAR' .OR. NAME.EQ.'FUNCTION' .OR.
     &            NAME.EQ.'PROGRAM' .OR. NAME.EQ.'SUB') THEN
C  When this routine is called, all declaration statements have been
C  processed. A VAR or FUNCTION statement is misplaced here.
            CALL ERROR(42,NAME)
         ELSE
C  Possibly any of the other MCL statements.
            NT = 0
            NA = 0
            NN = 0
            CALL EXE(NT,NA,NN)
         END IF
      ELSE
C  All MCL statements begin with an identifier.
         CALL ERROR(7,' ')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PFIX ( EXPR, TOKNUM, NEWTOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Transforms expressions from infix form to a hybrid postfix form.
C
C  Input arguments:
C    EXPR       Array holding the tokens of the expression.
C    TOKNUM     Number of tokens in array EXPR.
C    NEWTOK     Number of tokens left in EXPR upon return.
C
C  Notes:
C    (I) The hybrid notation mentioned earlier is as follows:
C        If the expression does not contain function references
C        or array elements, then it coincides with the standard
C        postfix form. For example:
C        (A+B)-C*2 (infix) becomes AB+C2*- (postfix)
C        If the expression contains either arrays or function
C        references, then the following is done:
C        A[1,2]*MAX[4,X] (infix) becomes A1,2]MAX4,X]* (postfix)
C        that is, it is postfix but the arguments and/or subscripts
C        remain in their order as the expression is
C        being transformed, and the final bracket also remains
C        to terminate the sequence (commas also remain). If we
C        face : A[X-2*B,1] (infix) we write AX2B*-,1] (postfix)
C        that is, the items enclosed in brackets, which are
C        simple expressions are transformed in standard postfix
C        but appear in their order in the final hybrid form.
C
C        From the above description it should be clear that
C        special processing is needed (when code is to be generated
C        from the hybrid postfix form) so that items enclosed
C        in brackets, be processed in a way to compensate for the
C        absence of full postfix arrangement.
C
C   (II) It is assumed that EXPR has been checked for validity by
C        subroutine XCHECK.
C
C   (III) The priority of operators is:
C                                      not        9
C                                       **        8
C                                      *,/        7
C                                      +,-        6
C                             >,<,:=:,#,<=,>=     5
C                                      and        4
C                                       or        3
C                                      xor        2
C                              parentheses        1
C
C         operands (i.e. numeric constants and identifiers ) typically
C         have priority 10, since they are appended to postfix once
C         they have been detected. The same is true for parenthesized
C         expressions, functions calls or array references.
C         Of course the validity of the last statement, is descriptory,
C         since "(" has priority 1. Note though, that this
C         is exactly the end result.
C
C   (IV)  the associativity of all the above operators is
C         from left to right.
C
C   (V)  the algorithm followed, can be summarized as follows:
C
C        For each token into expression do:
C           Check if token = operand, then
C               append token to postfix
C           else if token = leftparen or token = leftbra then
C               push token to stack
C           else if token = rightparen then
C               pop item from stack
C               while item not leftparen do
C                    append item to postfix
C                    pop item from stack
C           else if token = rightbra then
C               pop item from stack
C               while item not leftbra do
C                    append item to postfix
C                    pop item from stack
C               if item = leftbra then
C                    append token to postfix
C           else if token = comma then
C               pop item from stack
C               while item not leftbra and item not comma do
C                    append item to postfix
C                    pop item from stack
C               append token to postfix
C               if item = leftbra then
C                    push leftbra to stack
C               else if token = operator then
C                   pop item from stack
C                   if no item left then
C                       push token to stack
C                   else if tokenpriority > itempriority then
C                       push item to stack
C                       push token to stack
C                   else
C                       while item not leftparen do
C                            append token to postfix
C                            pop item from stack
C                            if no item left then
C                                push token to stack
C                            else if tokenpriority > itempriority then
C                                push item to stack
C                                push token to stack
C                        if item = leftparen then
C                            push leftparen to stack
C                            push token to stack
C             else
C                 stop
C        while not empty stack do
C           pop item from stack
C           append item to postfix
C        end
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER EXPR(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      INTEGER STACK(1:LINLEN)
      COMMON /EXPRST/ TOPBUF, STACK
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( ONE = 1000, ZERO = -1000 )
      DIMENSION PFBUF(1:LINLEN), PRIOR(1:31)
      LOGICAL OPERAN, OPRTOR, LFTPAR, RTPAR, EMPTY, RTBRA, LFTBRA
      LOGICAL ISCOMA
      DATA (PRIOR(I),I=1,31) / 1,0,0,0,5,5,5,0,5,5,6,6,7,7,8,
     &                         0,0,0,0,0,0,0,0,1,0,0,4,3,2,9,5/
      OPERAN() = (TEMPO.EQ.IDENT) .OR. (TEMPO.EQ.UNSI)   .OR.
     &           (TEMPO.EQ.ZERO)  .OR. (TEMPO.EQ.ONE)
      OPRTOR() = (TEMPO.EQ.PLUS)  .OR. (TEMPO.EQ.MINUS)  .OR.
     &           (TEMPO.EQ.STAR)  .OR. (TEMPO.EQ.BISTAR) .OR.
     &           (TEMPO.EQ.SLASH) .OR. (TEMPO.EQ.AND)    .OR.
     &           (TEMPO.EQ.OR)    .OR. (TEMPO.EQ.LT)     .OR.
     &           (TEMPO.EQ.LE)    .OR. (TEMPO.EQ.GT)     .OR.
     &           (TEMPO.EQ.GE)    .OR. (TEMPO.EQ.EQUAL)  .OR.
     &           (TEMPO.EQ.NE)    .OR. (TEMPO.EQ.XOR)    .OR.
     &           (TEMPO.EQ.NOT)
      LFTPAR () = (TEMPO.EQ.LEFTP)
      RTPAR ()  = (TEMPO.EQ.RIGHTP)
      RTBRA ()  = (TEMPO.EQ.RBRA)
      ISCOMA () = (TEMPO.EQ.COMMA)
      LFTBRA () = (TEMPO.EQ.LBRA)
C  ---------------------------------------------------------------------
C
      TOPBUF = 0
      BUFIND = 0
C
      DO 1000,K=1,TOKNUM
         TEMPO = EXPR ( K )
         IF (OPERAN()) THEN
C
C  Place operand at the rightmost empty position of postfix expression.
            BUFIND = BUFIND+1
            PFBUF(BUFIND) = TEMPO
C
C  Go to get next token.
         ELSE IF (LFTPAR()) THEN
C
C  Push left parenthesis into stack.
            CALL PUSH(LEFTP)
C
         ELSE IF (LFTBRA()) THEN
C
C  Push left bracket into stack
            CALL PUSH(LBRA)
C
         ELSE IF (RTPAR()) THEN
C
12          CONTINUE
C  Pop operator
            CALL POP(OPER,EMPTY)
            IF (OPER.EQ.LEFTP) THEN
               GOTO 1000
            ELSE
               BUFIND = BUFIND+1
               PFBUF(BUFIND) = OPER
C
C  Go to get next operator.
               GOTO 12
            END IF
C
         ELSE IF (RTBRA()) THEN
C
C  Bracket case. Continue to get a sequence of operators.
122         CONTINUE
            CALL POP(OPER,EMPTY)
            IF (OPER.EQ.LBRA) THEN
               BUFIND = BUFIND+1
               PFBUF(BUFIND) = RBRA
               GOTO 1000
            ELSE
               BUFIND = BUFIND+1
               PFBUF(BUFIND) = OPER
C
C   Another operator appended to postfix. Go to get next.
               GOTO 122
            END IF
C
         ELSE IF (ISCOMA()) THEN
C
C  For commas the processing is similar to rtbra.
152         CONTINUE
            CALL POP(OPER,EMPTY)
            IF (OPER.EQ.LBRA .OR. OPER.EQ.COMMA) THEN
               BUFIND = BUFIND + 1
               PFBUF(BUFIND) = COMMA
               IF (OPER.EQ.LBRA) THEN
                  CALL PUSH ( OPER )
               END IF
               GOTO 1000
            ELSE
               BUFIND = BUFIND + 1
               PFBUF(BUFIND) = OPER
C
C  Go to get next oper.
               GOTO 152
            END IF
C
         ELSE IF (OPRTOR()) THEN
C
C  Pop topstack-element and if not nil, check priorities.
            CALL POP(TOPSTK,EMPTY)
            IF (EMPTY) THEN
               CALL PUSH(TEMPO)
               GOTO 1000
            END IF
            IF (PRIOR(TEMPO) .GT. PRIOR(TOPSTK)) THEN
               CALL PUSH(TOPSTK)
               CALL PUSH(TEMPO)
               GOTO 1000
            ELSE
14             CONTINUE
               IF (TOPSTK.EQ.LEFTP) THEN
                  CALL PUSH(TOPSTK)
                  CALL PUSH(TEMPO)
                  GOTO 1000
               END IF
               BUFIND = BUFIND+1
               PFBUF(BUFIND) = TOPSTK
               CALL POP(TOPSTK,EMPTY)
               IF (EMPTY) THEN
                  CALL PUSH(TEMPO)
                  GOTO 1000
               END IF
               IF (PRIOR(TOPSTK).GE.PRIOR(TEMPO)) THEN
                  GOTO 14
               ELSE
                  CALL PUSH(TOPSTK)
                  CALL PUSH(TEMPO)
                  GOTO 1000
               END IF
            END IF
C
         ELSE
            CALL MABORT('Illegal item into expression.','PFIX')
         END IF
C
1000  CONTINUE
C
C  Push  components left at stack to postfix.
99    CONTINUE
C
      CALL POP(TOPSTK,EMPTY)
      IF (.NOT. EMPTY) THEN
         BUFIND = BUFIND+1
         PFBUF(BUFIND) = TOPSTK
C
C  Another item appended to postfix. Go to get next.
         GOTO 99
      END IF
C
      DO 8,L=1,BUFIND
8        EXPR(L) = PFBUF(L)
C
      NEWTOK = BUFIND
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POP ( ITEM , EMPTY )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is used by subroutine PFIX.
C    It pops out items from the expression stack.
C
C  Output arguments:
C    ITEM       The element poped out.
C    EMPTY      Logical variable:
C               EMPTY = .TRUE.  -> The stack is empty.
C               EMPTY = .FALSE. -> The stack is not empty.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      LOGICAL EMPTY
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      INTEGER STACK(1:LINLEN)
      COMMON /EXPRST/ TOPBUF, STACK
C  ---------------------------------------------------------------------
C
      TOP = TOPBUF
      IF (TOP.GE.1) THEN
         ITEM = STACK(TOP)
         TOP = TOP-1
         EMPTY = .FALSE.
      ELSE
         EMPTY = .TRUE.
      END IF
      TOPBUF = TOP
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POPBR ( ITEM )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is used by subroutine XCHECK.
C    It pops integer positions up from the BSTACK stack.
C
C  Output arguments:
C    ITEM       An integer position returned to XCHECK
C               to be used in brackets processing.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      INTEGER BSTACK(1:LINLEN)
      LOGICAL OK
      COMMON /TSTBST/ TOPBST, BSTACK, OK
C  ---------------------------------------------------------------------
C
      IF (TOPBST.GE.1) THEN
         ITEM = BSTACK(TOPBST)
         TOPBST = TOPBST-1
         OK = .TRUE.
      ELSE
         OK = .FALSE.
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE POPUP ( ITEM )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is called from subroutine XCHECK.
C    It pops numbers from the xcheck stack.
C
C  Output arguments:
C    ITEM       The number returned from the stack.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      INTEGER STACK(1:LINLEN)
      LOGICAL SUCCES
      COMMON /TSTXST/ TOPST, STACK, SUCCES
C  ---------------------------------------------------------------------
C
      IF (TOPST.GE.1) THEN
         ITEM = STACK(TOPST)
         TOPST = TOPST-1
         SUCCES = .TRUE.
      ELSE
         SUCCES = .FALSE.
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PUSH ( ITEM )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is used by subroutine PFIX.
C    It pushes items into the expression stack.
C
C  Input arguments:
C    ITEM       The element pushed.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      INTEGER STACK(1:LINLEN)
      COMMON /EXPRST/ TOPBUF, STACK
C  ---------------------------------------------------------------------
C
      TOP = TOPBUF
      IF (TOP.LT.LINLEN) THEN
         TOP = TOP+1
         STACK(TOP) = ITEM
      ELSE
         CALL MABORT('Stack overflow.','PUSH')
      END IF
      TOPBUF = TOP
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PUSHBR ( ITEM )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is used by subroutine XCHECK.
C    It pushes integer positions into the BSTACK stack.
C
C  Input arguments:
C    ITEM       Is the position of a bracket returned
C               to calling routine XCHECK.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      INTEGER BSTACK(1:LINLEN)
      LOGICAL OK
      COMMON /TSTBST/ TOPBST, BSTACK, OK
C  ---------------------------------------------------------------------
C
      IF (TOPBST.LT.LINLEN) THEN
         TOPBST = TOPBST+1
         BSTACK(TOPBST) = ITEM
         OK = .TRUE.
      ELSE
         OK = .FALSE.
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PUSHDN ( ITEM )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine is called from subroutine XCHECK.
C    It pushes items on the xcheck stack.
C
C  Input arguments:
C    ITEM       The pushed element.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      INTEGER STACK(1:LINLEN)
      LOGICAL SUCCES
      COMMON /TSTXST/ TOPST, STACK, SUCCES
C  ---------------------------------------------------------------------
C
      IF (TOPST.LT.LINLEN) THEN
         TOPST = TOPST+1
         STACK(TOPST) = ITEM
         SUCCES = .TRUE.
      ELSE
         SUCCES = .FALSE.
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QUIT ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the QUIT statement and generates appropriate object code.
C    The syntax is:
C      QUIT ( FLAG = <expression> ; EPILOG = 'yes' | 'no' )
C     Order of the keywords is not important.
C     The object code is:
C       code to evaluate <expression>
C       QUIT                 or          QUIT NOEPILOG
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME, STARG(1)
      INTEGER EX(LINLEN)
      LOGICAL CORRE
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
C  Expect a left parenthesis.
      NT = NT+1
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,'QUIT')
         RETURN
      END IF
C
      IFLA = 0
      IEPI = 0
      LINE = 'QUIT'
      LL = 4
100   CONTINUE
      NT = NT+1
C
C  A keyword is expected.
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         RETURN
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,NAME)
C
C  Is it  FLAG or EPILOG ?
      IF (ISCOMP(NAME,'FLAG').NE.0) THEN
         IF (IFLA.NE.0) THEN
            CALL ERROR(166,NAME)
            RETURN
         ELSE
            IFLA = 1
            IKEY = 1
         END IF
      ELSE IF (ISCOMP(NAME,'EPILOG').NE.0) THEN
         IF (IEPI.NE.0) THEN
            CALL ERROR(166,NAME)
            RETURN
         ELSE
            IEPI = 1
            IKEY = 2
         END IF
      ELSE
         CALL ERROR(33,NAME)
         RETURN
      END IF
C
C  Get the = sign.
      NT = NT+1
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(35,' ')
         RETURN
      END IF
C
C  If its the FLAG keyword, extract the expression.
      IF (IKEY.EQ.1) THEN
         CALL EXTRA(NT,EX,NEX,2,NA)
         CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C  If expression incorrect.
         IF (.NOT.CORRE) RETURN
      ELSE
C
C  Its the EPILOG keyword. Expect a string.
         NT = NT+1
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL ERROR(169,NAME)
            RETURN
         END IF
C
         NT = NT+1
         IF (TOKEN(NT).NE.STRING) THEN
            CALL ERROR(28,' ')
            RETURN
         END IF
         CALL GETIDS(NA,NN,IS,IE)
C
         NT = NT+1
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL ERROR(28,NAME)
            RETURN
         END IF
C
C  The string should be YES or NO.
         IF (ISCOMP(ALPHA(IS:IE),'NO').NE.0) THEN
            LINE = 'QUIT NOEPILOG'
            LL = 13
         ELSE IF (ISCOMP(ALPHA(IS:IE),'YES').EQ.0) THEN
            CALL ERROR(185,NAME)
            RETURN
         END IF
      END IF
C
      NT = NT+1
C
C  Another keyword ?
      CUR = TOKEN(NT)
      IF (CUR.EQ.SEMI) GOTO 100
C
C  End of statement ?
      IF (CUR.NE.RIGHTP) THEN
         CALL ERROR(36,' ')
         RETURN
      END IF
C
C  The quit keyword must be present.
      IF (IFLA.EQ.0) THEN
         CALL ERROR(41,'FLAG')
         RETURN
      END IF
C
C  Generate final object code.
      BLOC = BLOC+1
      IF (GENE) WRITE (SCR,'(A)') LINE(1:LL)
C
      END
C  ---------------------------------------------------------------------
      LOGICAL FUNCTION RESE ( NAME )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether an identifier can be used within a VAR or FUNCTION
C    declaration. The value returned is .TRUE. if the name belongs to
C    the reserved ones; .FALSE. Otherwise.
C
C  Input arguments:
C    NAME       Character variable. This is the name to be checked.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) NAME
C  ---------------------------------------------------------------------
C  Local variables:
C
C  Number of identifiers that may not be used as variables.
      PARAMETER ( NNUI = 5 )
      CHARACTER*4  NUI(NNUI)
      DATA NUI / 'BY  ',  'JUST',  'THEN', 'TO  ','FROM' /
C  ---------------------------------------------------------------------
C
      RESE = .TRUE.
      DO 10,I=1,NNUI
         IF (NAME.EQ.NUI(I)) RETURN
10    CONTINUE
C
      RESE = .FALSE.
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SDELIM  ( STR, LES, NEW, LL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SDELIM accepts an input string (STR) and constructs
C    an output string (NEW). If STR contains any special characters
C    (spaces, tabs, quotes), the resulting string in NEW will be
C    enclosed in quotes. In addition if STR contains any quotes, they
C    will be escaped (\').
C
C  Input arguments:
C    STR        The input string
C    LES        Effective length of STR.
C
C  Output arguments:
C    NEW        The output string.
C    LL         Effective length of NEW.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR, NEW
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C, QUOTE
C  ---------------------------------------------------------------------
C
      QUOTE = ''''
C
      NEW = ' '
      LL = 0
      NSP = 0
      LENACT = LEN(NEW)
      DO 10,I=1,LES
         C = STR(I:I)
         IF (C.EQ.' ') THEN
            IF (LL+1.GT.LENACT) GOTO 30
            NSP = NSP+1
            LL = LL+1
            NEW(LL:LL) = C
         ELSE IF (C.EQ.QUOTE) THEN
            IF (LL+2.GT.LENACT) GOTO 30
            NSP = NSP+1
            NEW(LL+1:LL+2) = ESC//QUOTE
            LL = LL+2
         ELSE IF (C.EQ.TAB) THEN
            IF (LL+1.GT.LENACT) GOTO 30
            NSP = NSP+1
            LL = LL+1
            NEW(LL:LL) = C
         ELSE
            IF (LL+1.GT.LENACT) GOTO 30
            LL = LL+1
            NEW(LL:LL) = C
         END IF
10    CONTINUE
C
C  Do we need to enclose NEW in quotes ?
30    CONTINUE
      IF (NSP.NE.0) THEN
C  Make room for the initial delimiter.
         IF (LL.GE.LENACT) RETURN
         DO 20,I=LL,1,-1
            I1 = I+1
            NEW(I1:I1) = NEW(I:I)
20       CONTINUE
C  Add delimiters.
         LL = LL+2
         NEW(1:1) = QUOTE
         NEW(LL:LL) = QUOTE
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SERLOC ( TABLE, ELEM, NAME, FOUND, POSIT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Locates a name in a character type array. A linear search is
C    applied to the given array.
C
C  Input arguments:
C    TABLE   The character array to be searched. It is assumed
C            to be dimensioned as: TABLE(1:ELEM)
C    ELEM    Number of elements in the above array.
C    NAME    The name whose position in the array we are looking for.
C
C  Output arguments:
C    FOUND   Logical variable indicating, whether the requested name
C            was (FOUND=.TRUE.), or was not found (FOUND=.FALSE.) in
C            the given array.
C    POSIT   If the name was found in the array, this is its position;
C            otherwise 0 is returned.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) TABLE(*), NAME
      LOGICAL FOUND
      INTEGER ELEM, POSIT
C  ---------------------------------------------------------------------
C
      FOUND = .TRUE.
      DO 10,POSIT=1,ELEM
         IF (TABLE(POSIT).EQ.NAME) RETURN
10    CONTINUE
C
C  The name was not found in the given array.
      FOUND = .FALSE.
      POSIT = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE STAFU
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will handle a statement function declaration.
C    The declaration has the form:
C
C    FUNCTION <name> [ ARG1, ARG2, ...ARGn ] = <expression>
C
C    where <name> is the symbolic name of the statement function,
C    ARG1, ARG2, ...ARGn is a list of dummy arguments to the function,
C    and <expression> is the actual expression that this function will
C    calculate. The arguments in the list are pushed in the statement
C    function argument table, in the order they appear in the list.
C    The object code generated by the <expression> is not actually
C    written in the scratch file, but in the statement function file
C    (unit STF) instead.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
C
C  Maximum number of arguments in a statement function.
      PARAMETER ( MXARG =(LINLEN-13)/2 )
      INTEGER EX(LINLEN)
      CHARACTER*(IDLEN) STAFID, VARID, STARG(MXARG)
      LOGICAL CORRE, FOUND, RESE
C
C  Initialize pointers to the TOKEN, ALPHA and NUM arrays respectively.
      NT = 1
      NA = 1
      NN = 2
C
C  An identifier is expected as the function name.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(60,' ')
         RETURN
      END IF
C
C  Get the function name.
      CALL GETIDF(NA,NN,STAFID)
C
C  Check if the function name should not be used.
      IF (RESE(STAFID)) CALL ERROR(12,STAFID)
C
C  Check if the function name is already present in the var table.
      CALL LOCAN(VAR,NVAR,STAFID,FOUND,IND)
      IF (FOUND) THEN
         CALL ERROR(65,STAFID)
         RETURN
      END IF
C
C  A left bracket should follow the function name.
      NT = NT+1
      IF (TOKEN(NT).NE.LBRA) THEN
         CALL ERROR(61,STAFID)
         RETURN
      END IF
C
C  NFARG is the number of arguments in the dummy argument list.
      NFARG = 0
C
C  Start a loop over all function arguments.
100   NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(62,' ')
         RETURN
      END IF
C
C  Get the argument name.
      CALL GETIDF(NA,NN,VARID)
C
C  Check if already present in the argument list.
      CALL SERLOC(STARG,NFARG,VARID,FOUND,II)
      IF (FOUND) THEN
         CALL ERROR(66,VARID)
         RETURN
      END IF
C
C  Push the argument in the table.
      NFARG = NFARG+1
      STARG(NFARG) = VARID
C
      NT = NT+1
      CUR = TOKEN(NT)
C
C  A comma denotes another argument in the list.
      IF (CUR.EQ.COMMA) GOTO 100
C
C  A right bracket denotes the end of the argument list.
      IF (CUR.EQ.RBRA) GOTO 200
C
C  Neither comma, nor right bracket found.
      CALL ERROR(63,' ')
      RETURN
C
C  An = sign should follow the right bracket.
200   NT = NT+1
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(64,' ')
         RETURN
      END IF
C
C  Push initial location of the statement function file in the
C  information table.
      IF (NINF+2.GT.MXINFO)
     &    CALL OVER('information','INFO','STAFU',MXINFO,'MXINFO')
      NINF = NINF+1
      INFO(NINF) = NSTL+1
C
C  Extract the <expression>.
      CALL EXTRA(NT,EX,NEX,0,NA)
      ISTF = STF
      INSTL = NSTL
C
C  Calculate <expression>. The object code will be written
C  in the statement function file. (Unit STF)
      CALL XPR(EX,NEX,NA,NN,CORRE,.TRUE.,INSTL,ISTF,STARG)
C
C  Count number of lines in the statement function file.
      NSTL = NSTL+INSTL
C
C  If <expression> incorrect.
      IF (.NOT.CORRE) RETURN
C
C  Push current (after code for the <expression> has been written)
C  location of the statement function file in the information table.
      NINF = NINF+1
      INFO(NINF) = NSTL
C
C  Insert information about this statement function in the
C  associated tables.
      CALL INSERT(STAFID,NFARG,MEMLOC,NINF-1,3,IND)
C
C  Reserve storage for the arguments of this statement function.
      MEMLOC = MEMLOC+NFARG
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SUBCAL ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SUBCAL parses the CALL statement, and generates
C    appropriate object code.
C    When a subprogram is called two numbers are passed from the stack:
C
C    1) On top of the stack: The current program counter. This is
C                            added at runtime by the $CALLSUB
C                            instruction.
C    2) On top-1           : Memory location, where the list of
C                            arguments for this call begins (INAD).
C                            This is added by appropriate abject
C                            code, written by this subroutine.
C
C    On location INAD, there is a block of NARGS memory locations:
C
C    Location:             Contents:
C    INAD                  Address of actual argument 1  (ADDR1)
C    INAD+1                Address of actual argument 2  (ADDR2)
C    INAD+2                Address of actual argument 3  (ADDR3)
C     ....                            ....
C    INAD+NARGS-1          Address of actual argument NARGS
C
C    ADDR1, ADDR2, ... depend to the argument passed:
C    1) For user variable or (user) array element, ADDRi is the actual
C       location where it is stored.
C    2) For a Merlin array, ADDRi is the address of a temporary block.
C       The Merlin array is copied onto this block, and passed to the
C       routine. When the subprogram returns, the block is deallocated.
C    3) For an expression, ADDRi is a memory address, allocated along
C       with any user variables, on the subprogram containing the CALL.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME, VNAME, STARG(1)
      INTEGER EX(LINLEN)
      LOGICAL FOUND, CORRE
      PARAMETER ( LLSCR = 80 )
      CHARACTER*(LLSCR) LINE
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(52,' ')
         RETURN
      END IF
C
C  Get the subprogram name.
      CALL GETIDF(NA,NN,NAME)
C
C  Calculate number of actual arguments (NARGS).
      NT = NT+1
      IF (NT.EQ.NTOK) THEN
         NARGS = 0
      ELSE
         IF (TOKEN(NT).NE.LEFTP) THEN
            CALL ERROR(52,' ')
            RETURN
         END IF
         NARGS = 1
         DO 10,I=NT,NTOK
10          IF (TOKEN(I).EQ.SEMI) NARGS = NARGS+1
      END IF
C
C  IMEM is the memory location where the list of addresses for
C  this call will reside.
      IMEM = MEMLOC+1
C
C  Allow at least one memory location for the argument list (even if
C  no arguments exist.
      MEMLOC = MEMLOC+MAX(NARGS,1)
      NME = 0
      NV = 0
      IF (NARGS.EQ.0) GOTO 1000
C
C  Loop over all arguments.
100   CONTINUE
      NV = NV+1
C
C  Determine what the next argument is:
C  IWHAT = 0 -> Some error.
C          1 -> Just a name.
C          2 -> A name followed by left bracket.
C          3 -> Expression.
C
C  Extract the argument as an expression.
      CALL EXTRA(NT,EX,NEX,2,NA)
      IF (NEX.EQ.0) THEN
         CALL I2STR(NV,STR,LES)
         CALL ERROR(70,STR(1:LES))
C  Error.
         IWHAT = 0
      ELSE IF (NEX.EQ.1) THEN
         IF (EX(1).EQ.IDENT) THEN
C  A variable.
            IWHAT = 1
         ELSE
C  An expression.
            IWHAT = 3
         END IF
      ELSE
C  A name followed by a bracket.
         IF (EX(1).EQ.IDENT .AND. EX(2).EQ.LBRA) THEN
            NBR = 1
            DO 30,IEX=3,NEX
               IF (EX(IEX).EQ.LBRA) NBR = NBR+1
               IF (EX(IEX).EQ.RBRA) NBR = NBR-1
               IF (NBR.EQ.0) THEN
                  IF (IEX.EQ.NEX) THEN
C  A name, followed by something in brackets.
                     IWHAT = 2
                  ELSE
C  An expression.
                     IWHAT = 3
                  END IF
                  GOTO 31
               END IF
30          CONTINUE
C  An expression.
            IWHAT = 3
31          CONTINUE
         ELSE
C  An expression.
            IWHAT = 3
         END IF
      END IF
C
C  Process the argument according to IWHAT.
C
C  Just a name.
      IF (IWHAT.EQ.1) THEN
C
C  Get the name and locate it in the var table.
         CALL GETIDF(NA,NN,VNAME)
         CALL LOCAN(VAR,NVAR,VNAME,FOUND,IND)
         IF (.NOT.FOUND) THEN
            CALL ERROR(19,VNAME)
            CALL INSERT(VNAME,0,0,0,-1,IND)
         END IF
C
C  Generate code according to its KIND.
         K = KIND(IND)
         IF (K.EQ.1) THEN
C  User variable (local).
            BLOC = BLOC+2
            IF (GENE) THEN
               WRITE (SCR,'(A)') '$PUSHA'
               WRITE (SCR,*) LOC(IND)
            END IF
            CALL WPOP(SCR,BLOC,IMEM+NV-1)
         ELSE IF (K.EQ.101) THEN
C  User variable (argument).
            CALL WPSC(SCR,BLOC,0)
            CALL WPSC(SCR,BLOC,LOC(IND))
            BLOC = BLOC+1
            IF (GENE) WRITE (SCR,'(A)') '$IOFFSET'
            CALL WPOP(SCR,BLOC,IMEM+NV-1)
         ELSE IF (K.EQ.2) THEN
C  User array (local).
            BLOC = BLOC+2
            IF (GENE) THEN
               WRITE (SCR,'(A)') '$PUSHA'
               WRITE (SCR,*) LOC(IND)
            END IF
            CALL WPOP(SCR,BLOC,IMEM+NV-1)
         ELSE IF (K.EQ.102) THEN
C  User array (argument).
            CALL WPSC(SCR,BLOC,0)
            CALL WPSC(SCR,BLOC,LOC(IND))
            BLOC = BLOC+1
            IF (GENE) WRITE (SCR,'(A)') '$IOFFSET'
            CALL WPOP(SCR,BLOC,IMEM+NV-1)
         ELSE IF (K.EQ.3) THEN
C  Statement function. Process it as an expression.
            NA = NA-1
            NN = NN-2
            IWHAT = 3
         ELSE IF (K.EQ.4) THEN
C  Merlin variable. Process it as an expression.
            NA = NA-1
            NN = NN-2
            IWHAT = 3
         ELSE IF (K.EQ.5) THEN
C  Merlin array. Pass a copy of the array. (Array by value !!!)
            NME = NME+1
            BLOC = BLOC+5
            IF (GENE) THEN
               WRITE (SCR,'(A)') '$DIM'
               WRITE (SCR,'(A)') '$ALLOC'
               CALL WPSC(SCR,BLOC,1)
               WRITE (SCR,'(A)') '$+'
               CALL WPOP(SCR,BLOC,IMEM+NV-1)
               CALL WPS(SCR,BLOC,IMEM+NV-1)
               WRITE (SCR,'(A)') '$BLOCKCOPY'
               WRITE (SCR,'(A)') VNAME
            END IF
         ELSE IF (K.EQ.6) THEN
C  Intrinsic function (ABS, COS, etc.). Process it as an expression.
            NA = NA-1
            NN = NN-2
            IWHAT = 3
         ELSE IF (K.NE.-1) THEN
C  An undeclared variable. Process it as an expression.
            NA = NA-1
            NN = NN-2
            IWHAT = 3
         END IF
      END IF
C
C  Name followed by a left bracket.
      IF (IWHAT.EQ.2) THEN
C
C  Get the name and locate it in the var table.
         CALL GETIDA(NA+1,VNAME)
         CALL LOCAN(VAR,NVAR,VNAME,FOUND,IND)
         IF (.NOT.FOUND) THEN
            CALL ERROR(19,VNAME)
            CALL INSERT(VNAME,0,0,0,-1,IND)
         END IF
         K = KIND(IND)
         IF (K.EQ.2 .OR. K.EQ. 102) THEN
            NLS = 0
            OPEN (SPARE,STATUS='SCRATCH',IOSTAT=IERR)
            IF (IERR.NE.0)
     &            CALL MABORT('Cannot open the spare file.','SUBCAL')
            REWIND SPARE
C  Extract the expression for the array index.
            CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,NLS,SPARE,STARG)
            IF (GENE) THEN
C  Copy the spare file to the scratch file. The final $GETCONT
C  instruction will not be written in the scratch file.
               REWIND SPARE
               DO 88,I=1,NLS-1
                  READ (SPARE,'(A)') LINE
                  WRITE (SCR,'(A)') LINE
88             CONTINUE
               BLOC = BLOC+NLS-1
            END IF
            CLOSE (SPARE,STATUS='DELETE',IOSTAT=IERR)
            IF (IERR.NE.0)
     &          CALL MABORT('Cannot open the spare file.','SUBCAL')
            CALL WPOP(SCR,BLOC,IMEM+NV-1)
         ELSE IF (K.NE.-1) THEN
C  Skip undeclared variables. A single statement function call
C  will be processed as an expression.
            IWHAT = 3
         END IF
      END IF
C
C  An expression.
      IF (IWHAT.EQ.3) THEN
         CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
         BLOC = BLOC+2
         MEMLOC = MEMLOC+1
         CALL WPOP(SCR,BLOC,MEMLOC)
         IF (GENE) THEN
            WRITE (SCR,'(A)') '$PUSHA'
            WRITE (SCR,*) MEMLOC
         END IF
         CALL WPOP(SCR,BLOC,IMEM+NV-1)
      END IF
C
C  Another argument ?
      NT = NT+1
      IF (TOKEN(NT).EQ.SEMI) GOTO 100
C
C  A right parenthesis terminates the statement.
      IF (TOKEN(NT).EQ.RIGHTP) THEN
         NT = NT+1
         IF (NT.NE.NTOK) CALL ERROR(9,' ')
      ELSE
         CALL ERROR(69,' ')
      END IF
C
C  Write final code for the call.
1000  CONTINUE
      BLOC = BLOC+4
      IF (GENE) THEN
         WRITE (SCR,'(A)') '$PUSHA'
         WRITE (SCR,*) IMEM
         WRITE (SCR,'(A)') '$CALLSUB'
         WRITE (SCR,*)
         NCAL = NCAL+1
         IF (NCAL.GT.MXCAL)
     &          CALL OVER('sub','CAL','SUBCAL',MXCAL,'MXCAL')
         CAL(NCAL) = NAME
         CALDEF(NCAL) = BLOC
         NARCAL(NCAL) = NARGS
      END IF
      BLOC = BLOC+2*NME
      IF (GENE) THEN
         DO 60,I=1,NME
            WRITE (SCR,'(A)') '$DIM'
            WRITE (SCR,'(A)') '$DEALLOC'
60       CONTINUE
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SUBR
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses a SUB declaration
C    A SUB declaration has the form:
C      SUB NAME [ ( arg1; arg2; ...; argN ) ]
C    where NAME is the name of the subprogram and arg1, arg2, ... are
C    as in the VAR declaration. Adjustable upper bounds for the last
C    dimension of an array, are allowed.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME
      LOGICAL FOUND
C  ---------------------------------------------------------------------
C
C  Initialize pointers to the TOKEN, ALPHA and NUM arrays respectively.
      NT = 1
      NA = 1
      NN = 2
C
C  Look for the name of the subprogram.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(52,' ')
         RETURN
      END IF
C
C  Get the name of the subprogram.
      CALL GETIDF(NA,NN,NAME)
C
C  Is it its first occurrence ?
      CALL SERLOC(EXT,NEXT,NAME,FOUND,POSIT)
      IF (FOUND) THEN
C  A multiply defined subprogram.
         CALL ERROR(30,NAME)
      ELSE
C  Enter the name in the externals table.
         NEXT = NEXT+1
         IF (NEXT.GT.MXEXT)
     &          CALL OVER('externals','EXT','SUBR',MXEXT,'MXEXT')
         EXT(NEXT) = NAME
         EXTDEF(NEXT) = BLOC+1
      END IF
C
C  Check for arguments.
      NT = NT+1
      IF (NT.EQ.NTOK) GOTO 1000
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(53,' ')
         RETURN
      END IF
C
C  Parse the argument definitions.
      CALL VARIA(1,NT,NA,NN,NVDECL)
      NARDEF(NEXT) = NVDECL
      NT = NT+1
      IF (TOKEN(NT).NE.RIGHTP) THEN
         CALL ERROR(69,NAME)
         RETURN
      END IF
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL ERROR(9,'SUB')
         RETURN
      END IF
C
C  Generate code for the subprogram.
1000  CONTINUE
      BLOC = BLOC+7
      IF (GENE) THEN
         WRITE (SCR,10) '$STACKSWAP'
         WRITE (SCR,10) '$PUSHINAD'
         WRITE (SCR,10) '$STACKSWAP'
         WRITE (SCR,10) '$POPINAD'
         WRITE (SCR,10) '$PUSHMB'
         WRITE (SCR,10) '$PUSHC'
         WRITE (SCR,*)
      END IF
C
C  Update the type and type-definition tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','SUBR',MXTYP,'MXTYP')
      TYPE(NTYPE) = 9
      TDEF(NTYPE) = BLOC
      BLOC = BLOC+2
      IF (GENE) THEN
         WRITE (SCR,10) '$ALLOC'
         WRITE (SCR,10) '$POPMB'
      END IF
C
10    FORMAT (A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SUBRET ( NT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the SUBRETURN statement. The object code generated is:
C      $MOVE               --->  Jump to the END of the subprogram.
C      <blank line>
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C
C  Nothing else should follow SUBRETURN.
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL ERROR(9,'SUBRETURN')
         RETURN
      END IF
C
C  You cant use SUBRETURN in a main program.
      IF (INWHAT.NE.2) THEN
         CALL ERROR(181,' ')
         RETURN
      END IF
C
C  SUBRETURN just performs a jump to the end of the subprogram.
C  Remember: we have to do some housekeeping before returning (memory
C  deallocation, etc.)
C  ".." is a fake label, defined by the END statement of the subprogram.
      NMOVE = NMOVE+1
      IF (NMOVE.GT.MXMOV)
     &    CALL OVER('move','MVELAB','SUBRET',MXMOV,'MXMOV')
      MVELAB(NMOVE) = '..'
C
C  Generate code.
      BLOC = BLOC+2
      IF (GENE) WRITE (SCR,'(A/)')'$MOVE'
C
C  Update the type and type-definition tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','SUBRET',MXTYP,'MXTYP')
      TYPE(NTYPE) = -7
      TDEF(NTYPE) = BLOC
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SUBX ( SIMPX, TOKN, CORR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine performs checking of subexpressions (or simple
C    expressions). A simple expression is defined as:
C
C         <simpex>  ::=  <operand> [<operator><operand>]
C
C    Clearly corresponds to the most inner part of an expression into
C    parentheses and brackets.
C
C  Input arguments:
C    SIMPX      Simple expression to be checked.
C    TOKN       Number of tokens in SIMPEX.
C
C  Output arguments:
C    CORR       CORR = .TRUE.  -> The subexpression is correct.
C               CORR = .FALSE. -> The subexpression is incorrect.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER SIMPX(*)
      LOGICAL CORR
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( ONE = 1000, ZERO = -1000 )
      LOGICAL OPERAN, OPRTOR
      OPERAN(TEMPO) = (TEMPO.EQ.IDENT) .OR. (TEMPO.EQ.UNSI)   .OR.
     &                (TEMPO.EQ.ZERO)  .OR. (TEMPO.EQ.ONE)
      OPRTOR(TEMPO) =(TEMPO.EQ.PLUS)   .OR. (TEMPO.EQ.MINUS)  .OR.
     &               (TEMPO.EQ.STAR)   .OR. (TEMPO.EQ.BISTAR) .OR.
     &               (TEMPO.EQ.SLASH)  .OR. (TEMPO.EQ.AND)    .OR.
     &               (TEMPO.EQ.OR)     .OR. (TEMPO.EQ.LT)     .OR.
     &               (TEMPO.EQ.LE)     .OR. (TEMPO.EQ.GT)     .OR.
     &               (TEMPO.EQ.GE)     .OR. (TEMPO.EQ.EQUAL)  .OR.
     &               (TEMPO.EQ.NE)     .OR. (TEMPO.EQ.XOR)    .OR.
     &               (TEMPO.EQ.NOT)
C  ---------------------------------------------------------------------
C
      IF (.NOT. OPERAN(SIMPX(1))) THEN
         CORR = .FALSE.
         CALL ERROR(43,' ')
         RETURN
      END IF
C
      IF (MOD(TOKN,2).EQ.0) THEN
         CORR = .FALSE.
         CALL ERROR(44,' ')
         RETURN
      END IF
C
      DO 1000,I=2,TOKN,2
         IF (.NOT. (OPRTOR(SIMPX(I))
     &         .AND. (SIMPX(I+1).EQ.IDENT .OR. SIMPX(I+1).EQ.UNSI .OR.
     &                SIMPX(I+1).EQ.ONE .OR. SIMPX(I+1).EQ.ZERO))) THEN
            CORR = .FALSE.
            CALL ERROR (45,' ')
            RETURN
         END IF
1000  CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TARGET ( NT, NA, NN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the target statement. Syntax:
C      TARGET ( VALUE = <expression> )
C    Object code:
C      code to evaluate <expression>
C      TARGET
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) NAME, STARG(1)
      INTEGER EX(LINLEN)
      LOGICAL CORRE
C  ---------------------------------------------------------------------
C
C  A left parenthesis should follow.
      NT = NT+1
      IF (TOKEN(NT).NE.LEFTP) THEN
         CALL ERROR(31,'TARGET')
         RETURN
      END IF
C
100   CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(32,' ')
         RETURN
      END IF
C
C  Get the keyword.
      CALL GETIDF(NA,NN,NAME)
C
C  Is it VALUE ?
      IF (ISCOMP(NAME,'VALUE').EQ.0) THEN
         CALL ERROR(33,NAME)
         RETURN
      END IF
C
C  Look for an = sign.
      NT = NT+1
      IF (TOKEN(NT).NE.EQ) THEN
         CALL ERROR(35,' ')
         RETURN
      END IF
C
C  Extract and evaluate the <expression>.
      CALL EXTRA(NT,EX,NEX,2,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If expression is incorrect.
      IF (.NOT.CORRE) RETURN
C
C  A right parenthesis terminates the statement.
      NT = NT+1
      IF (TOKEN(NT).NE.RIGHTP) THEN
         CALL ERROR(36,'TARGET')
         RETURN
      END IF
C
C  Anything left ?
      NT = NT+1
      IF (NT.NE.NTOK) CALL ERROR(9,'TARGET')
      BLOC = BLOC+1
      IF (GENE) WRITE (SCR,'(A)') 'TARGET'
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TERMIN
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will check for any block ifs, or loops that do not
C    terminate. (ENDIF or ENDLOOP absent)
C    An ENDIF statement causes the entries for the corresponding IF
C    and ELSE statement in the type table, to become negative
C    (-1 or -2). An ENDLOOP statement causes the entry for the
C    corresponding LOOP in the type table to become also negative
C    (-4).
C    Thus, when the first pass over the MCL source has been completed,
C    a value of 1 or 2, indicates a block if that does not terminate,
C    while a value of 4 indicates a loop that does not terminate.
C    The type table is searched backwards and if a value of 1 or 2 is
C    found, a false ENDIF is pushed in the type table. The above two
C    values become negative. If a 4 is reached, a false ENDLOOP is
C    pushed in the type table, and the previous value becomes also
C    negative. In both cases the location pushed in the type-definition
C    table, is the current location counter of the scratch file. Since
C    the first pass has been completed this is the last line of object
C    code written. (Note that no more object code will be written in the
C    scratch file for the current subprogram.) False ENDIF and ENDLOOP
C    statements are effectively "appended" at the end of the current
C    subprogram.
C    If any non-terminating block ifs or loops are found, appropriate
C    error messages are issued.
C
C  Abort conditions that may occur:
C    If the type table overflows, this will cause the program to
C    stop and issue an informative message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL NOLOOP, NOEND
C  ---------------------------------------------------------------------
C
C  NOLOOP indicates whether a non-terminating loop was found.
C  NOEND  indicates whether a non-terminating block if was found.
      NOLOOP = .FALSE.
      NOEND  = .FALSE.
      NTP    =  NTYPE
C
C  Search the type table backwards.
      DO 10,I=NTP,TSTART,-1
         TP = TYPE(I)
         IF (TP.EQ.1 .OR. TP.EQ.2) THEN
C  A block if that does not terminate was just found.
            NOEND = .TRUE.
            BLOC  = BLOC+1
C  Push a false ENDIF in the type table.
            NTYPE = NTYPE+1
            IF (NTYPE.GT.MXTYP)
     &              CALL OVER('type','TYPE','TERMIN',MXTYP,'MXTYP')
            TYPE(NTYPE) = -3
            TDEF(NTYPE) = BLOC
            TYPE(I)     = -TP
         ELSE IF (TP.EQ.4) THEN
C  A loop that does not terminate has just been found.
            NOLOOP = .TRUE.
            BLOC   = BLOC+1
C  Push a false ENDLOOP in the type table.
            NTYPE  = NTYPE+1
            IF (NTYPE.GT.MXTYP)
     &              CALL OVER('type','TYPE','TERMIN',MXTYP,'MXTYP')
            TYPE(NTYPE) = -5
            TDEF(NTYPE) = BLOC
            TYPE(I)     = -TP
         END IF
10    CONTINUE
C
C  If anything went wrong, issue informative messages.
      IF (NOLOOP) CALL ERROR(146,' ')
      IF (NOEND ) CALL ERROR(147,' ')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VALIJ
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will check the validity of the MOVE TO statements
C    in the program, in relation to block ifs and loops.
C    A block is defined to be a sequence of statements residing in
C    between:
C      IF ... ENDIF,
C      IF ... ELSE,
C      ELSE ... ENDIF and
C      LOOP ... ENDLOOP.
C    A jump inside a block is not allowed, unless the corresponding
C    MOVE TO statement, belongs to the block.
C    Jumps outside a block, are always allowed.
C    When this routine is entered, it is assumed that block ifs
C    and loops are properly nested and properly terminated.
C    It is also assumed that all MOVE TO statements have been processed.
C    (Linked with the corresponding labels) Thus only negative entries
C    occur in the type table.
C    The routine will start searching the type table. For each block
C    encountered it will assign a value of 7 for all MOVE TO statements
C    that belong to the block as well as their target label.
C    Then it will check the validity of the rest of the MOVE TO
C    statements, belonging to the block. Upon exit of the routine
C    all entries in the type table that indicate an IF, ELSE,
C    ENDIF, LOOP, ENDLOOP or MOVE TO statement will yield
C    positive values: 1, 2, 3, 4, 5 and 7 respectively.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  ---------------------------------------------------------------------
C  Local variables:
      INTEGER ELSE, END, ENDLOO
C
C  Start searching the type table.
      DO 900,I=TSTART,NTYPE
         IF (TYPE(I).EQ.-3) THEN
C  An ENDIF was found.
            END = I
C  Locate the corresponding IF and ELSE in the type table.
C  0 will be returned in ELSE if no ELSE is defined
C  for this block if.
            CALL LOCIF(IF1,ELSE,END)
C  Indicate that these IF and ENDIF statements have been processed.
            TYPE(IF1) = 1
            TYPE(END) = 3
            IF (ELSE.EQ.0) THEN
C  This is a block if with no ELSE part.
C  Jumps from this block, to this block are allowed.
               CALL ZEROM (IF1,END)
C  Jumps into the block caused by MOVE TO statements before the
C  block if are not allowed.
               CALL JUMPIN(1,IF1-1,IF1,END)
C  Jumps into the block caused by MOVE TO statements after the
C  block if are not allowed.
               CALL JUMPIN(IF1+1,NTYPE,IF1,END)
            ELSE
C  This is a block if with an ELSE part.
C  Indicate that this ELSE has been processed.
               TYPE(ELSE) = 2
C  Jumps from the IF...ELSE block into the same block, are allowed.
               CALL ZEROM (IF1,ELSE)
C  Jumps from the ELSE...ENDIF block into the same block, are allowed.
               CALL ZEROM (ELSE,END)
C  Jumps into the IF...ENDIF block caused by MOVE TO statements
C  before the block if are not allowed.
               CALL JUMPIN(1,IF1-1,IF1,END)
C  Jumps into the IF...ENDIF block caused by MOVE TO statements
C  after the block if are not allowed.
               CALL JUMPIN(END+1,NTYPE,IF1,END)
C  Jumps into the ELSE...ENDIF block caused by MOVE TO statements
C  residing in the IF...ELSE block are not allowed.
               CALL JUMPIN(IF1+1,ELSE-1,ELSE,END)
C  Jumps into the IF...ELSE block caused by MOVE TO statements
C  residing in the ELSE...ENDIF block are not allowed.
               CALL JUMPIN(ELSE+1,END-1,IF1,ELSE)
            END IF
         ELSE IF (TYPE(I).EQ.-5) THEN
C  An ENDLOOP was found.
            ENDLOO = I
C  Locate the corresponding LOOP in the type table.
            CALL LOCLOO(LOO,ENDLOO)
C  Jumps into the loop by MOVE TO statements residing in the loop
C  are allowed.
            CALL ZEROM (LOO,ENDLOO)
C  Jumps into the LOOP...ENDLOOP block caused by MOVE TO statements
C  residing before the block are not allowed.
            CALL JUMPIN(1,LOO-1,LOO,ENDLOO)
C  Jumps into the LOOP...ENDLOOP block caused by MOVE TO statements
C  residing after the block are not allowed.
            CALL JUMPIN(ENDLOO+1,NTYPE,LOO,ENDLOO)
C  Indicate that the LOOP and ENDLOOP have been processed.
            TYPE(LOO)    = 4
            TYPE(ENDLOO) = 5
         END IF
900   CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE VARIA ( IOPT, NT, NA, NN, NVDECL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes the VAR declaration statement. Reserved names, as well
C    as names already present in the var table, should not be declared.
C    The routine checks for syntactic validity and updates the related
C    tables. These are:
C      The var table,
C      the memory location table,
C      the dimensionality table,
C      the information table,
C      the position table and
C      the kind table.
C
C  Input arguments:
C    IOPT       IOPT = 0 -> We are parsing a VAR declaration.
C               IOPT = 1 -> We are parsing the dummy argument list of a
C                           SUB statement. * is allowed as the upper
C                           bound of the last dimension of an array.
C
C  Output arguments:
C    NVDECL     Number of variables declared in the current VAR
C               statement.
C
C  Input / Output arguments:
C    NT         Number of token-table entries that have been used by the
C               previously called parsing routines.
C    NN         Number of number-table entries that have been used by
C               the previously called parsing routines.
C    NA         Number of alpha-table entries (strings/identifiers),
C               that have been used by the previously called parsing
C               routines.
C
C  Notes:
C    If a syntactic error is encountered while a simple variable is
C    processed, the routine will continue parsing.
C    If an error is encountered while an array is processed, the
C    routine will return, since it has no means to determine in which
C    point of the input line, the next variable declaration starts.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*(IDLEN) VARID
      LOGICAL FOUND, ERR, ARR, RESE
C  ---------------------------------------------------------------------
C
C  NVDECL is the number of variables declared in the line that
C  is currently being parsed.
      NVDECL = 0
C
C  Start a loop for all variables declared in this line.
C  The next token encountered must be an identifier.
40    NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(3,' ')
         RETURN
      END IF
C
C  Get the name of the variable.
      NVDECL = NVDECL+1
      CALL GETIDF(NA,NN,VARID)
C
C  Check if it should not be used. (If it's reserved)
      IF (RESE(VARID)) CALL ERROR(12,VARID)
C
C  ININF is the next available position in the information table.
      ININF = NINF+1
C
C  Locate the variable in the var table.
      CALL LOCAN(VAR,NVAR,VARID,FOUND,IND)
C
C  Check if this is an array declaration. If yes its dimensionality
C  is returned.
      CALL CHARR(NT,NDIM,ERR,ARR,NN,IOPT,VARID)
C
      IF (FOUND) THEN
C  If the name is already in the table, this is an error.
         CALL ERROR(4,VARID)
C  However, if its a simple variable, don't return, but check the
C  rest of the line as well.
         IF (ARR .AND. ERR) RETURN
      ELSE
C
C  The name was not found in the table.
         IF (ARR) THEN
C  If its an array, calculate total memory space occupied.
            IF (ERR) RETURN
            IF (IOPT.EQ.0) SPACE = MEMSPC(ININF,NDIM)
            K = 2
         ELSE
            NDIM = 0
            SPACE = 1
            ININF = 0
            K = 1
         END IF
         IF (IOPT.NE.0) K = K+100
         IF (IOPT.EQ.0) THEN
            MEM = MEMLOC+1
         ELSE
            MEM = NVDECL
         END IF
C
C  Insert information in the related tables.
C  For this variable (or array) :
C  NDIM is its dimensionality,
C  MEM is the memory location to be assigned to it,
C  ININF is the starting position in the information table,
C  K is its kind and
C  IND is the position where all this information should be inserted.
         CALL INSERT(VARID,NDIM,MEM,ININF,K,IND)
C  Update the memory counter.
         IF (IOPT.EQ.0) MEMLOC = MEMLOC+SPACE
      END IF
C
C  Check whether another variable is declared. If not, just return.
      NT = NT+1
      IF (NT.EQ.NTOK) THEN
         IF (IOPT.NE.0) NT = NT-1
         RETURN
      END IF
      IF (TOKEN(NT).EQ.SEMI) GOTO 40
C  Neither semicolon (denoting another variable), nor EOL was found.
      IF (IOPT.NE.0) THEN
         NT = NT-1
         RETURN
      END IF
      CALL ERROR(5,' ')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WARN ( CODE, MESS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Handles warning messages. When a non fatal error is detected,
C    this routine is called to issue the appropriate message.
C    Subroutine WARN will set ISERR = 0 and call subroutine ERROR.
C    ISERR in common block /MCLDTA/ determines whether the error is
C    a fatal one, or just a warning.
C
C  Input arguments:
C    CODE       Integer variable. This is the error code number to be
C               issued.
C    MESS       Character variable. This is usually the name of a
C               variable, MCL statement, etc. that will be substituted
C               in place of the % sign in the error message.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) MESS
C  ---------------------------------------------------------------------
C  Common blocks:
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
C  ---------------------------------------------------------------------
C
      ISERR = 0
      CALL ERROR(CODE,MESS)
      ISERR = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE WHEN
C  ---------------------------------------------------------------------
C
C  Description:
C    Processes the WHEN statement. Its syntax is:
C    WHEN <MCL expression> JUST <ncnlcs>
C    where <MCL expression> is any valid MCL expression and
C    <ncnlcs> is any non conditional and non loop control statement.
C    The object code generated is:
C
C    Code to evaluate the <MCL expression>
C    $NOT
C    $TESTMOVE
C    <blank line>
C
C    If the syntax up to the JUST keyword is correct, a value of
C    -6 is pushed to the type table, to indicate the WHEN statement.
C    Subroutine EXE is called then to process the <ncnlcs> statement.
C    Upon return of the above routine, the number of object code lines,
C    that it has written, is pushed in the type-link table, in the
C    position corresponding to this WHEN.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
C
      LOGICAL CORRE
      INTEGER EX(LINLEN)
      CHARACTER*(IDLEN) STARG(1), IDF
C  ---------------------------------------------------------------------
C
C  Initialize pointers to the TOKEN, ALPHA and NUM arrays respectively.
      NT = 1
      NA = 1
      NN = 2
C
C  Extract and evaluate the <MCL expression>.
      CALL EXTRA(NT,EX,NEX,1,NA)
      CALL XPR(EX,NEX,NA,NN,CORRE,.FALSE.,BLOC,SCR,STARG)
C
C  If the expression is incorrect.
      IF (.NOT.CORRE) RETURN
C
C  Write code to skip the <ncnlcs> statement, in case the
C  expression yields a zero value. (Which means that the given
C  condition is not satisfied.)
      BLOC = BLOC+3
      IF (GENE) WRITE (SCR,'(A/A/)') '$NOT','$TESTMOVE'
C
C  A JUST should follow the <MCL expression>.
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL ERROR(6,' ')
         RETURN
      ELSE
         CALL GETIDF(NA,NN,IDF)
         IF (IDF.NE.'JUST') THEN
            CALL ERROR(21,IDF)
            RETURN
         END IF
      END IF
C
C  Update the type and type-definition tables.
      NTYPE = NTYPE+1
      IF (NTYPE.GT.MXTYP)
     &    CALL OVER('type','TYPE','WHEN',MXTYP,'MXTYP')
      TYPE(NTYPE) = -6
      TDEF(NTYPE) = BLOC
C
C  Save values, for later use.
      INILOC = BLOC
      OLDNT = NTYPE
C
C  Process the <ncnlcs> statement.
      CALL EXE(NT,NA,NN)
C
C  Calculate entry in the type-link table, for this WHEN.
      TREL(OLDNT) = BLOC-INILOC+1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE XCHECK ( EXPR, TOKNUM, CORR, NA, FDECLR, STVAR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine checks the validity of any MCL expression using the
C    following method:
C      1. The input expression, residing on array EXPR, is enclosed in
C         parentheses (this is done so as to use SUBX in any way)
C      2. A scanning is performed on EXPR and when a left parenthesis
C         is located, its position is pushed onto a stack.
C         If a left bracket is located its position is pushed in a
C         second stack (referenced here on as bstack).
C      3. When a right parenthesis is met the position of the
C         parenthesis most recently located, is poped and this position
C         is transformed so to become ident.
C         If a right bracket is met the position of the most recently
C         located bracket is poped from the bstack and ignore
C         transformation is performed.
C      4. All items from left-parenthesis'-position + 1 up to and
C         position of right par., are set ignore and scanning is
C         continued from the last position before "ignorization" where
C         subx was called.
C         All items from left-bracket-position up to and position of
C         right bracket are set ignore.
C
C    Ignore-transformation is a trick which is made so to skip all
C    correct constructs from inside to outside and determine the
C    illegal situations if any.
C
C  Input arguments:
C    EXPR       Integer array containing the tokens of the expression
C               to be processed.
C    TOKNUM     Number of tokens in EXPR.
C
C  Output arguments:
C    CORR       CORR = .TRUE.  -> The expression is correct.
C               CORR = .FALSE. -> The expression is incorrect.
C    FDECLR     FDECLR = .TRUE.  -> We are parsing a function
C                                   declaration.
C               FDECLR = .FALSE. -> We are parsing a normal expression.
C    STVAR      The statement function argument table.
C
C  Input / Output arguments:
C    NA         Pointer to the last used element of the alpha-table.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      LOGICAL CORR, FDECLR
      CHARACTER*(*)  STVAR(*)
      INTEGER EXPR(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      INTEGER STACK(1:LINLEN), BSTACK(1:LINLEN)
      COMMON /TSTXST/ TOPST, STACK, SUCCES
      COMMON /TSTBST/ TOPBST, BSTACK, OK
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( IGNORE = -283, ZERO = -1000, ONE = 1000 )
      PARAMETER ( DNTCAR = -1, VARIA = 1, ARR = 2, STFUN = 3 )
      PARAMETER ( INTRVR = 4, INTRFN = 6 )
      LOGICAL SUBCOR, SUCCES, OK, RTBRA, LFTBRA, ISCOMA, FOUND
      LOGICAL OPERAN, OPRTOR, LFTPAR, RTPAR, PFLAG, BFLAG, INTARR
      DIMENSION WORKEX(LINLEN+2), SUBEX(1:LINLEN)
      CHARACTER*(IDLEN) X, STRI
      INTARR (X) = X.EQ.'R'    .OR. X.EQ.'L'    .OR. X.EQ.'FIX'  .OR.
     &             X.EQ.'GRAD' .OR. X.EQ.'STEP' .OR. X.EQ.'MARG' .OR.
     &             X.EQ.'X'
      OPERAN() = (TEMPO.EQ.IDENT) .OR. (TEMPO.EQ.UNSI) .OR.
     &           (TEMPO.EQ.ZERO)  .OR. (TEMPO.EQ.ONE)
      OPRTOR() = (TEMPO.EQ.PLUS)  .OR. (TEMPO.EQ.MINUS)  .OR.
     &           (TEMPO.EQ.STAR)  .OR. (TEMPO.EQ.BISTAR) .OR.
     &           (TEMPO.EQ.SLASH) .OR. (TEMPO.EQ.AND)    .OR.
     &           (TEMPO.EQ.OR)    .OR. (TEMPO.EQ.LT)     .OR.
     &           (TEMPO.EQ.LE)    .OR. (TEMPO.EQ.GT)     .OR.
     &           (TEMPO.EQ.GE)    .OR. (TEMPO.EQ.EQUAL)  .OR.
     &           (TEMPO.EQ.NE)    .OR. (TEMPO.EQ.XOR)    .OR.
     &           (TEMPO.EQ.NOT)
      LFTPAR() = (TEMPO.EQ.LEFTP)
      RTPAR()  = (TEMPO.EQ.RIGHTP)
      LFTBRA() = (TEMPO.EQ.LBRA)
      RTBRA()  = (TEMPO.EQ.RBRA)
      ISCOMA()  = (TEMPO.EQ.COMMA)
C  ---------------------------------------------------------------------
C
      WORKEX(1) = LEFTP
      WORKEX(TOKNUM+2) = RIGHTP
C
      DO 200,K=2,TOKNUM+1
200      WORKEX(K) = EXPR(K-1)
C
      FINGER = 0
      TOPST = 0
      TOPBST = 0
      CORR = .TRUE.
      NNA = NA
C
1000  CONTINUE
C
      FINGER = FINGER+1
      PFLAG = .FALSE.
      BFLAG = .FALSE.
C
      IF (FINGER.GT.TOKNUM+2) THEN
         CALL POPUP(DUMMY)
         CALL POPBR(DUMMY)
         IF (SUCCES .OR. OK) THEN
            CORR = .FALSE.
            CALL ERROR(49,' ')
         END IF
         RETURN
      END IF
C
      TEMPO = WORKEX(FINGER)
C
      IF (LFTPAR()) THEN
         CALL PUSHDN(FINGER)
         GOTO 1000
C
      ELSE IF (LFTBRA()) THEN
C
C  Check the case of <oprtor> [<expr>] which is wrong.
C
         IF (WORKEX(FINGER-1).NE.IDENT) THEN
            CORR = .FALSE.
            CALL ERROR(50,' ')
            RETURN
         END IF
         CALL PUSHBR(FINGER)
         GOTO 1000
C
      ELSE IF (OPRTOR()) THEN
         GOTO 1000
C
      ELSE IF (OPERAN()) THEN
         IF (TEMPO.EQ.IDENT) THEN
            NNA = NNA+1
            CALL GETIDA(NNA,STRI)
            FOUND = .FALSE.
            IF (FDECLR) CALL SERLOC(STVAR,NFARG,STRI,FOUND,POS)
            IF (.NOT.FOUND) THEN
               CALL LOCAN(VAR,NVAR,STRI,FOUND,POS)
               IF (FOUND) THEN
                  KND = KIND(POS)
               ELSE
                  CALL INSERT(STRI,0,0,0,-1,POS)
                  CORR = .FALSE.
C
C  Sign error and announce the undeclared variable.
C
                  CALL ERROR(19,STRI)
                  RETURN
               END IF
            ELSE
               KND = VARIA
            END IF
C
C  Perform some special checks.
C
            IF (KND.EQ.ARR .AND. WORKEX(FINGER+1).NE.LBRA) THEN
               CALL ERROR(162,' ')
               CORR = .FALSE.
               RETURN
            ELSE IF (KND.EQ.STFUN .OR. KND.EQ.INTRFN) THEN
               IF (WORKEX(FINGER+1).NE.LBRA) THEN
                  CALL ERROR(161,' ')
                  CORR = .FALSE.
                  RETURN
               END IF
            ELSE IF (KND.EQ.INTRVR .OR. KND.EQ.VARIA) THEN
               W = WORKEX(FINGER+1)
               IF (W.EQ.LBRA .OR. W.EQ.LEFTP) THEN
                  IF (KND.EQ.VARIA) THEN
                     CALL ERROR(159,' ')
                  ELSE
                     CALL ERROR(160,' ')
                  END IF
                  CORR = .FALSE.
                  RETURN
               END IF
            ELSE IF (KND.EQ.DNTCAR) THEN
               GOTO 1000
C  Finally, check intrinsic arrays with no [  ].
            ELSE IF (INTARR(STRI) .AND. WORKEX(FINGER+1).NE.LBRA) THEN
               CALL ERROR(163,' ')
               CORR = .FALSE.
               RETURN
            END IF
         END IF
         GOTO 1000
C
      ELSE IF (RTPAR()) THEN
         CALL POPUP(BSTART)
         IF (.NOT. SUCCES) THEN
            CORR = .FALSE.
            CALL ERROR(48,' ')
            RETURN
         END IF
         PFLAG = .TRUE.
         GOTO 2000
C
      ELSE IF (RTBRA()) THEN
         CALL POPBR(BSTART)
         IF (.NOT. OK) THEN
            CORR = .FALSE.
            CALL ERROR(48,' ')
            RETURN
         END IF
         BFLAG = .TRUE.
         GOTO 2000
C
      ELSE IF (ISCOMA()) THEN
         CALL POPBR(BSTART)
         IF (.NOT. OK) THEN
            CORR = .FALSE.
            CALL ERROR(48,' ')
            RETURN
         END IF
         CALL PUSHBR(BSTART)
         GOTO 2000
C
      ELSE
         CALL MABORT('Illegal item into expression.','XCHECK')
      END IF
C
2000  CONTINUE
C
      IF (FINGER-BSTART.LT.2) THEN
         CORR = .FALSE.
         CALL ERROR(46,' ')
         RETURN
      END IF
C
      SUBCOR = .TRUE.
      EFECT = 0
      DO 555,J=BSTART+1,FINGER-1
         IF (WORKEX(J).EQ.IGNORE) GOTO 555
         EFECT = EFECT+1
         SUBEX(EFECT) = WORKEX(J)
555   CONTINUE
C
      IF (EFECT.NE.0) CALL SUBX(SUBEX,EFECT,SUBCOR)
C
      IF (.NOT. SUBCOR) THEN
         CORR = .FALSE.
         RETURN
      END IF
C
      IF (BFLAG) BSTART = BSTART-1
C
      DO 556,L=BSTART+1,FINGER
556      WORKEX(L) = IGNORE
C
      IF (PFLAG) WORKEX(BSTART) = IDENT
C
      GOTO 1000
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE XGEN ( EXPR, NEX, NA, NN, CORREC, FDECLR, BCNT,
     &                  WORKFL, STVAR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine generates the code needed for the evaluation of
C    expressions. It also checks for the presence of any calls for
C    functions or arrays with incorrect number of arguments.
C    It performs the calculations needed to locate position of
C    multi-dimensional arrays into the linear memory array of merlin 2.
C
C  Input arguments:
C    EXPR       The expression to be processed in a hybrid postfix form.
C               See the documentation of PFIX for information on the
C               form of EXPR.
C    NEX        Number of tokens residing into EXPR.
C               correct. If an error occurs it becomes false and
C               is returned to xpr routine.
C    FDECLR     FDECLR = .TRUE.  -> We are parsing a function
C                                   declaration.
C               FDECLR = .FALSE. -> We are parsing a normal expression.
C    STRVAR     The statement function argument table.
C    WORKFL     Work file unit where intermediate code is written.
C
C  Input / Output arguments:
C    NA         Pointer to the last used element of the alpha-table.
C    NN         Pointer to the last used element of the num-table.
C    CORREC     At entry is true signifying that up to now expr seems
C    BCNT       Holds the total number of lines of MOC written up to
C               now to unit WORKFL.
C
C  Notes:
C    (I) For arrays, processing is based in the following formulas:
C
C                    ----Array dimension   -+----+- (k-1)
C                   <                       !    !
C   (Arrel pos)  =   >   (subscript -lowb )*!    ! (Upperb -lowb +1)
C                   <              k     k  !    !        M     m
C                    ----k=1                      m=1
C
C                    + offset  = wsum + lowsum + offset ,
C
C                   ----array dimension
C                  <
C    where, wsum =  >    volume  * subscript  ,
C                  <           k            k
C                   ----k=1
C
C            ----array dimension              -+---+- (k-1)
C           <                                  !   !
C  Lowsum =  >   lowb  -volume   ,and volume = !   ! (Upperb -lowb +1)
C           <        k        k             k  !   !        M     m
C            ----k=1                                m=1
C
C     concerning semantics: (Arrel-pos) means array element
C     position, lowb means lower bound,
C     upperb means upper bound, and wsum weighted sum.
C
C     (II) A convention that is followed through the entire MCLCOM code
C          and associated with Merlin interface, is that all
C          MOC which needs handling from the interface, is prefixed
C          by a dollar sign. (See subroutine USEMCL of Merlin).
C
C     (III) Arrays which hold kinds ( VAR and STVAR ), use the following
C           coding:
C             -1  Signifies a don't-care situation,
C              1     >>     a user defined variable,
C              2     >>     a user defined array,
C              3     >>     a user defined function,
C              4     >>     an intrinsic function,
C              5     >>     an      >>   variable,
C              6     >>     an      >>   function.
C            101     >>     a simple (non-array) variable, declared in
C                           an argument list.
C            102     >>     an array, declared in an argument list.
C
C     (IV) If GENE is false, code is not generated.
C
C      (V)  This routine is called from XPR.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER EXPR(*)
      LOGICAL CORREC, FDECLR
      CHARACTER*(*) STVAR(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
      COMMON /MCLCNT/ NTOK, NLAB, NVAR, NTYPE, NMOVE, NINF, NFARG, NSTL,
     &                NLOO, NERR, NWARN, NEXT, NCAL, TSTART
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
      LOGICAL GENE, BOUNDS, DEBUG, EOPEN
      COMMON /MCLDTA/ NL, GENE, BLOC, MEMLOC, BOUNDS, DEBUG, INWHAT,
     &                EOPEN, ISERR
      COMMON /MCLUNT/ INP, ERL, BIN, SCR, STF, SPARE, FHELP, ABSFI
C  ---------------------------------------------------------------------
C  Local variables:
C
      PARAMETER ( LLSCR = 80 )
      PARAMETER ( NDIM = (LINLEN-13)/2 )
      INTEGER DIME(NDIM), QUALIT(NDIM), COMCNT(NDIM-1), OFFSET(NDIM)
      INTEGER VOLUME(NDIM), LOW(NDIM), UP(NDIM), POSIT(NDIM)
      CHARACTER OPERS(5:31)*(4), CODLIN*(LLSCR)
      CHARACTER*(IDLEN) ST, NMTEST
      LOGICAL FOUND, OPTEST
      PARAMETER ( VARIA = 1, ARR = 2, STFUN = 3, INTRVR = 4 )
      PARAMETER ( INTARR = 5, INTRFN = 6, DNTCAR = -1 )
      PARAMETER ( ZERO = -1000, ONE = 1000 )
C
C  All MCL-operators are here. Their names are used
C  when XGEN needs them for MOC generation.
      DATA OPERS / '$>  ' , '$<  ' , '$#  ' , '    ' , '$>= ' , '$<= ' ,
     &             '$+  ' , '$-  ' , '$*  ' , '$/  ' , '$** ' ,
     &         11* '    ' , '$AND' , '$OR ' , '$XOR' , '$NOT' , '$=  ' /
C
      ITE1(T) = (T-PLUS)*(T-MINUS)*(T-STAR)*(T-SLASH)
      ITE2(T) = (T-BISTAR)*(T-LT)*(T-LE)*(T-GT)*(T-GE)
      ITE3(T) = (T-EQUAL)*(T-NE)*(T-AND)*(T-OR)*(T-XOR)*(T-NOT)
      OPTEST(T) = ITE1(T).EQ.0 .OR. ITE2(T).EQ.0 .OR. ITE3(T).EQ.0
C  ---------------------------------------------------------------------
C
      SCRUNT = WORKFL
      STACK = 0
      ARRSTK = 0
      SAVEB = BCNT
      IF (FDECLR) THEN
C
C  XGEN called for function declaration processing, so
C  all code should be written to a completely dummy file
C  and then transformed to stfcod. This is done so to
C  allow backward reference to already declared functions
C  for which code is already into stfcod.
         BCNT = 0
         SAVEB = 0
         SCRUNT = FHELP
         REWIND SCRUNT
      END IF
C
C  Loop over all tokens of EXPR.
      DO 1000,I=1,NEX
         TEMPO = EXPR(I)
C
         IF (TEMPO.EQ.IDENT) THEN
            CALL GETIDF(NA,NN,ST)
            IF (FDECLR) THEN
C
C  Enter here only if called for function declaration processing.
C
C  Check dummy names table first.
C
               CALL SERLOC(STVAR,NFARG,ST,FOUND,POS)
               IF (.NOT. FOUND) GOTO 110
               KND = VARIA
               DIMEN = 0
C
C  FWA stands for first word address, and MEMLOC is position of
C  merlin memory array where above it, MCL-functions are
C  linked (pushed and poped).
C  Since STVAR is loaded with the order of the dummy
C  arguments preserved, pos also holds argument order.
C
C  Get fwa value for current dummy argument:
               FWA = MEMLOC + POS
C
               BNDFWA = 0
               GOTO 111
            END IF
110         CALL LOCAN(VAR,NVAR,ST,FOUND,POS)
C
C  If XCHECK performed wrong, XGEN aborts.
            IF (.NOT.FOUND) CALL MABORT('Name not found.','XGEN')
C
            KND = KIND(POS)
            DIMEN = DIM(POS)
            FWA = LOC(POS)
            BNDFWA = IPOS(POS)
111         CONTINUE
            IF (KND.EQ.DNTCAR) THEN
C
C  Previously undeclared identifier found again.
               RETURN
            ELSE IF (KND.EQ.VARIA) THEN
               CALL WPS(SCRUNT,BCNT,FWA)
            ELSE IF (KND.EQ.101) THEN
               BCNT = BCNT+2
               IF (GENE) THEN
                  WRITE (SCRUNT,'(''$IPUSH'')')
                  WRITE (SCRUNT,*) FWA
               END IF
            ELSE IF (KND.EQ.ARR .OR. KND.EQ.102) THEN
               STACK = STACK+1
C
C  A stack scheme is used here for processing due to the
C  special structure of EXPR. All information is kept until a
C  ket is located, and thereafter they are processed in reverse
C  order. ( "Ket" signifies a right bracket. Remember Dirac).
               COMCNT(STACK) = 0
               DIME(STACK) = DIMEN
               QUALIT(STACK) = KND
               OFFSET(STACK) = FWA
               POSIT(STACK) = POS
               ARRSTK = ARRSTK +1
               LOW(ARRSTK) = BNDFWA
               UP(ARRSTK) = BNDFWA+1
               IF (BOUNDS .AND. DIMEN.NE.-1) THEN
C
C  Lower and upper bounds of first dimension are pushed, so
C  to be checked via loccheck for violation during run-time.
                  CALL WPSC(SCRUNT,BCNT,INFO(LOW(ARRSTK)))
                  CALL WPSC(SCRUNT,BCNT,INFO(UP(ARRSTK)))
               END IF
            ELSE IF (KND.EQ.STFUN.OR.KND.EQ.INTRFN) THEN
               STACK = STACK+1
               COMCNT(STACK) = 0
               DIME(STACK) = DIMEN
               POSIT(STACK) = POS
               QUALIT(STACK) = KND
               OFFSET(STACK) = FWA
            ELSE IF (KND.EQ.INTRVR) THEN
               BCNT = BCNT+1
               IF (GENE) WRITE (SCRUNT,'(''$'',A)') VAR(POS)
            ELSE IF (KND.EQ.INTARR) THEN
               STACK = STACK+1
               COMCNT(STACK) = 0
               DIME(STACK) = DIMEN
               QUALIT(STACK) = KND
               POSIT(STACK) = POS
               IF (BOUNDS) THEN
                  CALL WPSC(SCRUNT,BCNT,1)
                  IF (GENE) WRITE (SCRUNT,'(''$DIM'')')
                  BCNT = BCNT+1
               END IF
            END IF
C
         ELSE IF (TEMPO.EQ.ZERO) THEN
            CALL WPSC(SCRUNT,BCNT,0)
C
         ELSE IF (TEMPO.EQ.ONE) THEN
C
C  Ignore everything since "one" is used just to make "not" a binary
C  operator (for checks in xcheck).
C  Processing proceeds to the next token.
         ELSE IF (TEMPO.EQ.UNSI) THEN
            NN = NN+1
            BCNT = BCNT+2
            IF (GENE) THEN
               WRITE (SCRUNT,'(''$PUSHC'')')
               WRITE (SCRUNT,*) NUM(NN)
            END IF
C
         ELSE IF (OPTEST(TEMPO)) THEN
            BCNT = BCNT+1
            IF (GENE) WRITE (SCRUNT,'(A)') OPERS(TEMPO)
C
         ELSE IF (TEMPO.EQ.COMMA) THEN
            COMCNT(STACK) = COMCNT(STACK)+1
            QTEST = QUALIT(STACK)
            IF ((QTEST.EQ.ARR .OR. QTEST.EQ.102) .AND. BOUNDS) THEN
               BCNT = BCNT+1
               IF (GENE) WRITE (SCRUNT,'(A)') '$LOCCHECK'
               IF (GENE) THEN
                  IF (DIME(STACK).GT.0 .OR.
     &                     ABS(DIME(STACK)).NE.COMCNT(STACK)+1) THEN
                  ICSS2 = COMCNT(STACK)*2
                  CALL WPSC(SCRUNT,BCNT,INFO(LOW(ARRSTK)+ICSS2))
                  CALL WPSC(SCRUNT,BCNT,INFO(UP(ARRSTK)+ICSS2))
               END IF
            END IF
         ELSE IF (QTEST.EQ.INTARR) THEN
            CORREC = .FALSE.
            CALL ERROR(164,' ')
            RETURN
         END IF
C
      ELSE IF (TEMPO.EQ.RBRA) THEN
C
C  Main reverse-order processing activated.
         IF (COMCNT(STACK)+1.EQ.ABS(DIME(STACK))) THEN
            QTEST = QUALIT(STACK)
            BNDFWA = IPOS(POSIT(STACK))
            IF (QTEST.EQ.STFUN) THEN
C
C  Link argument values with the positions where the
C  following MCL-functions code addresses.
               DO 118,L=DIME(STACK),1,-1
118               CALL WPOP(SCRUNT,BCNT,OFFSET(STACK)+L)
C
C   Read MCL-function code from scratch and write it to
C   the intermediate MOC file.
               IF (.NOT.GENE) GOTO 123
               REWIND STF
               DO 121,SKIP=1,INFO(BNDFWA)-1
121               READ (STF,'(A)') CODLIN
               DO 122,J=1,INFO(BNDFWA+1)-INFO(BNDFWA)+1
                  READ (STF,'(A)') CODLIN
                  WRITE (SCRUNT,'(A)') CODLIN
                  BCNT = BCNT+1
122            CONTINUE
123            CONTINUE
            ELSE IF (QTEST.EQ.INTRFN) THEN
               NMTEST = VAR(POSIT(STACK))
               BCNT = BCNT+1
               IF (GENE) WRITE (SCRUNT,'(A11)') '$'//NMTEST
            ELSE IF (QTEST.EQ.ARR .OR. QTEST.EQ.102) THEN
               IF (BOUNDS .AND. DIME(STACK).GT.0) THEN
                  BCNT = BCNT+1
                  IF (GENE) WRITE (SCRUNT,'(''$LOCCHECK'')')
               END IF
C
C  Perform array position processing as presented in the
C  introductory comments of this routine.
               IF ( .NOT. GENE ) GOTO 818
C
               MM = LOW( ARRSTK )
               DMS = ABS(DIME(STACK))
               DO 101,VL=1,DMS
                  VOLUME(VL) = 1
C
C  Note the (MM+1) below. It is written like this, so to be consistent
C  with the formulas in the introductory comments.
                  DO 101,L=MM+1,MM+2*(VL-1),2
101                  VOLUME(VL) = VOLUME(VL)*(1+INFO(L)-INFO(L-1))
                  LOWSUM = 0
                  MM = MM-2
                  DO 102,M=1,DMS
                     MM = MM+2
                     LOWSUM = LOWSUM+INFO(MM)*VOLUME(M)
102               CONTINUE
C
                  DO 103,N=1,DMS
103                  CALL WPSC(SCRUNT,BCNT,VOLUME(N))
                  BCNT = BCNT+2
                  WRITE (SCRUNT,'(''$WSUM'')')
                  WRITE (SCRUNT,*) DMS
                  IF (QTEST.EQ.2) THEN
                     BCNT = BCNT+4
                     WRITE (SCRUNT,'(''$PUSHA'')')
                     WRITE (SCRUNT,*) OFFSET(STACK)-LOWSUM
                     WRITE (SCRUNT,'(''$+'')')
                     WRITE (SCRUNT,'(''$GETCONT'')')
                  ELSE
C  102
                     BCNT = BCNT+3
                     CALL WPSC(SCRUNT,BCNT,-LOWSUM)
                     WRITE (SCRUNT,'(''$+'')')
                     CALL WPSC(SCRUNT,BCNT,OFFSET(STACK))
                     WRITE (SCRUNT,'(''$IOFFSET'')')
                     WRITE (SCRUNT,'(''$GETCONT'')')
                  END IF
818               ARRSTK = ARRSTK -1
C
C  One array lower...
               ELSE IF (QTEST.EQ.INTARR) THEN
                  BCNT = BCNT+1
                  IF (BOUNDS) THEN
                     BCNT = BCNT+1
                     IF (GENE) WRITE (SCRUNT,'(''$LOCCHECK'')')
                  END IF
                  IF (GENE) WRITE (SCRUNT,'(''$'',A)')
     &                                  VAR(POSIT(STACK))
               END IF
C
               STACK = STACK-1
C
            ELSE
C
C  The number of commas found, does not coincide with the expected
C  number implied from the declaration, so...
               NMTEST = VAR(POSIT(STACK))
               IF (NMTEST.EQ.'MIN' .OR. NMTEST.EQ.'MAX' .OR.
     &                      NMTEST.EQ.'MEAN') THEN
               BCNT = BCNT+2
               IF (GENE) THEN
                  WRITE (SCRUNT,'(A11)') '$'//NMTEST
                  WRITE (SCRUNT,*) COMCNT(STACK)+1
               END IF
C
               STACK = STACK-1
C
               GOTO 1000
            END IF
            CORREC = .FALSE.
            CALL ERROR(78,NMTEST)
            RETURN
         END IF
C
      ELSE
         CALL MABORT('Illegal item into expression.','XGEN')
      END IF
1000  CONTINUE
C
      IF ((.NOT. FDECLR) .OR. (.NOT.GENE)) GOTO 9000
C
      REWIND STF
      REWIND SCRUNT
      DO 1001,KL=1,NSTL
         READ (STF,'(A)') CODLIN
1001  CONTINUE
      DO 1002,ML=1,BCNT-SAVEB
         READ (SCRUNT,'(A)') CODLIN
         WRITE (STF,'(A)') CODLIN
1002  CONTINUE
C
9000  CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE XPR ( EX, NEX, NA, NN, CORR, FDECLR, BLOC, SCRUNT,
     &                 STVAR )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine, orchestrates expression-processing.
C
C  Input arguments:
C    EX         Integer array containing the tokens of the expression
C               to be processed.
C    NEX        Number of tokens, EX is composed of.
C    FDECLR     FDECLR = .TRUE.  -> We are parsing a function
C                                   declaration.
C               FDECLR = .FALSE. -> We are parsing a normal expression.
C    SCRUNT     The output unit.
C    STVAR      The statement function argument table.
C
C  Output arguments:
C    CORR       CORR = .TRUE.  -> The expression is correct.
C               CORR = .FALSE. -> The expression is incorrect.
C
C  Input / Output arguments:
C    NA         Pointer to the last used element of the alpha-table.
C    NN         Pointer to the last used element of the num-table.
C    BLOC       Lines written to the scratch file.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER EX(*)
      CHARACTER*(*) STVAR(*)
      LOGICAL CORR, FDECLR
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( ONE = 1000, ZERO = -1000 )
      DIMENSION  NEWEX(2*LINLEN), EXP2(3*LINLEN)
      LOGICAL OPERAN
      OPERAN(XPART) = (XPART.EQ.UNSI) .OR. (XPART.EQ.IDENT)
C  ---------------------------------------------------------------------
C
      NTOKS = NEX
      IF (NTOKS.EQ.0) THEN
C  XPR was called but no expression was present.
         CALL ERROR (132,' ')
         CORR = .FALSE.
         RETURN
      END IF
C
C  Zero-insertion for negations:
C  Zeros are inserted in front of every + or - when they are used as
C  signs. The resulting code is not that efficient, so do not place an
C  unnecessary + in front of an expression.
      INDEX = 1
      IF (EX(1).EQ.PLUS .OR. EX(1).EQ.MINUS) THEN
         NEWEX(INDEX) = ZERO
         INDEX = INDEX+1
      END IF
      NEWEX(INDEX) = EX(1)
      IF (NTOKS.EQ.1) GOTO 224
      DO 212,I=2,NTOKS-1
         IF (EX(I).EQ.PLUS .OR. EX(I).EQ.MINUS) THEN
            IF (.NOT.(
     &               (OPERAN(EX(I-1)) .AND. (OPERAN(EX(I+1)) .OR.
     &                                       EX(I+1).EQ.LEFTP .OR.
     &                                       EX(I+1).EQ.NOT)) .OR.
     &                                       EX(I-1).EQ.RIGHTP .OR.
     &                                       EX(I-1).EQ.RBRA )) THEN
               INDEX = INDEX+1
               NEWEX(INDEX) = ZERO
            END IF
         END IF
         INDEX = INDEX+1
         NEWEX(INDEX) = EX(I)
212   CONTINUE
      INDEX = INDEX+1
      NEWEX(INDEX) = EX(NTOKS)
C
C  Zeros inserted. Continue to insert ones for "not" processing.
C  During the code-generation stage of XGEN, "one"s are ignored.
C  They serve the purpose of transforming "not" to a pseudo
C  binary operator.
224   NTOKS = INDEX
      DO 190,J=1,NTOKS
190      EXP2(J) = NEWEX(J)
      INDEX = 1
      IF (EXP2(1).EQ.NOT) THEN
         NEWEX(INDEX) = ONE
         INDEX = INDEX+1
      END IF
      NEWEX(INDEX) = EXP2(1)
      IF (NTOKS.EQ.1) GOTO 225
      DO 213,K=2,NTOKS-1
         IF (EXP2(K).EQ.NOT) THEN
            INDEX = INDEX+1
            NEWEX(INDEX) = ONE
         END IF
         INDEX = INDEX+1
         NEWEX(INDEX) = EXP2(K)
213   CONTINUE
      INDEX = INDEX+1
      NEWEX(INDEX) = EXP2(NTOKS)
C
C  Insertions completed. Continue for validity-checking of
C  the new layout of expression. Note that:
C  If the expression without the inserted items is correct
C  it remains so and after the insertions.
225   CONTINUE
      NNA = NA
      NNN = NN
      CALL XCHECK(NEWEX,INDEX,CORR,NNA,FDECLR,STVAR)
      IF (.NOT.CORR) GOTO 500
C
C  If CORR is true, then the sole situation left unchecked
C  is the call to a function with the wrong number of arguments, or
C  the reference to an array element with a wrong number
C  of subscripts. Those check-ups are performed via XGEN.
C  Now call PFIX to transform NEWEX to the hybrid postfix form.
C  (See documentation of PFIX for details).
      CALL PFIX(NEWEX,INDEX,FINTOK)
C
C  Generate object code.
      CALL XGEN(NEWEX,FINTOK,NNA,NNN,CORR,FDECLR,BLOC,SCRUNT,STVAR)
C
C  Advance NA, NN pointers to their correct position.
500   CONTINUE
      DO 510,I=1,NEX
         EXI = EX(I)
         IF (EXI.EQ.IDENT .OR. EXI.EQ.STRING) THEN
            NA = NA+1
            NN = NN+2
         ELSE IF (EXI.EQ.UNSI) THEN
            NN = NN+1
         END IF
510   CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE ZEROM ( IND1, IND2 )
C  ---------------------------------------------------------------------
C
C  Description:
C    This routine will assign a value of 7 to all MOVE TO entries
C    in the type table, residing in a block specified by two limits,
C    if their target label lies also in the block.
C
C  Input arguments:
C    IND1       Integer variable. This is a pointer to the type table,
C               and is the lower limit from where the search for
C               any MOVE TO statements will start.
C    IND2       Integer variable. This is a pointer to the type table,
C               and is the upper limit to where the search for
C               any MOVE TO statements will terminate.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      PARAMETER ( IDLEN = 30 )
      PARAMETER ( LENONE = 120, MXCONT = 10, LINLEN = LENONE*MXCONT )
C -- Tables --
      PARAMETER ( MXTOK = LINLEN+1, MXNUM = (LINLEN+1)/2 )
      PARAMETER ( MXLAB = 100, MXLOO = 100, MXTYP = 1000 )
      PARAMETER ( MXVAR = 500, MXINFO = 500, MXMOV = 500 )
      PARAMETER ( MXEXT = 200, MXCAL = 500 )
C  The lexical analysis tables.
      INTEGER TOKEN(MXTOK)
      CHARACTER*(LINLEN) ALPHA
      $(TYPE) NUM(MXNUM)
C  Tables related to variables.
      CHARACTER*(IDLEN) VAR(MXVAR)
      INTEGER KIND(MXVAR), LOC(MXVAR), DIM(MXVAR), IPOS(MXVAR)
      INTEGER INFO(MXINFO)
C  Label related tables.
      CHARACTER*(IDLEN) LABEL(MXLAB), MVELAB(MXMOV)
      INTEGER LABLOC(MXLAB)
C  The type (and related) tables.
      INTEGER TYPE(MXTYP), TDEF(MXTYP), TREL(MXTYP)
C  Loop related tables.
      INTEGER LLOC(MXLOO), LMEM(MXLOO)
C  Subprogram related tables.
      CHARACTER*(IDLEN) EXT(MXEXT), CAL(MXCAL)
      INTEGER EXTDEF(MXEXT), CALDEF(MXCAL), NARDEF(MXEXT), NARCAL(MXCAL)
      INTEGER SUBUSE(MXEXT)
      COMMON /ITABLE/ TOKEN, KIND, LOC, DIM, IPOS, INFO, LABLOC, TYPE,
     &                TDEF, TREL, LLOC, LMEM, EXTDEF, CALDEF, NARDEF,
     &                NARCAL, SUBUSE
      COMMON /CTABLE/ ALPHA, VAR, LABEL, MVELAB, EXT, CAL
      COMMON /RTABLE/ NUM
C -- End of tables --
C  ---------------------------------------------------------------------
C
C  Get locations for the block limits.
C  LOC1 is the location for the lower limit.
C  LOC2 is the location for the upper limit.
      LOC1 = TDEF(IND1)
      LOC2 = TDEF(IND2)
C
C  Search the type table.
      DO 10,I=IND1,IND2
         IF (TYPE(I).EQ.-7) THEN
C  A MOVE TO statement was found.
C  Since the type-link table holds only relative numbers, calculate
C  the absolute location of the target label.
            IR = TREL(I)
            IF (IR.LT.0) IR = IR-1
            JT = TDEF(I)+IR
C  Check if the location lies inside the allowable bounds.
            IF (JT.GE.LOC1 .AND. JT.LE.LOC2) TYPE(I) = 7
         END IF
10    CONTINUE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONECP ( CONFIL, IERRU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Copies the error file produced by the configuration parsing
C    routine MCONF to the standard output (*). The error file is then
C    removed.
C    This routine is called whenever there are configuration problems
C    concerning the output file.
C
C  Input arguments:
C    CONFIL     Name of the configuration file.
C    IERRU      Unit number of the error file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CONFIL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
C  Initial message.
      CALL DASHO
      WRITE (*,150) CONFIL(1:LENGTH(CONFIL))
150   FORMAT ($(FC)'The following error(s) have been detected while '/
     &        $(FC)'processing the Merlin configuration file "',A,'":')
      CALL DASHO
C
C  Copy error file.
      REWIND IERRU
10    READ (IERRU,'(A)',END=20) LINE
      LE = LENGTH(LINE)
      IF (LE.EQ.0) THEN
         WRITE (*,*)
      ELSE
         WRITE (*,'($(FC)A)') LINE(1:LE)
      END IF
      GOTO 10
C
C  Remove error file.
20    CONTINUE
      CLOSE (IERRU,STATUS='DELETE')
      CALL DASHO
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONEOU ( CONFIL, IUOUT, IERRU )
C  ---------------------------------------------------------------------
C
C  Description:
C    Copies the error file produced by the configuration parsing
C    routine MCONF to the Merlin output file. The error file is then
C    removed.
C
C  Input arguments:
C    CONFIL     Name of the configuration file.
C    IUOUT      The Merlin output unit.
C    IERRU      Unit number of the error file.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CONFIL
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*80 LINE
C  ---------------------------------------------------------------------
C
C  Initial message.
      CALL DASH
      WRITE (IUOUT,150) CONFIL(1:LENGTH(CONFIL))
150   FORMAT ($(FC)'The following error(s) have been detected while '/
     &        $(FC)'processing the Merlin configuration file "',A,'":')
      CALL DASH
C
C  Copy error file.
      REWIND IERRU
10    READ (IERRU,'(A)',END=20) LINE
      LE = LENGTH(LINE)
      IF (LE.EQ.0) THEN
         WRITE (IUOUT,*)
      ELSE
         WRITE (IUOUT,'($(FC)A)') LINE(1:LE)
      END IF
      GOTO 10
C
C  Remove error file.
20    CONTINUE
      CLOSE (IERRU,STATUS='DELETE')
      CALL DASH
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE DASHO
C  ---------------------------------------------------------------------
C
C  Description:
C    Prints a horizontal line of dashes. Subroutine DASHO is used before
C    the output unit is opened; hence it cannot write to IUOUT.
C
C  ---------------------------------------------------------------------
C
      WRITE (*,10)
C
10    FORMAT ($(FC)79('-'))
      END
C  ---------------------------------------------------------------------
      SUBROUTINE I2STR ( N, STR, LES )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts an integer (N) to its string representation (STR).
C
C  Input arguments:
C    N          The integer to be converted.
C
C  Output arguments:
C    STR        The first LES characters of STR contain the string
C               representation of N.
C    LES        Effective length of STR.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*5 FORM
      DATA FORM / '(I__)' /
C  ---------------------------------------------------------------------
C
      LE = LEN(STR)
      WRITE (FORM(3:4),'(I2.2)') LE
      WRITE (STR,FORM) N
C
      DO 10,I=1,LE
         IF (STR(I:I).NE.' ') THEN
            DO 20,J=I,LE
               JI = J-I+1
               STR(JI:JI) = STR(J:J)
20          CONTINUE
            STR(LE-I+2:) = ' '
            LES = LE-I+1
            RETURN
         END IF
10    CONTINUE
      LES = LE
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IRANGE ( TOKEN, NUM, ALPHA, NT, NN, IERR, PIR, NPIR,
     &                    MXPIR, PDESCF, NPL, COMM, KEY, IUOUT, LEP,
     &                    WEAK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine IRANGE parses the definition of a range or set of
C    integers and fills in the appropriate data structures.
C    A range of integers has the form:
C      L S,E R
C    where: L is ( or [ indicating an open/closed interval on the left.
C           R is ) or ] indicating an open/closed interval on the right.
C           S is an integer (start of the interval) or -inf (for
C             infinity).
C           E is an integer (end of the interval) or inf (for infinity).
C    Examples: (-1,1)  (-inf,0]  [10,100)  [70,80]
C
C    A set (list) of integers has the form:
C      { I1, I2, I3, ... }
C    where I1, I2,... etc. are integers.
C
C    Panel related data structures are described in subroutine PARSP.
C    The lexical analysis tables (TOKEN, NUM, ALPHA) are described in
C    subroutine LEX.
C
C  Input arguments:
C    TOKEN      Tokens in the current line (see subroutine LEX).
C    NUM        Numbers in the current line (see subroutine LEX).
C    ALPHA      Identifiers/strings in the current line (see subroutine
C               LEX).
C    MXPIR      Maximum allowed number of entries in the PIR table.
C    PDESCF     Name of the panel description file.
C    NPL        Number of lines read so far from the panel description
C               file.
C    COMM       The panel currently under processing.
C    KEY        The keyword currently under processing.
C    IUOUT      The Merlin output unit.
C    LEP        Effective length of PDESCF.
C    WEAK       Error messages are printed only of WEAK = .TRUE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  Input / Output arguments:
C    NT         Number of tokens (TOKEN) used so far.
C    NN         Number of numbers (NUM) used so far.
C    PIR        The PIR table (see subroutine PARSP).
C    NPIR       Number of entries in the PIR array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(*)
      $(TYPE) NUM(*), PIR(MXPIR)
      CHARACTER*(*) ALPHA, PDESCF, COMM, KEY
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
      $(TYPE) XLE, XLI, XRI
C  ---------------------------------------------------------------------
C
      IERR = 0
      NT = NT+1
      INIT = TOKEN(NT)
      IF (INIT.EQ.LBRACE) THEN
C  A list of values is expected.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = 2
C  Allocate PIR space for the # of items.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         ISTLI = NPIR
10       CONTINUE
         NT = NT+1
         ITOK = TOKEN(NT)
         ISI = 1
         IF (ITOK.EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (ITOK.EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NN = NN+1
         XLI = NUM(NN)
         IF (XLI.NE.INT(XLI)) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = ISI*NINT(XLI)
         NT = NT+1
         IF (TOKEN(NT).EQ.COMMA) GOTO 10
         IF (TOKEN(NT).EQ.RBRACE) THEN
C  End of list ok.
            NLI = NPIR-ISTLI
            PIR(ISTLI) = NLI
            RETURN
         END IF
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE IF (INIT.EQ.LEFTP .OR. INIT.EQ.LBRA) THEN
C  A left open/closed interval is expected.
         LEF = 0
         IRI = 0
         IF (INIT.EQ.LEFTP) ITPL = 1
         IF (INIT.EQ.LBRA) ITPL = 2
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPL = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,273) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XLE = NUM(NN)
            IF (XLE.NE.INT(XLE)) THEN
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,274) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
            LEF = ISI*NINT(XLE)
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,273) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.COMMA) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,276) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         ITPR = 0
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPR = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,277) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XRI = NUM(NN)
            IF (XRI.NE.INT(XRI)) THEN
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,278) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
            IRI = ISI*NINT(XRI)
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,277) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         NT = NT+1
         IF (TOKEN(NT).EQ.RIGHTP) THEN
            IF (ITPR.EQ.0) ITPR = 1
         ELSE IF (TOKEN(NT).EQ.RBRA) THEN
            IF (ITPR.EQ.0) ITPR = 2
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,280) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         IF (NPIR+5.GT.MXPIR)
     &          CALL OVER('pir','PIR','IRANGE',MXPIR,'MXPIR')
         PIR(NPIR+1) = 1
         PIR(NPIR+2) = ITPL
         PIR(NPIR+3) = LEF
         PIR(NPIR+4) = ITPR
         PIR(NPIR+5) = IRI
         NPIR = NPIR+5
         IF (ITPL.EQ. 1 .OR.ITPL.EQ.2) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = LEF.LE.IRI
            ELSE
               OK = ITPR.EQ.3
            END IF
         ELSE IF (ITPL.EQ.-3) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = .TRUE.
            ELSE
               OK = ITPL.LT.ITPR
            END IF
         ELSE
            OK = .FALSE.
         END IF
         IF (.NOT.OK) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,281) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
      ELSE IF (INIT.EQ.EOL) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,282) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,283) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
C
270   FORMAT ($(FC)'Integers only should be present in the ',
     &        'list of allowed values'/
     &        $(FC)'for the ',A,' keyword in the ',A,' panel.')
272   FORMAT ($(FC)'The list of integer values for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
273   FORMAT ($(FC)'The left limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
274   FORMAT ($(FC)'The left limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'should be an integer.')
276   FORMAT ($(FC)'A comma is expected after the left',
     &        ' limit for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
277   FORMAT ($(FC)'The right limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
278   FORMAT ($(FC)'The right limit for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'should be an integer.')
280   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
281   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect. Left limit > right limit.')
282   FORMAT ($(FC)'There is no description of allowed values ',
     &        'for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
283   FORMAT ($(FC)'The specification of allowed values for the ',A,
     &        ' keyword'
     &        /$(FC)'in the ',A,' panel, is incorrect.')
      END
C  ---------------------------------------------------------------------
      LOGICAL FUNCTION ISCAP ( C )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns .TRUE. if the input character C is a capital letter;
C    .FALSE. otherwise.
C
C  Input arguments:
C    C          Character to check.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER C
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*26 LETT
      DATA LETT / 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' /
C  ---------------------------------------------------------------------
C
      ISCAP = INDEX(LETT,C).NE.0
      END
C  ---------------------------------------------------------------------
      FUNCTION ISCOMP ( A, B )
C  ---------------------------------------------------------------------
C
C  Description:
C    Performs a case insensitive comparison of the two character strings
C    A and B. Returns 0 if they don't match, non-zero if they match.
C
C  Input arguments:
C    A          The first of the two strings to be compared.
C    B          The second of the two strings to be compared.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) A, B
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CA, CB
      LOGICAL BLANK
C  ---------------------------------------------------------------------
C
      LA = LEN(A)
      LB = LEN(B)
C
C  Compare all characters in A and B, one by one.
      LE = MIN(LA,LB)
      DO 10,I=1,LE
         CA = A(I:I)
         CB = B(I:I)
C  Convert to upper case before comparison.
         CALL UPPER(CA,1)
         CALL UPPER(CB,1)
         IF (CA.NE.CB) THEN
            ISCOMP = 0
            RETURN
         END IF
10    CONTINUE
C
C  Blanks at the end of a string are ignored.
      IF (LA.GT.LB) THEN
         BLANK = A(LB+1:) .EQ. ' '
      ELSE IF (LB.GT.LA) THEN
         BLANK = B(LA+1:) .EQ. ' '
      ELSE
         BLANK = .TRUE.
      END IF
C
      IF (BLANK) THEN
         ISCOMP = 1
      ELSE
         ISCOMP = 0
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION LENGTH ( STRING )
C  ---------------------------------------------------------------------
C
C  Description:
C    Returns the position of the last non-blank non-tab character in
C    the input string, ie. its effective length.
C
C  Input arguments:
C    STRING    A character variable whose length we seek.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER STRING*(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
C  ---------------------------------------------------------------------
C
      DO 10,I=LEN(STRING),1,-1
         C = STRING(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            LENGTH = I
            RETURN
         END IF
10    CONTINUE
      LENGTH = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE MCONF ( CFILE, ICONU, IFILEU, IERRU, NMCOM, MXPLUG,
     &                   COMMAN, EXCLU, NEXCLU, NURANG, IERROR, HEADER,
     &                   MODE, PROLOG, EPILOG, FILIN, FILOUT, PDESCF,
     &                   STRO, WEAK, IPRECO, OPRECO, BI, SM, IUINP,
     &                   IUOUT, MINU, MAXU, HASAPP, OUTAPP, NSCHA,
     &                   NSINT, NSREAL, HFILE, MFILE, EFILE, OFILE,
     &                   NDIG, NEOFA )
C  ---------------------------------------------------------------------
C
C  Purpose:
C    Reads the Merlin configuration file and changes the defaults
C    according to the configuration directives. Errors are written
C    to a scratch file with unit number IERRU. If any errors
C    actually occur (IERROR.NE.0) this file remains open when
C    subroutine MCONF returns.
C
C  Input arguments:
C    CFILE     The name of the configuration file to be read.
C
C    ICONU     Unit number to be assigned to the configuration file
C              in the corresponding OPEN statement.
C
C    IFILEU    Unit number to be assigned to the file specified in
C              a "FILE" configuration directive.
C
C    IERRU     Unit number to be assigned to the a scratch file
C              that receives all error messages.
C
C    NMCOM     Number of Merlin commands in the COMMAN array. The
C              actual length of this array is NMCOM+MXPLUG
C
C    MXPLUG    Maximum number of plugs that may be defined by the user.
C
C    COMMAN    Character array, containing all the Merlin commands and
C              plug-in names. This array may or may not be sorted.
C
C    EXCLU     Character array. Contains additional commands that must
C              be excluded form being valid command names in a PLUG
C              directive. Mainly used by the MCL compiler to exclude
C              IF, WHEN, LOOP etc. from being used as plug-in names.
C
C    NEXCLU    Number of entries in the EXCLU array.
C
C    NURANG    Minimum acceptable number of units that are assigned to
C              the unit allocator.
C
C  Output arguments:
C    IERROR    Indicates whether any errors occurred while processing
C              the configuration file. Possible values are:
C                IERROR = 0  -->  No error.
C                IERROR = 1  -->  Some error (not related to the
C                                 configuration directives that
C                                 affect the Merlin output file/unit).
C                IERROR = 2  -->  An error occurred while processing
C                                 configuration directives that affect
C                                 the Merlin output file/unit).
C              When IERROR = 2 Merlin must not open the output file.
C
C  Input / Output arguments:
C    The arguments listed below, on entry have the default values
C    assigned by Merlin/MCL. They are not referenced in any way, other
C    than by assignment statements that change their values when a
C    corresponding configuration directive is encountered.
C
C    Variable:      Configuration directive:
C    HEADER         HEADER
C    MODE           MODE
C    PROLOG         PLOLOG
C    EPILOG         EPILOG
C    FILIN          INPUT_FILE
C    FILOUT         OUTPUT_FILE
C    PDESCF         PDESC_FILE
C    STRO           PRINTOUT
C    WEAK           PRINTOUT
C    INPRECO        INPUT_PRECONN
C    OPRECO         OUTPUT_PRECONN
C    BI             BIGGER
C    SM             SMALLER
C    IUINP          INPUT_UNIT
C    IUOUT          OUTPUT_UNIT
C    MINU           UNIT_RANGE
C    MAXU           UNIT_RANGE
C    HASAPP         HAS_APPEND
C    OUTAPP         OUTPUT_FILE
C    NSCHA          SIZE_CHAR
C    NSINT          SIZE_INT
C    NSREAL         SIZE_REAL
C    HFILE          HELP_FILE
C    EFILE          MCL_ERROR_FILE
C    OFILE          MCL_OBJECT_FILE
C    NDIG           MACHINE_DIGITS
C    IEOF           ONEOF
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
      CHARACTER*(*) CFILE, MODE, PROLOG, EPILOG, FILIN, FILOUT, PDESCF
      CHARACTER*(*) COMMAN(*), EXCLU(NEXCLU), HFILE, MFILE, EFILE, OFILE
      LOGICAL HEADER, STRO, WEAK, IPRECO, OPRECO, HASAPP, OUTAPP
C
      PARAMETER ( LCLEN = 80 )
      CHARACTER*(LCLEN) LINE, ORIG, NAME, FL, ITEM
      INTEGER IS(LCLEN/2), IE(LCLEN/2)
      CHARACTER*20 STR
      LOGICAL EOF, EXI, OPE
C
      PARAMETER ( NDIR = 27 )
      CHARACTER*15 DIR(NDIR)
      INTEGER NITAL(NDIR)
C
      DATA (DIR(I),NITAL(I),I=1,10) /
     &     'HAS_APPEND     ', 2,
     &     'HEADER         ', 2,
     &     'SIZE_CHAR      ', 2,
     &     'MODE           ', 2,
     &     'FILE           ', 2,
     &     'PROLOG         ', 2,
     &     'EPILOG         ', 2,
     &     'INPUT_FILE     ', 2,
     &     'INPUT_UNIT     ', 2,
     &     'OUTPUT_FILE    ', -1 /
      DATA (DIR(I),NITAL(I),I=11,20) /
     &     'OUTPUT_UNIT    ', -1,
     &     'INPUT_PRECONN  ', 2,
     &     'BIGGER         ', 2,
     &     'OUTPUT_PRECONN ', -1,
     &     'UNIT_RANGE     ', 3,
     &     'SIZE_INT       ', 2,
     &     'SIZE_REAL      ', 2,
     &     'HELP_FILE      ', 2,
     &     'SMALLER        ', 2,
     &     'PRINTOUT       ', 2 /
      DATA (DIR(I),NITAL(I),I=21,NDIR) /
     &     'PLUG           ', 3 ,
     &     'PDESC_FILE     ', 2 ,
     &     'MACRO_FILE     ', 2 ,
     &     'MCL_ERROR_FILE ', 2 ,
     &     'MCL_OBJECT_FILE', 2 ,
     &     'MACHINE_DIGITS ', 2 ,
     &     'ONEOF          ', 2 /
C
      IERROR = 0
      IF (CFILE.EQ.' ') RETURN
C
      INQUIRE (FILE=CFILE,EXIST=EXI,OPENED=OPE,NUMBER=ICOI,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         OPEN (IERRU,STATUS='SCRATCH')
         REWIND IERRU
         CALL I2STR(IERR,STR,LES)
         CALL CONERR(IERRU,0,LINE,ORIG,IS,IE,0,
     &           'Cannot inquire the Merlin configuration file. '//
     &           'IOSTAT = '//STR(1:LES)//'.')
         IERROR = 1
         RETURN
      END IF
      IF (.NOT.EXI) RETURN
      IF (OPE) THEN
         OPEN (IERRU,STATUS='SCRATCH')
         REWIND IERRU
         CALL I2STR(ICOI,STR,LES)
         CALL CONERR(IERRU,0,LINE,ORIG,IS,IE,0,
     &           'The Merlin configuration file is already open.'//
     &           ' UNIT = '//STR(1:LES)//'.')
         IERROR = 1
         RETURN
      END IF
C
      OPEN (ICONU,FILE=CFILE,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         OPEN (IERRU,STATUS='SCRATCH')
         REWIND IERRU
         CALL I2STR(IERR,STR,LES)
         CALL CONERR(IERRU,0,LINE,ORIG,IS,IE,0,
     &           'The Merlin configuration file cannot be opened. '//
     &           'IOSTAT = '//STR(1:LES)//'.')
         IERROR = 1
         RETURN
      END IF
      REWIND ICONU
      OPEN (IERRU,STATUS='SCRATCH')
      REWIND IERRU
C
      EOF = .FALSE.
      NL = 0
100   CONTINUE
      IF (EOF) GOTO 200
      READ (ICONU,'(A)',END=200) ORIG
      NL = NL+1
      LL = LENGTH(LINE)
      CALL QLCR(ORIG,LL,LINE,IS,IE,NIT,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &                  'Syntax error (Probably a missing quote).')
         IERROR = 1
         GOTO 100
      END IF
C
C  Is this an empty line with no items ?
      IF (NIT.EQ.0) GOTO 100
C
C  Is this a comment line ?
      IF (LINE(IS(1):IS(1)).EQ.'%') GOTO 100
C
C  Check if this is a correct configuration directive.
      ITEM = LINE(IS(1):IE(1))
      LEI = IE(1)-IS(1)+1
C
      DO 500,I=1,NDIR
         IF (ISCOMP(ITEM,DIR(I)).NE.0) THEN
            NI = NITAL(I)
            IF (NI.NE.-1 .AND. NIT.NE.NI) THEN
               CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
               IERROR = 1
               GOTO 100
            ELSE
C  Check for arguments with zero length.
               DO 91,J=1,NIT
                  LEP = IE(J)-IS(J)+1
                  IF (LEP.EQ.0) THEN
                     CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'One or more parameters of "'//ITEM(1:LEI)//
     &            '" have zero length.')
                     IERROR = 1
                     GOTO 100
                  END IF
91             CONTINUE
               GOTO (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,
     &                        20,21,22,23,24,25,26,27),I
            END IF
         END IF
500   CONTINUE
C
      CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'There is no such configuration directive: "'//
     &            ITEM(1:LEI)//'".')
      IERROR = 1
      GOTO 100
C
C  ..................................................... HAS_APPEND
C
C  Syntax: HAS_APPEND yes | no
C
1     CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'YES').NE.0) THEN
         HASAPP = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         HASAPP = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "yes" or "no".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... HEADER
C
C  Syntax: HEADER on | off
C
2     CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'ON').NE.0) THEN
         HEADER = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'OFF').NE.0) THEN
         HEADER = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "on" or "off".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... SIZE_CHAR
C
C  Syntax: SIZE_CHAR int
C
3     CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NSIZE,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NSIZE.LE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer > 0.')
         IERROR = 1
      ELSE
         NSCHA = NSIZE
      END IF
      GOTO 100
C
C  ..................................................... MODE
C
C  Syntax: MODE iaf | batch
C
4     CONTINUE
      LL = IE(2)-IS(2)+1
      IF (ISCOMP(LINE(IS(2):IE(2)),'IAF').NE.0) THEN
         MODE = LINE(IS(2):IE(2))
         CALL UPPER(MODE,LL)
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'BATCH').NE.0) THEN
         MODE = LINE(IS(2):IE(2))
         CALL UPPER(MODE,LL)
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "iaf" or "batch".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... FILE
C
C  Syntax: FILE filename
C                  data lines
C          ...END
C
5     CONTINUE
      NAME = LINE(IS(2):IE(2))
      LEF = IE(2)-IS(2)+1
      INQUIRE (FILE=NAME(1:LEF),OPENED=OPE,NUMBER=IUFI,IOSTAT=IERR)
      IF (OPE) THEN
         CALL I2STR(IUFI,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,'File "'//
     &                  NAME(1:LEF)//'" is already open. UNIT = '//
     &                  STR(1:LES)//'.')
         IERR = -1
      ELSE
         OPEN (IFILEU,FILE=NAME(1:LEF),IOSTAT=IERR)
         IF (IERR.NE.0) THEN
            CALL I2STR(IERR,STR,LES)
            CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,'File "'//
     &                        NAME(1:LEF)//'" could not be opened. '//
     &                        'IOSTAT = '//STR(1:LES)//'.')
         END IF
      END IF
      IF (IERR.EQ.0) REWIND IFILEU
185   READ (ICONU,'(A)',END=195) FL
      IF (ISCOMP(FL(1:6),'...END').NE.0) GOTO 196
      IF (IERR.EQ.0) THEN
         LE = LENGTH(FL)
         IF (LE.EQ.0) THEN
            WRITE (IFILEU,*)
         ELSE
            WRITE (IFILEU,'(A)') FL(1:LE)
         END IF
      END IF
      GOTO 185
C
195   EOF = .TRUE.
196   CONTINUE
      IF (IERR.EQ.0) THEN
         REWIND IFILEU
         CLOSE (IFILEU)
      ELSE
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... PROLOG
C
C  Syntax: PROLOG command
C
6     CONTINUE
      PROLOG = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... EPILOG
C
C  Syntax: EPILOG command
C
7     CONTINUE
      EPILOG = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... INPUT_FILE
C
C  Syntax: INPUT_FILE file
C
8     CONTINUE
      FILIN = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... INPUT_UNIT
C
C  Syntax: INPUT_UNIT unit_number
C
9     CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NU,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 1
      ELSE IF (NU.LT.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 1
      ELSE
         IUINP = NU
      END IF
      GOTO 100
C
C  ..................................................... OUTPUT_FILE
C
C  Syntax: OUTPUT_FILE file [ append ]
C
10    CONTINUE
      IF (NIT.NE.2 .AND. NIT.NE.3) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
         IERROR = 2
      ELSE
         FILOUT = LINE(IS(2):IE(2))
         IF (NIT.EQ.3) THEN
            FL = LINE(IS(3):IE(3))
            IF (ISCOMP(FL,'APPEND').NE.0) THEN
               OUTAPP = .TRUE.
            ELSE
               CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &                              'Only "append" is allowed.')
               IERROR = 2
            END IF
         END IF
      END IF
      GOTO 100
C
C  ..................................................... OUTPUT_UNIT
C
C  Syntax: OUTPUT_UNIT unit_number
C
11    CONTINUE
      IF (NIT.NE.2) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
         IERROR = 2
         GOTO 100
      END IF
C
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NU,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 2
      ELSE IF (NU.LT.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number >= 0.')
         IERROR = 2
      ELSE
         IUOUT = NU
      END IF
      GOTO 100
C
C  ..................................................... INPUT_PRECONN
C
C  Syntax: INPUT_PRECONN yes | no
C
12    CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'YES').NE.0) THEN
         IPRECO = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         IPRECO = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "yes" or "no".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... BIGGER
C
C  Syntax: BIGGER real
C
13    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOREAL(LINE(IS(2):IE(2)),LE,X,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a positive real.')
         IERROR = 1
      ELSE IF (X.LE.0.) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a positive real.')
         IERROR = 1
      ELSE
         BI = X
      END IF
      GOTO 100
C
C  ..................................................... OUTPUT_PRECONN
C
C  Syntax: OUTPUT_PRECONN yes | no
C
14    CONTINUE
      IF (NIT.NE.2) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &            'The number of parameters of "'//ITEM(1:LEI)//
     &            '" is incorrect.')
         IERROR = 2
         GOTO 100
      END IF
C
      IF (ISCOMP(LINE(IS(2):IE(2)),'YES').NE.0) THEN
         OPRECO = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         OPRECO = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "yes" or "no".')
         IERROR = 2
      END IF
      GOTO 100
C
C  ..................................................... UNIT_RANGE
C
C  Syntax: UNIT_RANGE start end
C
15    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NU1,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      ELSE IF (NU1.LT.0 .OR. NU1.GT.999) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      END IF
C
      LE = IE(3)-IS(3)+1
      CALL TOINT(LINE(IS(3):IE(3)),LE,NU2,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      ELSE IF (NU2.LT.0 .OR. NU2.GT.999) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &                  'Must be a unit number: 0 <= n <= 999.')
         IERROR = 1
         GOTO 100
      ELSE IF (NU2-NU1+1.LT.NURANG) THEN
         CALL I2STR(NURANG,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,0,
     &                  'You must provide a range of at least '//
     &                  STR(1:LES)//' unit numbers.')
         IERROR = 1
         GOTO 100
      ELSE
         MINU = NU1
         MAXU = NU2
      END IF
      GOTO 100
C
C  ..................................................... SIZE_INT
C
C  Syntax: SIZE_INT int
C
16    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NSIZE,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NSIZE.LE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer > 0.')
         IERROR = 1
      ELSE
         NSINT = NSIZE
      END IF
      GOTO 100
C
C  ..................................................... SIZE_REAL
C
C  Syntax: SIZE_REAL int
C
17    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NSIZE,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NSIZE.LE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer > 0.')
         IERROR = 1
      ELSE
         NSREAL = NSIZE
      END IF
      GOTO 100
C
C  ..................................................... HELP_FILE
C
C  Syntax: HELP_FILE file
C
18    CONTINUE
      HFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... SMALLER
C
C  Syntax: SMALLER real
C
19    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOREAL(LINE(IS(2):IE(2)),LE,X,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a real number: 0 < x < 1.')
         IERROR = 1
      ELSE IF (X.LE.0. .OR. X.GT.1.) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be a real number: 0 < x < 1.')
         IERROR = 1
      ELSE
         SM = X
      END IF
      GOTO 100
C
C  ..................................................... PRINTOUT
C
C  Syntax: PRINTOUT full | half | no
C
20    CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'FULL').NE.0) THEN
         WEAK = .TRUE.
         STRO = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'HALF').NE.0) THEN
         STRO = .FALSE.
         WEAK = .TRUE.
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'NO').NE.0) THEN
         WEAK = .FALSE.
         STRO = .FALSE.
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "no", "half" or "full".')
         IERROR = 1
      END IF
      GOTO 100
C
C  ..................................................... PLUG
C
C  Syntax: PLUG number name
C
21    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NPLU,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer: n > 0.')
         IERROR = 1
         GOTO 100
      ELSE IF (NPLU.LT.1) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be an integer: n > 0.')
         IERROR = 1
         GOTO 100
      ELSE IF (NPLU.GT.MXPLUG) THEN
         CALL I2STR(MXPLUG,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,'There are only '//
     &                  STR(1:LES)//'plugs in the command list.')
         IERROR = 1
         GOTO 100
      END IF
C
C  Search the command array for the position of the plug.
      CALL I2STR(NPLU,STR,LES)
      NAME = 'PLUG'//STR(1:LES)
      NCOM = NMCOM+MXPLUG
      DO 600,ICM=1,NCOM
         IF (ISCOMP(COMMAN(ICM),NAME).NE.0) GOTO 603
600   CONTINUE
      CALL MABORT('Plug not found in command list.','MCONF')
C
603   CONTINUE
      LE = IE(3)-IS(3)+1
      LECO = LEN(COMMAN(1))
      IF (LE.GT.LECO) THEN
         CALL I2STR(LECO,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'The command name is too long. Maximum is '//
     &            STR(1:LES)//' characters.')
         IERROR = 1
         GOTO 100
      END IF
      NAME = LINE(IS(3):IE(3))
      LENA = IE(3)-IS(3)+1
      IF (ISCOMP(NAME(1:4),'PLUG').NE.0) THEN
         CALL I2STR(LECO,STR,LES)
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'The command cannot start with "plug".')
         IERROR = 1
         GOTO 100
      END IF
      DO 604,I=1,NCOM
         IF (ISCOMP(NAME,COMMAN(I)).NE.0) THEN
            CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'This command exists already.')
            IERROR = 1
            GOTO 100
         END IF
604   CONTINUE
      DO 605,I=1,NEXCLU
         IF (ISCOMP(NAME,EXCLU(I)).NE.0) THEN
            CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,3,
     &            'This command exists already.')
            IERROR = 1
            GOTO 100
         END IF
605   CONTINUE
      CALL UPPER(NAME,LENA)
      COMMAN(ICM) = NAME
      GOTO 100
C
C  ..................................................... PDESC_FILE
C
C  Syntax: PDESC_FILE file
C
22    CONTINUE
      PDESCF = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... MACRO_FILE
C
C  Syntax: MACRO_FILE file
C
23    CONTINUE
      MFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... MCL_ERROR_FILE
C
C  Syntax: MCL_ERROR_FILE file
C
24    CONTINUE
      EFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... MCL_OBJECT_FILE
C
C  Syntax: MCL_OBJECT_FILE file
C
25    CONTINUE
      OFILE = LINE(IS(2):IE(2))
      GOTO 100
C
C  ..................................................... ONEOF
C
C  Syntax: MACHINE_DIGITS int
C
26    CONTINUE
      LE = IE(2)-IS(2)+1
      CALL TOINT(LINE(IS(2):IE(2)),LE,NN,IERR)
      IF (IERR.NE.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE IF (NN.LT.0) THEN
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must an integer > 0.')
         IERROR = 1
      ELSE
         NDIG = NN
      END IF
      GOTO 100
C
C  ..................................................... MACHINE_DIGITS
C
C  Syntax: ONEOF rewind | ignore | return
C
27    CONTINUE
      IF (ISCOMP(LINE(IS(2):IE(2)),'REWIND').NE.0) THEN
         NEOFA = 1
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'IGNORE').NE.0) THEN
         NEOFA = 2
      ELSE IF (ISCOMP(LINE(IS(2):IE(2)),'RETURN').NE.0) THEN
         NEOFA = 3
      ELSE
         CALL CONERR(IERRU,NL,LINE,ORIG,IS,IE,2,
     &                  'Must be "rewind", "ignore" or "return".')
         IERROR = 1
      END IF
      GOTO 100
C
200   CONTINUE
C
C  An End-of-File has been encountered.
      CLOSE (ICONU)
      IF (IERROR.EQ.0) CLOSE (IERRU,STATUS='DELETE')
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE CONERR ( IERRU, NL, LINE, ORIG, IS, IE, IPARNO, MSG )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine MCONF. It writes messages to the error file,
C    for incorrect configuration directives.
C
C  Input arguments:
C    IERRU      Error messages are written to this unit.
C    NL         The line number of the configuration file, that is
C               currently being processed.
C    LINE       Items of the current line (see subroutine QLCR).
C    ORIG       The line of the configuration file that is being
C               processed.
C    IS         IS(I) is the starting character position of item I
C               in LINE.
C    IE         IE(I) is the ending character position of item I
C               in LINE.
C    IPARNO     The item # that is in error.
C    MSG        Short message to be displayed.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER IS(*), IE(*)
      CHARACTER*(*) LINE, ORIG, MSG
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*15 STR, BLANK
      DATA BLANK / '               ' /
C  ---------------------------------------------------------------------
C
      IF (NL.NE.0) THEN
         CALL I2STR(NL,STR,LES)
         WRITE (IERRU,10) STR(1:LES), ORIG(1:LENGTH(ORIG))
C
         LEB = LES+7
         IF (IPARNO.NE.0) WRITE (IERRU,20) BLANK(1:LEB),
     &              LINE(IS(IPARNO):IE(IPARNO)), LINE(IS(1):IE(1))
         LEM = LENGTH(MSG)
         IF (LEM.NE.0) WRITE (IERRU,30) BLANK(1:LEB), MSG(1:LEM)
      ELSE
         LEM = LENGTH(MSG)
         IF (LEM.NE.0) THEN
            WRITE (IERRU,40) MSG(1:LEM)
         ELSE
            CALL MABORT('Some error in config. No msg available.',
     &                        'CONERR')
         END IF
      END IF
C
10    FORMAT ($(FC)/'Line ',A,': ',A)
20    FORMAT ($(FC)A,'Parameter "',A,'" is incorrect for ',
     &        'the directive "',A,'".')
30    FORMAT ($(FC)A,A)
40    FORMAT ($(FC)A)
      END
C  ---------------------------------------------------------------------
      SUBROUTINE LEX ( LINE, LENG, TOKEN, NTOK, NUM, ALPHA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine LEX is the lexical analyzer of the Merlin/MCL package.
C    It scans the input line LINE and determines its structure. LINE is
C    analyzed in a sequence of lexical items, or tokens. Tokens are
C    stored in succession in array TOKEN, along with some additional
C    information:
C      For an identifier (TOKEN(I)=IDENT)
C        The identifier is converted to upper case and stored in ALPHA.
C        Two successive entries in array NUM point to the starting and
C        ending character positions of the identifier in ALPHA.
C      For an unsigned number (TOKEN(I)=UNSI)
C        The number is stored in array NUM.
C      For a string (TOKEN(I)=STRING)
C        The string is stored in ALPHA. Two successive entries in array
C        NUM point to the starting and ending character position of the
C        string in ALPHA. Single quotes can be embedded in a string
C        using the escape sequence \'
C
C    The complete list of all possible tokens is kept in common
C    block MCLSCD. The last entry in array TOKEN is always the EOL
C    token. Ie: TOKEN(NTOK) = EOL
C    Underscores in an identifier are ignored. (HELLO_THERE is stored
C    as HELLOTHERE.)
C
C  Input arguments:
C    LINE       The line to be analyzed.
C    LENG       Effective length of LINE.
C
C  Output arguments:
C    TOKEN      Integer array. Contains the numeric codes of the tokens
C    NTOK       Number of tokens in TOKEN.
C    NUM        Real array. It contains all unsigned numbers, starting
C               and ending character positions for identifiers and
C               strings.
C    ALPHA      Stores all identifiers and strings. It should be at
C               least as long as the input line.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      INTEGER TOKEN(*)
      CHARACTER*(*) LINE, ALPHA
      $(TYPE) NUM(*)
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER UCHA, UPPER1, CHA
      LOGICAL DOTFND, EOLIN, ALPHAB, NMERIC, SPCIAL, UNDER, LIT
      $(TYPE) NUMBER, RE, TEN
      EOLIN() = (FINGER.GE.LENG)
      ALPHAB(CHA) = ('A'.LE.CHA) .AND. (CHA.LE.'Z')
      NMERIC() = ('0'.LE.CHA) .AND. (CHA.LE.'9')
      SPCIAL() = (CHA.EQ.'=') .OR. (CHA.EQ.'(') .OR. (CHA.EQ.')').OR.
     &           (CHA.EQ.'.') .OR. (CHA.EQ.',') .OR. (CHA.EQ.';').OR.
     &           (CHA.EQ.'+') .OR. (CHA.EQ.'-') .OR. (CHA.EQ.'/').OR.
     &           (CHA.EQ.'[') .OR. (CHA.EQ.']') .OR. (CHA.EQ.':').OR.
     &           (CHA.EQ.'#') .OR. (CHA.EQ.'{') .OR. (CHA.EQ.'}').OR.
     &           (CHA.EQ.'?')
      UNDER() = (CHA.EQ.'_')
C  ---------------------------------------------------------------------
C
      IC0 = ICHAR('0')
      TEN = 10.
      FINGER = 0
      LLEN = LEN(LINE)
C
C  NT is the number of entries in TOKEN.
      NT = 0
C  NN is the number of entries in NUM.
      NN = 0
C  NAL is the effective length of ALPHA.
      NAL = 0
C
100   CONTINUE
      IF (EOLIN()) THEN
         GOTO 9000
      ELSE
         FINGER = FINGER+1
      END IF
C
C  CHA is always the character under processing.
      CHA = LINE(FINGER:FINGER)
      UCHA = UPPER1(CHA)
C
C  We jump to 200, when there is further processing to
C  be done from other parts of LEX during scanning.
200   CONTINUE
      IF (ALPHAB(UCHA)) THEN
C
C  Alphabetic character located.
C  The logical operators (and, or, xor, not) are being identified
C  here too.
         NAL = NAL+1
         NALST = NAL
         ALPHA(NALST:NALST) = UCHA
300      CONTINUE
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF ((ALPHAB(UCHA) .OR. NMERIC() .OR. UNDER())) THEN
C  Ignore underscores when they appear into identifiers.
               IF (UNDER()) GOTO 300
               NAL = NAL+1
               ALPHA(NAL:NAL) = UCHA
               GOTO 300
            ELSE
C  Obtain operator code (if operator).
               IF (ALPHA(NALST:NAL).EQ.'AND') THEN
                  CALL PSHSTR(AND,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE IF (ALPHA(NALST:NAL).EQ.'OR') THEN
                  CALL PSHSTR(OR,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE IF (ALPHA(NALST:NAL).EQ.'XOR') THEN
                  CALL PSHSTR(XOR,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE IF (ALPHA(NALST:NAL).EQ.'NOT') THEN
                  CALL PSHSTR(NOT,TOKEN,NT,LLEN)
                  NAL = NALST-1
               ELSE
                  CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                  RE = NALST
                  CALL PSHNUM(RE,NUM,NN,LLEN)
                  RE = NAL
                  CALL PSHNUM(RE,NUM,NN,LLEN)
               END IF
               GOTO 200
            END IF
         ELSE
            IF (ALPHA(NALST:NAL).EQ.'AND') THEN
               CALL PSHSTR(AND,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE IF (ALPHA(NALST:NAL).EQ.'OR') THEN
               CALL PSHSTR(OR,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE IF (ALPHA(NALST:NAL).EQ.'XOR') THEN
               CALL PSHSTR(XOR,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE IF (ALPHA(NALST:NAL).EQ.'NOT') THEN
               CALL PSHSTR(NOT,TOKEN,NT,LLEN)
               NAL = NALST-1
            ELSE
               CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
               RE = NALST
               CALL PSHNUM(RE,NUM,NN,LLEN)
               RE = NAL
               CALL PSHNUM(RE,NUM,NN,LLEN)
            END IF
         END IF
C
      ELSE IF (NMERIC()) THEN
C
C  Numeric character located. Proceed to transform
C  the sequence of characters to real number.
         DOTFND = .FALSE.
         NUMBER = ICHAR(CHA)-IC0
         EXPO = 0
         POWER = -1
500      CONTINUE
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (NMERIC()) THEN
               IF (POWER.EQ.-1) THEN
                  RE = ICHAR(CHA)-IC0
                  NUMBER = NUMBER*TEN+RE
                  GOTO 500
               ELSE
                  POWER = POWER+1
                  RE = ICHAR(CHA)-IC0
                  NUMBER = NUMBER+RE/(TEN**POWER)
                  GOTO 500
               END IF
            ELSE IF (CHA.EQ.'.') THEN
               POWER = 0
               IF (DOTFND) THEN
                  CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                  CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                  CALL PSHSTR(PERIOD,TOKEN,NT,LLEN)
                  GOTO 100
               ELSE
                  DOTFND = .TRUE.
                  GOTO 500
               END IF
            ELSE IF (UCHA.EQ.'E') THEN
               IF (.NOT.EOLIN()) THEN
                  FINGER = FINGER+1
                  CHA = LINE(FINGER:FINGER)
                  UCHA = UPPER1(CHA)
                  IF (NMERIC()) THEN
                     EXPO = EXPO*10+(ICHAR(CHA)-IC0)
600                  CONTINUE
                     IF (.NOT.EOLIN()) THEN
                        FINGER = FINGER+1
                        CHA = LINE(FINGER:FINGER)
                        UCHA = UPPER1(CHA)
                        IF (NMERIC()) THEN
                           EXPO = EXPO*10+(ICHAR(CHA)-IC0)
                           GOTO 600
                        ELSE
                           NUMBER = NUMBER*(TEN**EXPO)
                           CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                           CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           GOTO 200
                        END IF
                     ELSE
                        NUMBER = NUMBER*(TEN**EXPO)
                        CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                        CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                     END IF
                  ELSE IF (CHA.EQ.'+') THEN
                     IF (.NOT.EOLIN()) THEN
                        FINGER = FINGER+1
                        CHA = LINE(FINGER:FINGER)
                        UCHA = UPPER1(CHA)
                        IF (NMERIC()) THEN
                           EXPO = EXPO*10+(ICHAR(CHA)-IC0)
800                        CONTINUE
                           IF (.NOT.EOLIN()) THEN
                              FINGER = FINGER+1
                              CHA = LINE(FINGER:FINGER)
                              UCHA = UPPER1(CHA)
                              IF (NMERIC()) THEN
                                 EXPO = EXPO*10+
     &                                                  (ICHAR(CHA)-IC0)
                                 GOTO 800
                              ELSE
                                 NUMBER = NUMBER*(TEN**EXPO)
                                 CALL PSHSTR(UNSI,
     &                                      TOKEN,NT,LLEN)
                                 CALL PSHNUM(NUMBER,
     &                                      NUM,NN,LLEN)
                                 GOTO 200
                              END IF
                           ELSE
                              NUMBER = NUMBER*(TEN**EXPO)
                              CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                              CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           END IF
                        ELSE
                           CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                           CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                           CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                           CALL PSHSTR(PLUS,TOKEN,NT,LLEN)
                           GOTO 200
                        END IF
                     ELSE
                        CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                        CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                        CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                        CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                        CALL PSHSTR(PLUS,TOKEN,NT,LLEN)
                     END IF
                  ELSE IF (CHA.EQ.'-') THEN
                     IF (.NOT.EOLIN()) THEN
                        FINGER = FINGER+1
                        CHA = LINE(FINGER:FINGER)
                        UCHA = UPPER1(CHA)
                        IF (NMERIC()) THEN
                           EXPO = EXPO*10+(ICHAR(CHA)-IC0)
820                        CONTINUE
                           IF (.NOT.EOLIN()) THEN
                              FINGER = FINGER+1
                              CHA = LINE(FINGER:FINGER)
                              UCHA = UPPER1(CHA)
                              IF (NMERIC()) THEN
                                 EXPO = EXPO*10+
     &                                               (ICHAR(CHA)-IC0)
                                 GOTO 820
                              ELSE
                                 NUMBER = NUMBER*(TEN**(-EXPO))
                                 CALL PSHSTR(UNSI,
     &                                                TOKEN,NT,LLEN)
                                 CALL PSHNUM(NUMBER,
     &                                                NUM,NN,LLEN)
                                 GOTO 200
                              END IF
                           ELSE
                              NUMBER = NUMBER*(TEN**(-EXPO))
                              CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                              CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           END IF
                        ELSE
                           CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                           CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                           CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                           CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                           CALL PSHSTR(MINUS,TOKEN,NT,LLEN)
                           GOTO 200
                        END IF
                     ELSE
                        CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                        CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                        CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                        CALL PSHALP('E',ALPHA,NAL,NUM,NN,LLEN)
                        CALL PSHSTR(MINUS,TOKEN,NT,LLEN)
                     END IF
                  ELSE
                     FINGER = FINGER-1
C  Minus because we unget CHA.
                     CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                     CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                     GOTO 200
                  END IF
               ELSE
                  CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
                  CALL PSHNUM(NUMBER,NUM,NN,LLEN)
                  CALL PSHSTR(IDENT,TOKEN,NT,LLEN)
                  CALL PSHALP(CHA,ALPHA,NAL,NUM,NN,LLEN)
               END IF
            ELSE
               CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
               CALL PSHNUM(NUMBER,NUM,NN,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(UNSI,TOKEN,NT,LLEN)
            CALL PSHNUM(NUMBER,NUM,NN,LLEN)
         END IF
C
      ELSE IF (SPCIAL()) THEN
C
C  Special situation.
         IF (CHA.EQ.'=') THEN
            IF (.NOT. EOLIN()) THEN
               FINGER = FINGER+1
               CHA = LINE (FINGER:FINGER)
               UCHA = UPPER1(CHA)
               IF (UCHA.EQ.'=') THEN
C  Equality operator detected.
                  CALL PSHSTR(EQUAL,TOKEN,NT,LLEN)
                  GOTO 100
               ELSE
                  CALL PSHSTR(EQ,TOKEN,NT,LLEN)
                  GOTO 200
               END IF
            ELSE
               CALL PSHSTR(EQ,TOKEN,NT,LLEN)
            END IF
         ELSE
C  Special character detected.
            SPCODE = LXTRAN(CHA)
            CALL PSHSTR(SPCODE,TOKEN,NT,LLEN)
            GOTO 100
         END IF
C
      ELSE IF (CHA.EQ.' ' .OR. CHA.EQ.TAB) THEN
C
C  Ignore spaces and tabs.
         GOTO 100
      ELSE IF (CHA.EQ.'*') THEN
C
C  Star located.
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (UCHA.EQ.'*') THEN
               CALL PSHSTR(BISTAR,TOKEN,NT,LLEN)
               GOTO 100
            ELSE
               CALL PSHSTR(STAR,TOKEN,NT,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(STAR,TOKEN,NT,LLEN)
         END IF
C
      ELSE IF (CHA.EQ.'>') THEN
C
C  GT located.
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (UCHA.EQ.'=') THEN
               CALL PSHSTR(GE,TOKEN,NT,LLEN)
               GOTO 100
            ELSE
               CALL PSHSTR(GT,TOKEN,NT,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(GT,TOKEN,NT,LLEN)
         END IF
C
      ELSE IF (CHA.EQ.'<') THEN
C
C  LT located.
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
            UCHA = UPPER1(CHA)
            IF (UCHA.EQ.'=') THEN
               CALL PSHSTR(LE,TOKEN,NT,LLEN)
               GOTO 100
            ELSE
               CALL PSHSTR(LT,TOKEN,NT,LLEN)
               GOTO 200
            END IF
         ELSE
            CALL PSHSTR(LT,TOKEN,NT,LLEN)
         END IF
C
      ELSE IF (CHA.EQ.'''') THEN
C
C  Quote located. Process string.
         STFING = 0
         CALL PSHSTR(QUOTE,TOKEN,NT,LLEN)
         NALST = NAL+1
         LIT = .FALSE.
987      CONTINUE
         IF (.NOT.EOLIN()) THEN
            FINGER = FINGER+1
            CHA = LINE(FINGER:FINGER)
988         CONTINUE
            IF (LIT) THEN
               LIT = .FALSE.
               IF (CHA.EQ.'''') THEN
                  NAL = NAL+1
                  ALPHA(NAL:NAL) = CHA
                  STFING = STFING+1
               ELSE
                  NAL = NAL+1
                  ALPHA(NAL:NAL) = ESC
                  STFING = STFING+1
                  GOTO 988
               END IF
               GOTO 987
            ELSE
               IF (CHA.EQ.ESC) THEN
                  LIT = .TRUE.
                  GOTO 987
               ELSE IF (CHA.EQ.'''') THEN
                  IF (STFING.GT.0) THEN
                     RE = NALST
                     CALL PSHNUM(RE,NUM,NN,LLEN)
                     RE = NAL
                     CALL PSHNUM(RE,NUM,NN,LLEN)
                     CALL PSHSTR(STRING,TOKEN,NT,LLEN)
                  END IF
                  CALL PSHSTR(QUOTE,TOKEN,NT,LLEN)
C  In case of a null string we sent just two quote codes.
                  GOTO 100
               ELSE
                  NAL = NAL+1
                  ALPHA(NAL:NAL) = CHA
                  STFING = STFING+1
                  GOTO 987
               END IF
            END IF
         ELSE
C
C  End-of-Line reached.
            IF (LIT) THEN
               NAL = NAL+1
               ALPHA(NAL:NAL) = ESC
               STFING = STFING+1
            END IF
            IF (STFING.GT.0) THEN
               RE = NALST
               CALL PSHNUM(RE,NUM,NN,LLEN)
               RE = NAL
               CALL PSHNUM(RE,NUM,NN,LLEN)
               CALL PSHSTR(STRING,TOKEN,NT,LLEN)
            END IF
         END IF
      ELSE IF (CHA.EQ.'%') THEN
C
C  % is used as a comment character. Line scanning is terminated here.
      ELSE
C
C  Unrecognizable character.
         CALL PSHSTR(NOTREC,TOKEN,NT,LLEN)
C  Continue scanning.
         GOTO 100
C
      END IF
C
C  End of scanning.
9000  CONTINUE
C
C  Add the End-of-Line token.
      CALL PSHSTR(EOL,TOKEN,NT,LLEN)
      NTOK = NT
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSHALP ( ITEM, ALPHA, NAL, NUM, NN, LINLEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Stores ITEM in the character storage ALPHA.
C
C  Input arguments:
C    ITEM       Character item to be stored in ALPHA.
C    LINLEN     Length of the line under processing, as declared
C               in the program unit that called subroutine LEX.
C               IE: LINLEN = LEN(LINE)
C
C  Input / Output arguments:
C    ALPHA      The ALPHA character storage (see subroutine LEX).
C    NAL        Effective length of ALPHA.
C    NUM        The NUM array. (see subroutine LEX).
C    NN         Number of entries in array NUM.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) ALPHA, ITEM
      $(TYPE) NUM(*)
C  ---------------------------------------------------------------------
C  Local variables:
      $(TYPE) RE
C  ---------------------------------------------------------------------
C
      LES = LENGTH(ITEM)
      RE = NAL+1
      CALL PSHNUM(RE,NUM,NN,LINLEN)
      RE = NAL+LES
      CALL PSHNUM(RE,NUM,NN,LINLEN)
      ALPHA(NAL+1:NAL+LES) = ITEM
      NAL = NAL+LES
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSHNUM ( ITEM, NUM, NN, LINLEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Stores ITEM in array NUM.
C
C  Input arguments:
C    ITEM       The item to be stored.
C    LINLEN     Length of the line under processing, as declared
C               in the program unit that called subroutine LEX.
C               IE: LINLEN = LEN(LINE)
C
C  Input / Output arguments:
C    NUM        The NUM array. (see subroutine LEX).
C    NN         Number of entries in array NUM.
C
C  ---------------------------------------------------------------------
C  Arguments:
      $(TYPE) NUM(*), ITEM
C  ---------------------------------------------------------------------
C
      NN = NN+1
      IF (NN.LE.LINLEN) THEN
         NUM(NN) = ITEM
      ELSE
C  This should never happen.
         CALL MABORT ('Stack overflow.','PSHNUM')
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PSHSTR ( ITEM, TOKEN, NT, LINLEN )
C  ---------------------------------------------------------------------
C
C  Description:
C    Stores the token code ITEM in the array of tokens TOKEN.
C
C  Input arguments:
C    ITEM       The item to be stored.
C    LINLEN     Length of the line under processing, as declared
C               in the program unit that called subroutine LEX.
C               IE: LINLEN = LEN(LINE)
C
C  Input / Output arguments:
C    TOKEN      Tokens of the line under processing (see subroutine
C               LEX).
C    NT         Number of entries in TOKEN.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(LINLEN+1)
C  ---------------------------------------------------------------------
C
      NT = NT+1
      IF (NT.LE.LINLEN+1) THEN
         TOKEN(NT) = ITEM
      ELSE
C  This should never happen.
         CALL MABORT ('Stack overflow.','PSHSTR')
      END IF
C
      END
C  ---------------------------------------------------------------------
      FUNCTION LXTRAN ( C )
C  ---------------------------------------------------------------------
C
C  Description:
C    Transforms some special characters to their corresponding token
C    codes to be used in subroutine LEX.
C
C  Input arguments:
C    C          The character to be transformed.
C
C  ---------------------------------------------------------------------
      IMPLICIT INTEGER (A-Z)
C  Arguments:
      CHARACTER*(*) C
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C
      IF (C.EQ.'(') THEN
         LXTRAN = LEFTP
      ELSE IF (C.EQ.')') THEN
         LXTRAN = RIGHTP
      ELSE IF (C.EQ.'.') THEN
         LXTRAN = PERIOD
      ELSE IF (C.EQ.',') THEN
         LXTRAN = COMMA
      ELSE IF (C.EQ.'#') THEN
         LXTRAN = NE
      ELSE IF (C.EQ.':') THEN
         LXTRAN = COLON
      ELSE IF (C.EQ.'+') THEN
         LXTRAN = PLUS
      ELSE IF (C.EQ.'-') THEN
         LXTRAN = MINUS
      ELSE IF (C.EQ.'/') THEN
         LXTRAN = SLASH
      ELSE IF (C.EQ.'[') THEN
         LXTRAN = LBRA
      ELSE IF (C.EQ.']') THEN
         LXTRAN = RBRA
      ELSE IF (C.EQ.';') THEN
         LXTRAN = SEMI
      ELSE IF (C.EQ.'{') THEN
         LXTRAN = LBRACE
      ELSE IF (C.EQ.'}') THEN
         LXTRAN = RBRACE
      ELSE IF (C.EQ.'?') THEN
         LXTRAN = QMARK
      ELSE
         CALL MABORT ('Illegal input character.','LXTRAN')
      END IF
C
      END
C  ---------------------------------------------------------------------
      BLOCK DATA LEXINI
C  ---------------------------------------------------------------------
C
C  Description:
C    Assigns initial values to the integer token codes in common
C    block MCLSCD.
C
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C
      DATA LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE, LE, PLUS /
     &         1,      2,      3,     4,  5,  6,  7,  8,  9, 10,   11 /
      DATA MINUS, STAR, SLASH, BISTAR, COLON, EOL, IDENT, UNSI, NOTREC /
     &        12,   13,    14,     15,    16,  17,    18,   19,     20 /
      DATA STRING, QUOTE, LBRA, RBRA, SEMI, PERCE, AND, OR, XOR, NOT /
     &         21,    22,   23,   24,   25,    26,  27, 28,  29,  30 /
      DATA EQUAL, LBRACE, RBRACE, QMARK /
     &        31,     32,     33,    34 /
C
      END
C  ---------------------------------------------------------------------
      FUNCTION NUNIT ( )
C  ---------------------------------------------------------------------
C
C  Description:
C    Function NUNIT will search a range of unit numbers in order to
C    locate an unused unit. (A unit not assigned to an open file).
C    Function NUNIT is a necessity since Fortran requires both, a
C    unit number and a file name in an OPEN statement. Nevertheless
C    function NUNIT and all of the Merlin file handling routines rely
C    on the INQUIRE statement and especially on the OPENED, EXIST,
C    NUMBER and IOSTAT specifiers.
C    The range of numbers to be searched can be set from the
C    configuration file. For efficiency, the search for the unit number
C    starts from the number returned in the previous NUNIT call, not
C    from the beginning of the available range.
C
C  ---------------------------------------------------------------------
C  Common blocks:
      COMMON /UALLOC/ MINU, MAXU, LASTU
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OPEN
C  ---------------------------------------------------------------------
C
C  Start the search from the previously returned unit.
      DO 10,I=LASTU+1,MAXU
         INQUIRE (UNIT=I,OPENED=OPEN,IOSTAT=IERR)
         IF (IERR.EQ.0) THEN
            IF (.NOT.OPEN) THEN
               LASTU = I
               NUNIT = I
               RETURN
            END IF
         END IF
10    CONTINUE
C
C  Wrap around the end of the available range.
      DO 20,I=MINU,LASTU
         INQUIRE (UNIT=I,OPENED=OPEN,IOSTAT=IERR)
         IF (IERR.EQ.0) THEN
            IF (.NOT.OPEN) THEN
               LASTU = I
               NUNIT = I
               RETURN
            END IF
         END IF
20    CONTINUE
C
C  No more unit numbers are available.
      NUNIT = -1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE OLREAD ( NU, MAXCON, LINE, LLINE, ILL, NL, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Reads a number of lines (physical lines) from unit NU, until a line
C    without the continuation symbol & is reached, or until EOF. The
C    lines are concatenated and output in the large character variable
C    LLINE (logical line). If the maximum number of continuation lines
C    (MAXCON) is exceeded, subroutine OLREAD will continue reading
C    from unit NU; LLINE will not be enlarged however.
C
C  Input arguments:
C    NU         Unit to read from.
C    MAXCON     Maximum number of continuation lines.
C
C  Output arguments:
C    LINE       Character variable. Should have the length of one
C               physical line. Eg: CHARACTER*80 LINE.
C               LINE is used as a temporary variable and upon return
C               contains the last line read. If ILL = 0 the contents
C               of LINE are undefined.
C    LLINE      The input lines are concatenated and disposed here.
C               LLINE should be at least MAXCON*LEN(LINE) characters
C               long.
C    ILL        Effective length of LLINE.
C    NL         Number of lines read.
C    IERR       Error indicator:
C               IERR = 0 -> Everything ok.
C               IERR = 1 -> Too many continuation lines.
C               IERR = 2 -> EOF on input. If ILL.NE.0 this means that
C                           the continuation symbol & is present in the
C                           last line of the file, which is probably an
C                           error.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE, LLINE
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL ADDIT
C  ---------------------------------------------------------------------
C
      ILL = 0
      ADDIT = .TRUE.
      NL = 0
      LLINE = ' '
C
104   READ (NU,'(A)',END=200) LINE
      NL = NL+1
      IF (NL.GT.MAXCON) THEN
         ADDIT = .FALSE.
         IERR = 1
      END IF
      LE1 = LENGTH(LINE)
      IF (LE1.EQ.0) THEN
         IF (ADDIT) IERR = 0
         RETURN
      END IF
C
C  Is this a continuation line ?
      IF (LINE(LE1:LE1).EQ.'&') THEN
         IF (ADDIT) THEN
            LLINE(ILL+1:ILL+LE1-1) = LINE(1:LE1-1)
            ILL = ILL+LE1-1
         END IF
         GOTO 104
      END IF
      IF (ADDIT) THEN
         LLINE(ILL+1:ILL+LE1) = LINE(1:LE1)
         ILL = ILL+LE1
         IF (ILL.LT.LEN(LLINE)) LLINE(ILL+1:) = ' '
         IERR = 0
      END IF
      RETURN
C
200   CONTINUE
      IERR = 2
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PARSP ( PDESCF, IUOUT, STRO, WEAK, ICODE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Parses the panel description file and fills in the appropriate
C    data structures. This routine is called after the configuration
C    file has been processed, and assumes that the output file (unit
C    IUOUT) is open.
C    Panel related data are kept in a combination of real/integer
C    arrays and character variables. The following paragraphs describe
C    the data structures used for storing panel related data.
C
C    For the I-th panel command the following information is stored:
C      1) The name of the panel PCOM(I). PCOM is the panel table; a
C         CHARACTER*10 array.
C      2) Two integer pointers NCHELP(2*I-1), NCHELP(2*I) indicating
C         the starting and ending help lines for this panel, in the
C         panel description file.
C      3) An integer PONF(I), indicating the default status of the
C         panel: PONF(I) = 0 -> the panel is off.
C                PONF(I) = 1 -> the panel is off.
C      4) A pointer KPOINT(I), indicating the first keyword entry for
C         this panel, in the keyword table PKEY. After all panels
C         have been read in, a dummy entry is added at the end of the
C         KPOINT array, to indicate the "start" of a fake (non-
C         existent) panel. This is necessary since the number of
C         keywords in panel I is determined as: KPOINT(I+1)-KPOINT(I).
C
C    For the J-th keyword the following information is stored:
C      5) The name of the keyword PKEY(J). PKEY is the keyword table;
C         a CHARACTER*10 array. Keywords from all panels are stored
C         here.
C      6) The keyword type KTYPE(J). KTYPE(J) may be:
C           KTYPE(J) = 'I' for a keyword that accepts integer values.
C           KTYPE(J) = 'R' for a keyword that accepts real values.
C           KTYPE(J) = 'S' for a keyword that accepts string values.
C           KTYPE(J) = 'C' for a keyword that accepts string values;
C                          character case not being significant.
C      7) Two integer pointers NKHELP(2*J-1), NKHELP(2*J) indicating
C         the starting and ending help lines for this keyword, in the
C         panel description file.
C      8) A character KIO(J), indicating whether this is a normal or
C         an MCL returned keyword:
C           KIO(J)   =  ' ' for a normal keyword.
C           KIO(J) .NE. ' ' for an MCL returned keyword.
C      9) A pointer TPOINT(J), indicating the first entry for this
C         keyword in the PIR table.
C
C     10) The PIR table stores information related to the allowed
C         values of a keyword. (Assume TPOINT(J) = K)
C           a) For a keyword that accepts integer or real values:
C                PIR(K) = 1 This keyword has a range of allowed values.
C                PIR(K) = 2 This keyword has a set of allowed values.
C
C                If PIR(K) = 1 (range) then
C                  PIR(K+1) = 1 Interval is open at the left.
C                  PIR(K+1) = 2 Interval is closed at the left.
C                  PIR(K+1) = 3 Left limit is -Infinity.
C
C                  PIR(K+2) is the left limit; 0 for -Infinity.
C
C                  PIR(K+3) = 1 Interval is open at the right.
C                  PIR(K+3) = 2 Interval is closed at the right.
C                  PIR(K+3) = 3 Right limit is Infinity.
C
C                  PIR(K+4) is the right limit; 0 for -Infinity.
C                  PIR(K+5) is the current value of the keyword.
C                  PIR(K+6) is a pointer to the SPOINT array
C                           (description of the keyword).
C
C                If PIR(K) = 2 (set of values) then
C                  PIR(K+1) is the number of items in the set.
C                  PIR(K+2)...PIR(K+1+PIR(K+1)) is the set of allowed
C                                    values.
C                  PIR(K+2+PIR(K+1)) is the current value of the
C                                    keyword.
C                  PIR(K+3+PIR(K+1)) is a pointer to the SPOINT array
C                                    (description of the keyword).
C
C           b) For a keyword that accepts string or character values:
C                PIR(K) = 1 Any string is allowed as value.
C                PIR(K) = 2 This keyword has a set of allowed values.
C
C                If PIR(K) = 1 (any string) then
C                  PIR(K+1) is a pointer to the SPOINT array
C                           (current value).
C                  PIR(K+2) is a pointer to the SPOINT array
C                           (description of the keyword).
C
C                If PIR(K) = 2 (set of values) then
C                  PIR(K+1) = Number of items in the set.
C                  PIR(K+2)...PIR(K+1+PIR(K+1)) are pointers to the
C                             SPOINT array (allowed string items)
C                  PIR(K+2+PIR(K+1)) is a pointer to the SPOINT array
C                                    (current value).
C                  PIR(K+3+PIR(K+1)) is a pointer to the SPOINT array
C                                    (description of the keyword).
C
C     11) The SPOINT array stores the starting and ending character
C         positions for all panel related strings. For the L-th string
C         SPOINT(2*L-1) is the starting character position of the
C         string in the character storage PANSTR.
C         SPOINT(2*L) is the ending character position. Hence the
C         contents of the L-th string are:
C         PANSTR(SPOINT(2*L-1):SPOINT(2*L))
C
C     12) PANSTR is a large character, used to store all panel related
C         strings. Allowed values for string keywords, descriptions
C         of keywords, etc. are stored here. PANSTR is always addressed
C         through the pointers in array SPOINT.
C
C  Input arguments:
C    PDESCF     Name of the panel description file.
C    IUOUT      The Merlin output unit.
C    STRO       Informative messages are printed only if STRO = .TRUE.
C    WEAK       Error messages are printed only if WEAK = .TRUE.
C
C  Output arguments:
C    ICODE      Error indicator.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PDESCF
      LOGICAL STRO, WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      PARAMETER ( LENONE = 120, MXCONT = 10, LPLEN = LENONE*MXCONT )
      CHARACTER*(LENONE) LINE
      CHARACTER*(LPLEN) LLINE
      INTEGER TOKEN(LPLEN+1)
      $(TYPE) NUM(LPLEN)
      CHARACTER*(LPLEN) ALPHA
      LOGICAL EX
      CHARACTER*10 STR
      CHARACTER UPPER1
C  These types can be handled by a panel.
      PARAMETER ( NTYPES = 4 )
      CHARACTER TYPE(NTYPES)
      DATA TYPE / 'I', 'R', 'S', 'C' /
C  ---------------------------------------------------------------------
C
C  Number of panel commands.
      NPCOM = 0
C  Number of keywords.
      NPKEY = 0
C  Number of entries in the "pir" table.
      NPIR = 0
C  Number of string pointers.
      NPSP = 0
C  Number of characters stored.
      NPCH = 0
C
C  Zero help pointers.
      DO 10,I=1,2*MXPCOM
10       NCHELP(I) = 0
      DO 20,I=1,2*MXPKEY
20       NKHELP(I) = 0
C
C  Return if there's no panel description file.
      IF (PDESCF.EQ.' ') THEN
         ICODE = 0
         RETURN
      END IF
C
C  Get info about the panel description file.
      LEP = LENGTH(PDESCF)
      INQUIRE (FILE=PDESCF, EXIST=EX, IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LE1)
         IF (WEAK) WRITE (IUOUT,250) PDESCF(1:LEP), STR(1:LE1)
         ICODE = -1
         RETURN
      END IF
C
C  Does it exist ?
      IF (.NOT.EX) THEN
         IF (STRO) WRITE (IUOUT,251) PDESCF(1:LEP)
         ICODE = 0
         RETURN
      END IF
C
C  Allocate a unit.
      NU = NUNIT()
      IF (NU.EQ.-1) THEN
         IF (WEAK) WRITE (IUOUT,292) PDESCF(1:LEP)
         ICODE = -1
         RETURN
      END IF
C
C  Open the panel description file.
      OPEN (NU,FILE=PDESCF,IOSTAT=IERR)
      IF (IERR.NE.0) THEN
         CALL I2STR(IERR,STR,LE1)
         IF (WEAK) WRITE (IUOUT,252) PDESCF(1:LEP), STR(1:LE1)
         ICODE = -1
         RETURN
      END IF
C
C  NPL is the number of lines read so far.
      NPL = 0
C
C  Start the main loop. Read a line (along with all continuations).
100   CONTINUE
      CALL OLREAD(NU,MXCONT,LINE,LLINE,ILL,NL,IERR)
      NPL = NPL+NL
      IF (IERR.EQ.2) GOTO 200
      IF (IERR.EQ.1) THEN
         CALL I2STR(MXCONT,STR,LE1)
         IF (WEAK) WRITE (IUOUT,253) PDESCF(1:LEP), STR(1:LE1)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Convert the input line LLINE to tokens.
      CALL LEX(LLINE,ILL,TOKEN,NTOK,NUM,ALPHA)
C  Empty line ?
      IF (TOKEN(1).EQ.EOL) GOTO 100
C  Comment line ?
      IF (TOKEN(1).EQ.NE) GOTO 100
C  Start parsing the tokens.
      NT = 0
      NN = 0
C  Check for "." lines. NTOK is at least 1 here.
      NT = NT+1
      IF (TOKEN(NT).EQ.PERIOD) THEN
         NT = NT+1
         IF (TOKEN(NT).NE.IDENT) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,254)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
C  Is this is a .PANEL ?
         IF (ISCOMP(ALPHA(IST:IEND),'PANEL').NE.0) GOTO 501
C  Is this a .HELP ?
         IF (ISCOMP(ALPHA(IST:IEND),'HELP').NE.0) GOTO 502
      END IF
C
C  ------------------------- Keyword definition ------------------------
C
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,255)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LEK = IEND-IST+1
      IF (NPCOM.EQ.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,249) ALPHA(IST:IEND)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      IF (LEK.GT.KEYLEN) THEN
         CALL I2STR(KEYLEN,STR,LES)
         LE1 = LENGTH(PCOM(NPCOM))
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,248) ALPHA(IST:IEND),
     &                       PCOM(NPCOM)(1:LE1), STR(1:LES)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      DO 81,I=KPOINT(NPCOM),NPKEY
         IF (ISCOMP(PKEY(I),ALPHA(IST:IEND)).NE.0) THEN
            LE1 = LENGTH(PCOM(NPCOM))
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,256) ALPHA(IST:IEND),
     &                                        PCOM(NPCOM)(1:LE1)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
81    CONTINUE
      NPKEY = NPKEY+1
      IF (NPKEY.GT.MXPKEY)
     &    CALL OVER('pkey','PKEY','PARSP',MXPKEY,'MXPKEY')
      PKEY(NPKEY) = ALPHA(IST:IEND)
C
C  Get keyword type.
      NT = NT+1
      KIO(NPKEY) = ' '
      IF (TOKEN(NT).EQ.QMARK) THEN
         NT = NT+1
         KIO(NPKEY) = '?'
      END IF
      IF (TOKEN(NT).NE.IDENT) THEN
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,257) PKEY(NPKEY)(1:LE1),
     &                                  PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      DO 60,I=1,NTYPES
         IF (ISCOMP(TYPE(I),ALPHA(IST:IEND)).NE.0) GOTO 61
60    CONTINUE
      LE1 = LENGTH(PKEY(NPKEY))
      LE2 = LENGTH(PCOM(NPCOM))
      CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
      IF (WEAK) WRITE (IUOUT,258) PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
      CLOSE (NU)
      ICODE = -1
      RETURN
C
61    CONTINUE
      ITY = I
      KTYPE(NPKEY) = UPPER1(ALPHA(IST:IEND))
C
C  Get range of allowed values.
      IF (ITY.EQ.1) THEN
         TPOINT(NPKEY) = NPIR+1
         CALL IRANGE(TOKEN,NUM,ALPHA,NT,NN,IERR,PIR,NPIR,MXPIR,
     &                  PDESCF,NPL,PCOM(NPCOM),PKEY(NPKEY),IUOUT,
     &                  LEP,WEAK)
      ELSE IF (ITY.EQ.2) THEN
         TPOINT(NPKEY) = NPIR+1
         CALL RRANGE(TOKEN,NUM,ALPHA,NT,NN,IERR,PIR,NPIR,MXPIR,
     &                  PDESCF,NPL,PCOM(NPCOM),PKEY(NPKEY),IUOUT,
     &                  LEP,WEAK)
      ELSE IF (ITY.EQ.3 .OR. ITY.EQ.4) THEN
         TPOINT(NPKEY) = NPIR+1
         CALL SRANGE(TOKEN,NUM,ALPHA,NT,NN,IERR,PIR,NPIR,MXPIR,
     &                  SPOINT,NPSP,MXPPST,PANSTR,NPCH,MXPSTR,PDESCF,
     &                  NPL,PCOM(NPCOM),PKEY(NPKEY),IUOUT,LEP,WEAK)
      END IF
      IF (IERR.NE.0) THEN
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Get default value.
      NT = NT+1
      IF (ITY.EQ.1) THEN
C  Default value for an integer.
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,259)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         XDE = NUM(NN)
         IF (XDE.NE.INT(XDE)) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,259)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         IDEF = ISI*NINT(XDE)
         CALL PRANI(IDEF,NPKEY,IOK)
         IF (IOK.EQ.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,261)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
         PIR(NPIR) = IDEF
      ELSE IF (ITY.EQ.2) THEN
C  Default value for a real.
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,262)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         XDE = ISI*NUM(NN)
         CALL PRANR(XDE,NPKEY,IOK)
         IF (IOK.EQ.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,261)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
         PIR(NPIR) = XDE
      ELSE IF (ITY.EQ.3 .OR. ITY.EQ.4) THEN
C  Default value for a string.
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,264)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.STRING) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,264)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.QUOTE) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,264)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
         LE1 = IEND-IST+1
         CALL PRANS(ALPHA(IST:IEND),LE1,NPKEY,IOK)
         IF (IOK.EQ.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(NPKEY))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,261)
     &                      PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
         PIR(NPIR) = NPSP+1
         IF (NPSP+2.GT.MXPPST)
     &          CALL OVER('spoint','SPOINT','PARSP',MXPPST,'MXPPST')
         SPOINT(NPSP+1) = NPCH+1
         SPOINT(NPSP+2) = NPCH+LE1
         NPSP = NPSP+2
         IF (NPCH+LE1.GT.MXPSTR)
     &          CALL OVER('panstr','PANSTR','PARSP',MXPSTR,'MXPSTR')
         PANSTR(NPCH+1:NPCH+LE1) = ALPHA(IST:IEND)
         NPCH = NPCH+LE1
      END IF
C
C  Get description.
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,268)
     &                PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.STRING) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,268)
     &             PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,268)
     &                PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','PARSP',MXPIR,'MXPIR')
      PIR(NPIR) = NPSP+1
      IF (NPSP+2.GT.MXPPST)
     &    CALL OVER('spoint','SPOINT','PARSP',MXPPST,'MXPPST')
      SPOINT(NPSP+1) = NPCH+1
      SPOINT(NPSP+2) = NPCH+LE1
      NPSP = NPSP+2
      IF (NPCH+LE1.GT.MXPSTR)
     &    CALL OVER('panstr','PANSTR','PARSP',MXPSTR,'MXPSTR')
      PANSTR(NPCH+1:NPCH+LE1) = ALPHA(IST:IEND)
      NPCH = NPCH+LE1
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PKEY(NPKEY))
         LE2 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,271)
     &                PKEY(NPKEY)(1:LE1), PCOM(NPCOM)(1:LE2)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Read the next line from the panel description file.
      GOTO 100
C
C  -------------------------- Panel definition -------------------------
501   CONTINUE
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
C  Check against other panel names. (Should check all Merlin commands)
      CALL PLOCP(ALPHA(IST:IEND),IPA)
      IF (IPA.NE.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,274) ALPHA(IST:IEND)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NPCOM = NPCOM+1
      IF (NPCOM.GT.MXPCOM)
     &    CALL OVER('pcom','PCOM','PARSP',MXPCOM,'MXPCOM')
      PCOM(NPCOM) = ALPHA(IST:IEND)
      KPOINT(NPCOM) = NPKEY+1
C
C  Get panel status. Should be "on" or "off".
      NT = NT+1
      IF (TOKEN(NT).NE.IDENT) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,242)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      IF (ISCOMP(ALPHA(IST:IEND),'ON').NE.0) THEN
         PONF(NPCOM) = 1
      ELSE IF (ISCOMP(ALPHA(IST:IEND),'OFF').NE.0) THEN
         PONF(NPCOM) = 0
      ELSE
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,242)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
      NT = NT+1
      IF (NT.NE.NTOK) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,273)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
C
C  Read the next line from the panel description file.
      GOTO 100
C
C
C  -------------------------- Help definition --------------------------
502   CONTINUE
      IF (NPCOM.EQ.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,279)
         CLOSE (NU)
         ICODE = -1
         RETURN
      END IF
      NT = NT+1
      IF (NT.EQ.NTOK) THEN
         IF (NCHELP(2*NPCOM-1).NE.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,284) PCOM(NPCOM)(1:LE1)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         IWHAT = 1
      ELSE
C  Get key name here.
         IF (TOKEN(NT).NE.IDENT) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,281)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
         LE1 = IEND-IST+1
         NT = NT+1
         IF (NT.NE.NTOK) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,280)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         DO 82,IK=KPOINT(NPCOM),NPKEY
            IF (ISCOMP(PKEY(IK),ALPHA(IST:IEND)).NE.0) GOTO 83
82       CONTINUE
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         LE1 = LENGTH(PCOM(NPCOM))
         IF (WEAK) WRITE (IUOUT,283) ALPHA(IST:IEND),
     &                                  PCOM(NPCOM)(1:LE1)
         CLOSE (NU)
         ICODE = -1
         RETURN
C
83       CONTINUE
         IF (NKHELP(2*IK-1).NE.0) THEN
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            LE1 = LENGTH(PKEY(IK))
            LE2 = LENGTH(PCOM(NPCOM))
            IF (WEAK) WRITE (IUOUT,285) PKEY(IK)(1:LE1),
     &                                        PCOM(NPCOM)(1:LE2)
            CLOSE (NU)
            ICODE = -1
            RETURN
         END IF
         IWHAT = 2
      END IF
C
      IF (IWHAT.EQ.1) THEN
         NCHELP(2*NPCOM-1) = NPL+1
      ELSE
         NKHELP(2*IK-1) = NPL+1
      END IF
C
C  Scan file until .end
202   CONTINUE
      READ (NU,'(A)',END=201) LINE
      NPL = NPL+1
      IF (ISCOMP(LINE,'.END').EQ.0) GOTO 202
      IF (IWHAT.EQ.1) THEN
         NCHELP(2*NPCOM) = NPL-1
      ELSE
         NKHELP(2*IK) = NPL-1
      END IF
      GOTO 100
C
201   CONTINUE
      CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
      IF (WEAK) WRITE (IUOUT,282)
      CLOSE (NU)
      ICODE = -1
      RETURN
C
200   CONTINUE
      CLOSE (NU)
      IF (ILL.NE.0) THEN
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,275)
         ICODE = -1
         RETURN
      END IF
      IF (NPCOM+1.GT.MXPCOM)
     &    CALL OVER('kpoint','KPOINT','PARSP',MXPCOM,'MXPCOM')
      KPOINT(NPCOM+1) = NPKEY+1
      ICODE = 0
C
C  The following WRITE statements display the amount of storage used
C  by the current panel description file. Used for development purposes.
C     IF (STRO) THEN
C         WRITE (IUOUT,*) 'Number of panel commands. NPCOM = ', NPCOM
C         WRITE (IUOUT,*) 'Number of keywords. NPKEY = ', NPKEY
C         WRITE (IUOUT,*) 'Number of entries in the "pir" table. ',
C     &                    'NPIR = ',NPIR
C         WRITE (IUOUT,*) 'Number of string pointers. NPSP = ', NPSP
C         WRITE (IUOUT,*) 'Number of characters stored. NPCH = ', NPCH
C     END IF
C
250   FORMAT (/$(FC)'An I/O error occurred while inquiring for the ',
     &        'existence'
     &        /$(FC)'of the panel description file "',A,
     &        '". IOSTAT = ',A,'.')
251   FORMAT (/$(FC)'Warning: The panel description file "',A,
     &        '" does not exist.')
292   FORMAT (/$(FC)'Cannot allocate a unit number for the ',
     &        'panel description file "',A,'".')
252   FORMAT (/$(FC)'An I/O error occurred while opening the ',
     &        'panel description ',
     &        /$(FC)'file "',A,'". IOSTAT = ',A)
253   FORMAT (/$(FC)'Too many continuation lines in the panel ',
     &        'description file "',A,'".'
     &        /$(FC)'Maximum number of continuation lines is ',A,'.')
254   FORMAT ($(FC)'An identifier is expected after the dot.')
255   FORMAT ($(FC)'An identifier is expected at the begining ',
     &        'of the line.')
249   FORMAT ($(FC)'The keyword ',A,' was found outside a panel ',
     &        'definition.')
248   FORMAT ($(FC)'The keyword ',A, ' in the ',A,' panel'/
     &        $(FC)'is too long. Maximum is ',A,' characters.')
256   FORMAT ($(FC)'The keyword ',A, ' has already ',
     &        'been defined in the ',A,' panel.')
257   FORMAT ($(FC)'The keyword type (I, R, S or C) is expected ',
     &        'after the ',A,' keyword'
     &         /$(FC)'in the ',A,' panel.')
258   FORMAT ($(FC)'The type for the ',A,' keyword in the ',A,' panel'
     &        /$(FC)'is unrecognized. It should be I, R, S or C.')
259   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'should be an integer.')
261   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'is out of the specified range.')
262   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'should be a real.')
264   FORMAT ($(FC)'The default value for the ',A,' keyword in the ',
     &        A,' panel'
     &        /$(FC)'should be a string.')
268   FORMAT ($(FC)'A string is expected as a description for the ',
     &        A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
271   FORMAT ($(FC)'Too many items in the input line for the ',
     &        A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
272   FORMAT ($(FC)'A panel name is expected after ".panel".')
274   FORMAT ($(FC)'The ',A,' panel has alredy been defined.')
242   FORMAT ($(FC)'The panel status ("on" or "off") is expected ',
     &        'after the panel name.')
273   FORMAT ($(FC)'Too many items in the ".panel" line.')
279   FORMAT ($(FC)'Help was found outside a panel definition.')
284   FORMAT ($(FC)'You already specified ".help" for the ',
     &        A,' panel.')
281   FORMAT ($(FC)'Keywords only should follow ".help".')
280   FORMAT ($(FC)'Too many items in the ".help" line.')
283   FORMAT ($(FC)'There is no such keyword ',A,' in the ',
     &        A,' panel.')
285   FORMAT ($(FC)'You already specified ".help" for the ',
     &        A,' keyword in the ',A,' panel.')
282   FORMAT ($(FC)'A ".help" has no corresponding ".end".')
275   FORMAT ($(FC)'A continuation mark was specified in the last line'
     &        /$(FC)'of the panel description file.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PERRMS ( PDESCF, LEP, WEAK, NPL, IUOUT )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PARSP. Prints an informative header
C    before subroutine PARSP issues any error message.
C
C  Input arguments:
C    PDESCF     Name of the panel description file.
C    LEP        Effective length of PDESCF.
C    WEAK       This routine prints only if WEAK = .TRUE.
C    NPL        Number of lines read so far from the panel description
C               file.
C    IUOUT      The Merlin output unit.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) PDESCF
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER*10 STR
C  ---------------------------------------------------------------------
C
      IF (WEAK) THEN
         CALL I2STR(NPL,STR,LES)
         CALL DASH
         WRITE (IUOUT,10) PDESCF(1:LEP), STR(1:LES)
      END IF
C
10    FORMAT ($(FC)'An error occurred in the panel description file "',
     &        A,'", line ',A,'.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PLOCK ( IPA, KEY, IKE )
C  ---------------------------------------------------------------------
C
C  Description:
C    Determines the position of the keyword KEY in the keyword table
C    PKEY. For more information on the structures that store panel
C    related information, see the discussion in subroutine PARSP.
C
C  Input arguments:
C    IPA        Index of the panel, KEY belongs to.
C    KEY        The keyword to search for.
C
C  Output arguments:
C    IKE        Index of KEY in the PKEY table. IKE is set to 0
C               if KEY is not a valid keyword for the panel with
C               index IPA.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) KEY
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C
      KP1 = KPOINT(IPA)
      KP2 = KPOINT(IPA+1)-1
      DO 10,I=KP1,KP2
         IF (ISCOMP(KEY,PKEY(I)).NE.0) THEN
            IKE = I
            RETURN
         END IF
10    CONTINUE
C
C  Keyword not found.
      IKE = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PLOCP ( PNAME, IPOS )
C  ---------------------------------------------------------------------
C
C  Description:
C    Determines the position IPOS of the panel PNAME in the panel table.
C
C  Input arguments:
C    PNAME      Name of the panel.
C
C  Output arguments:
C    IPOS       Position of PNAME in the panel table. IPOS is set to 0
C               if PNAME is not a valid panel.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) PNAME
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C
      DO 10,I=1,NPCOM
         IF (ISCOMP(PNAME,PCOM(I)).NE.0) THEN
            IPOS = I
            RETURN
         END IF
10    CONTINUE
C
C  Panel not found.
      IPOS = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRANI ( IVAL, NK, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the given integer IVAL is allowed as a value for a
C    keyword.
C
C  Input arguments:
C    IVAL       The integer to check.
C    NK         The index of the corresponding keyword in the keyword
C               table PKEY.
C
C  Output arguments:
C    IOK        IOK = 0 -> Value not allowed for this keyword.
C               IOK = 1 -> Value ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(KTYPE(NK),'I').EQ.0)
     &    CALL MABORT('Not an integer key: '//PKEY(NK),'PRANI')
C
      IP = TPOINT(NK)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IL = NINT(PIR(IP+1))
         NLE = NINT(PIR(IP+2))
         IR = NINT(PIR(IP+3))
         NRI = NINT(PIR(IP+4))
         CALL IPI(IVAL,IL,NLE,IR,NRI,OK)
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         OK = .FALSE.
         DO 10,I=1,NITE
            IP = IP+1
            ITEM = NINT(PIR(IP))
            OK = OK .OR. ITEM.EQ.IVAL
10       CONTINUE
      END IF
      IOK = 0
      IF (OK) IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IPI ( IVAL, IL, NLE, IR, NRI, OK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PRANI. Checks whether IVAL is within the
C    interval defined by NLE and NRI.
C
C  Input arguments:
C    IVAL       The integer to check.
C    IL         IL = 1 -> The interval is open on the left.
C               IL = 2 -> The interval is closed on the left.
C               IL = 3 -> -Infinity was specified as the left limit.
C    NLE        The left limit of the interval.
C    IR         IL = 1 -> The interval is open on the right.
C               IL = 2 -> The interval is closed on the right.
C               IL = 3 -> Infinity was specified as the right limit.
C    NRI        The right limit of the interval.
C
C  Output arguments:
C    OK         OK = .TRUE.  -> IVAL is within the specified interval.
C               OK = .FALSE. -> IVAL is not within the interval.
C
C  ---------------------------------------------------------------------
C  Arguments:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (IL.EQ.1) THEN
         IF (IR.EQ.1) THEN
            OK = IVAL.GT.NLE .AND. IVAL.LT.NRI
         ELSE IF (IR.EQ.2) THEN
            OK = IVAL.GT.NLE .AND. IVAL.LE.NRI
         ELSE IF (IR.EQ.3) THEN
            OK = IVAL.GT.NLE
         END IF
      ELSE IF (IL.EQ.2) THEN
         IF (IR.EQ.1) THEN
            OK = IVAL.GE.NLE .AND. IVAL.LT.NRI
         ELSE IF (IR.EQ.2) THEN
            OK = IVAL.GE.NLE .AND. IVAL.LE.NRI
         ELSE IF (IR.EQ.3) THEN
            OK = IVAL.GE.NLE
         END IF
      ELSE IF (ABS(IL).EQ.3) THEN
         IF (IR.EQ.1) THEN
            OK = IVAL.LT.NRI
         ELSE IF (IR.EQ.2) THEN
            OK = IVAL.LE.NRI
         ELSE IF (IR.EQ.3) THEN
            OK = .TRUE.
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRANR ( XVAL, NK, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the given real XVAL is allowed as a value for a
C    keyword.
C
C  Input arguments:
C    XVAL       The real to check.
C    NK         The index of the corresponding keyword in the keyword
C               table PKEY.
C
C  Output arguments:
C    IOK        IOK = 0 -> Value not allowed for this keyword.
C               IOK = 1 -> Value ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(KTYPE(NK),'R').EQ.0)
     &    CALL MABORT('Not a real key: '//PKEY(NK),'PRANR')
C
      IP = TPOINT(NK)
      IRL = NINT(PIR(IP))
      IF (IRL.EQ.1) THEN
         IL = NINT(PIR(IP+1))
         XLE = PIR(IP+2)
         IR = NINT(PIR(IP+3))
         XRI = PIR(IP+4)
         CALL IPR(XVAL,IL,XLE,IR,XRI,OK)
         IP = IP+4
      ELSE IF (IRL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         OK = .FALSE.
         DO 10,I=1,NITE
            IP = IP+1
            XITEM = PIR(IP)
            OK = OK .OR. XITEM.EQ.XVAL
10       CONTINUE
      END IF
      IOK = 0
      IF (OK) IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE IPR ( XVAL, IL, XLE, IR, XRI, OK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Called by subroutine PRANR. Checks whether XVAL is within the
C    interval defined by XLE and XRI.
C
C  Input arguments:
C    XVAL       The real to check.
C    IL         IL = 1 -> The interval is open on the left.
C               IL = 2 -> The interval is closed on the left.
C               IL = 3 -> -Infinity was specified as the left limit.
C    XLE        The left limit of the interval.
C    IR         IL = 1 -> The interval is open on the right.
C               IL = 2 -> The interval is closed on the right.
C               IL = 3 -> Infinity was specified as the right limit.
C    XRI        The right limit of the interval.
C
C  Output arguments:
C    OK         OK = .TRUE.  -> XVAL is within the specified interval.
C               OK = .FALSE. -> XVAL is not within the interval.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      LOGICAL OK
C  ---------------------------------------------------------------------
C
      IF (IL.EQ.1) THEN
         IF (IR.EQ.1) THEN
            OK = XVAL.GT.XLE .AND. XVAL.LT.XRI
         ELSE IF (IR.EQ.2) THEN
            OK = XVAL.GT.XLE .AND. XVAL.LE.XRI
         ELSE IF (IR.EQ.3) THEN
            OK = XVAL.GT.XLE
         END IF
      ELSE IF (IL.EQ.2) THEN
         IF (IR.EQ.1) THEN
            OK = XVAL.GE.XLE .AND. XVAL.LT.XRI
         ELSE IF (IR.EQ.2) THEN
            OK = XVAL.GE.XLE .AND. XVAL.LE.XRI
         ELSE IF (IR.EQ.3) THEN
            OK = XVAL.GE.XLE
         END IF
      ELSE IF (ABS(IL).EQ.3) THEN
         IF (IR.EQ.1) THEN
            OK = XVAL.LT.XRI
         ELSE IF (IR.EQ.2) THEN
            OK = XVAL.LE.XRI
         ELSE IF (IR.EQ.3) THEN
            OK = .TRUE.
         END IF
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE PRANS ( SVAL, LES, NK, IOK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Checks whether the given string SVAL is allowed as a value for a
C    keyword.
C
C  Input arguments:
C    SVAL       The string to check.
C    LES        Effective length of SVAL.
C    NK         The index of the corresponding keyword in the keyword
C               table PKEY.
C
C  Output arguments:
C    IOK        IOK = 0 -> Value not allowed for this keyword.
C               IOK = 1 -> Value ok.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) SVAL
C  ---------------------------------------------------------------------
C  Common blocks:
C  Maximum length of a keyword.
      PARAMETER ( KEYLEN = 10 )
C  Maximum number of panel commands.
      PARAMETER ( MXPCOM = 80 )
C  Maximum number of keywords in all panel commands.
      PARAMETER ( MXPKEY = 1000 )
C  Maximum storage for integer/real keywords.
      PARAMETER ( MXPIR = 5000 )
C  Maximum storage for string keywords and descriptions.
      PARAMETER ( MXPSTR = 20000 )
C  Maximum number of pointers in the above string storage.
      PARAMETER ( MXPPST = 5000 )
C
      CHARACTER*(KEYLEN) PCOM(MXPCOM)
      INTEGER            PONF(MXPCOM)
      INTEGER            KPOINT(MXPCOM)
      INTEGER            NCHELP(2*MXPCOM)
      CHARACTER*(KEYLEN) PKEY(MXPKEY)
      INTEGER            NKHELP(2*MXPKEY)
      CHARACTER          KTYPE(MXPKEY)
      INTEGER            TPOINT(MXPKEY)
      CHARACTER          KIO(MXPKEY)
      $(TYPE)            PIR(MXPIR)
      INTEGER            SPOINT(MXPPST)
      CHARACTER*(MXPSTR) PANSTR
      COMMON /PCOMS1/ KPOINT, NCHELP, NKHELP, TPOINT, SPOINT, PONF
      COMMON /PCOMS2/ PIR
      COMMON /PCOMS3/ PKEY, KTYPE, PANSTR, PCOM, KIO
C
C  NPCOM: Number of panel commands in the panel description file.
C  NPKEY: Number of keywords in all panel commands.
C  NPIR:  Number of positions filled in array PIR.
C  NPSP:  Number of positions filled in the char-pointer array SPOINT.
C  NPCH:  Number of character positions filled in PANSTR.
      COMMON /PCOMS4/ NPCOM, NPKEY, NPIR, NPSP, NPCH
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
      LOGICAL STRI
C  ---------------------------------------------------------------------
C
      IF (ISCOMP(KTYPE(NK),'S').NE.0) THEN
         STRI = .TRUE.
      ELSE IF (ISCOMP(KTYPE(NK),'C').NE.0) THEN
         STRI = .FALSE.
      ELSE
         CALL MABORT('Not a string key: '//PKEY(NK),'PRANS')
      END IF
C
      IP = TPOINT(NK)
      IAL = NINT(PIR(IP))
      IF (IAL.EQ.1) THEN
         OK = .TRUE.
      ELSE IF (IAL.EQ.2) THEN
         IP = IP+1
         NITE = NINT(PIR(IP))
         OK = .FALSE.
         DO 10,I=1,NITE
            IP = IP+1
            NS = NINT(PIR(IP))
            IST = SPOINT(NS)
            IEND = SPOINT(NS+1)
            LEC = IEND-IST+1
            IF (LES.EQ.0) THEN
               OK = OK .OR. LEC.EQ.0
            ELSE
               IF (LEC.NE.0) THEN
                  IF (STRI) THEN
                     OK = OK .OR. SVAL(1:LES).EQ.
     &                                           PANSTR(IST:IEND)
                  ELSE
                     OK = OK .OR. ISCOMP(SVAL(1:LES),
     &                                           PANSTR(IST:IEND)).NE.0
                  END IF
               END IF
            END IF
10       CONTINUE
      END IF
C
      IOK = 0
      IF (OK) IOK = 1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE QLCR ( LINE, L, NEW, IS, IE, NIT, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Given a string of characters (LINE), this subroutine will
C    separate ("crack") them into a number of items. An item is
C    a sequence of non-blank non-tab characters. For example the
C    following line:
C      I love Merlin.
C    contains the 3 items: I
C                          love
C                          Merlin.
C    Blanks and/or tabs can be embedded in an item using single quotes.
C    The following line:
C      An 'extra fast' routine.
C    contains the 3 items: An
C                          extra fast
C                          routine.
C    A single quote can be entered using the escape sequence \'
C    (even within a pair of quotes). For example the line:
C      A single quote \'
C    contains the 4 items: A
C                        single
C                        quote
C                        '
C    The characters of all items are concatenated and returned in the
C    output string NEW. In order to separate one item from another
C    two arrays of pointers (IS, IE) are returned as well.
C    IS(I) and IE(I) contain the starting and ending positions of
C    item I, correspondingly.
C    For example the input line:
C      Hello it\'s me again.
C    contains 4 items and sets NEW to:
C      Helloit'smeagain.
C    IS(I), IE(I) are:  I    IS(I)    IE(I)    Item
C                       -----------------------------
C                       1      1        5      Hello
C                       2      6        9      it's
C                       3     10       11      me
C                       4     12       17      again.
C
C  Input arguments:
C    LINE       The line to be cracked.
C    L          Effective length of LINE.
C
C  Output arguments:
C    NEW        All items concatenated.
C    IS         IS(I) is the starting character position of item I in
C               the output character NEW.
C    IE         IE(I) is the ending character position of item I in the
C               output character NEW.
C    NIT        Number of items.
C    IERR       Error indicator.
C               IERR =  0 -> Everything ok.
C               IERR = -2 -> A closing quote is missing.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) LINE, NEW
      INTEGER IS(*), IE(*)
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER QUOTE, SPACE
      PARAMETER ( QUOTE = '''', SPACE = ' ' )
      LOGICAL INWORD, INSTR
      CHARACTER C
      LOGICAL WHITE
      WHITE(C) = C.EQ.SPACE .OR. C.EQ.TAB
C  ---------------------------------------------------------------------
C
      NIT = 0
      IERR = 0
      IF (L.EQ.0) RETURN
C
      NN = 0
C
C  INWORD is set to .TRUE. as long as the character under examination
C  belongs to an item. When examining item separators (blanks/tabs),
C  INWORD is set to .FALSE.
      INWORD = .FALSE.
C
C  INSTR is set to .TRUE. while examining characters after a quote.
C  (not an escaped quote.)
      INSTR = .FALSE.
C
C  The following loop examines all characters in LINE, one by one.
      I = 1
10    CONTINUE
      C = LINE(I:I)
      IF (C.EQ.QUOTE) THEN
         IF (INSTR) THEN
C  Closing quote.
            INSTR = .FALSE.
         ELSE
C  Opening quote.
            INSTR = .TRUE.
            IF (.NOT.INWORD) THEN
               INWORD = .TRUE.
               IF (I.EQ.L) THEN
                  GOTO 20
               ELSE
                  NIT = NIT+1
                  IS(NIT) = NN+1
               END IF
            END IF
         END IF
      ELSE IF (WHITE(C)) THEN
C  Space or tab.
         IF (INWORD) THEN
            IF (INSTR) THEN
C  A new item starts.
               NN = NN+1
               NEW(NN:NN) = C
            ELSE
               INWORD = .FALSE.
               IF (IS(NIT).GT.NN) THEN
                  NIT = NIT-1
               ELSE
                  IE(NIT) = NN
               END IF
            END IF
         END IF
      ELSE IF (C.EQ.ESC) THEN
C  The escape character (backslash).
         IF (I.EQ.L) THEN
C  The escape character is the last one in LINE.
            NN = NN+1
            NEW(NN:NN) = C
            IF (.NOT.INWORD) THEN
               NIT = NIT+1
               IS(NIT) = NN
               INWORD = .TRUE.
            END IF
            GOTO 20
         ELSE
C  Something follows the escape character.
            I = I+1
            C = LINE(I:I)
            IF (.NOT.INWORD) THEN
               NIT = NIT+1
               IS(NIT) = NN+1
               INWORD = .TRUE.
            END IF
            IF (C.EQ.QUOTE) THEN
C  A quote follows the escape character. Add it to the current item.
               NN = NN+1
               NEW(NN:NN) = QUOTE
            ELSE
C  A character other than a quote follows the escape character.
               NN = NN+1
               NEW(NN:NN) = ESC
               GOTO 10
            END IF
         END IF
      ELSE
C  Any other character.
         IF (INWORD) THEN
C  Add the character to the current item.
            NN = NN+1
            NEW(NN:NN) = C
         ELSE
C  Start a new item.
            INWORD = .TRUE.
            NN = NN+1
            NEW(NN:NN) = C
            NIT = NIT+1
            IS(NIT) = NN
         END IF
      END IF
      I = I+1
      IF (I.LE.L) GOTO 10
C
C  All characters in LINE have been processed. Set the ending pointer
C  for the last item. Also check whether a closing quote is missing.
20    CONTINUE
      IF (INWORD) THEN
         IF (IS(NIT).GT.NN) THEN
            NIT = NIT-1
         ELSE
            IE(NIT) = NN
         END IF
         IF (INSTR) IERR = -2
      END IF
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE RRANGE ( TOKEN, NUM, ALPHA, NT, NN, IERR, PIR, NPIR,
     &                    MXPIR, PDESCF, NPL, COMM, KEY, IUOUT, LEP,
     &                    WEAK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine RRANGE parses the definition of a range or set of
C    reals and fills in the appropriate data structures.
C    A range of reals has the form:
C      L S,E R
C    where: L is ( or [ indicating an open/closed interval on the left.
C           R is ) or ] indicating an open/closed interval on the right.
C           S is a real (start of the interval) or -inf (for
C             infinity).
C           E is a real (end of the interval) or inf (for infinity).
C    Examples: (-1,1)  (-inf,0]  [10,100)  [70,80]
C
C    A set (list) of reals has the form:
C      { R1, R2, R3, ... }
C    where R1, R2,... etc. are reals.
C
C    Panel related data structures are described in subroutine PARSP.
C    The lexical analysis tables (TOKEN, NUM, ALPHA) are described in
C    subroutine LEX.
C
C  Input arguments:
C    TOKEN      Tokens in the current line (see subroutine LEX).
C    NUM        Numbers in the current line (see subroutine LEX).
C    ALPHA      Identifiers/strings in the current line (see subroutine
C               LEX).
C    MXPIR      Maximum allowed number of entries in the PIR table.
C    PDESCF     Name of the panel description file.
C    NPL        Number of lines read so far from the panel description
C               file.
C    COMM       The panel currently under processing.
C    KEY        The keyword currently under processing.
C    IUOUT      The Merlin output unit.
C    LEP        Effective length of PDESCF.
C    WEAK       Error messages are printed only of WEAK = .TRUE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  Input / Output arguments:
C    NT         Number of tokens (TOKEN) used so far.
C    NN         Number of numbers (NUM) used so far.
C    PIR        The PIR table (see subroutine PARSP).
C    NPIR       Number of entries in the PIR array.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(*)
      $(TYPE) NUM(*), PIR(MXPIR)
      CHARACTER*(*) ALPHA, PDESCF, COMM, KEY
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C  Local variables:
      LOGICAL OK
      $(TYPE) XLI, XLE, XRI
C  ---------------------------------------------------------------------
C
      IERR = 0
      NT = NT+1
      INIT = TOKEN(NT)
      IF (INIT.EQ.LBRACE) THEN
C  A list of values is expected.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = 2
C  Allocate PIR space for the # of items.
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         ISTLI = NPIR
10       CONTINUE
         NT = NT+1
         ITOK = TOKEN(NT)
         ISI = 1
         IF (ITOK.EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (ITOK.EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).NE.UNSI) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NN = NN+1
         XLI = ISI*NUM(NN)
         NPIR = NPIR+1
         IF (NPIR.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         PIR(NPIR) = XLI
         NT = NT+1
         IF (TOKEN(NT).EQ.COMMA) GOTO 10
         IF (TOKEN(NT).EQ.RBRACE) THEN
C  End of list ok.
            NLI = NPIR-ISTLI
            PIR(ISTLI) = NLI
            RETURN
         END IF
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,271) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE IF (INIT.EQ.LEFTP .OR. INIT.EQ.LBRA) THEN
C  A left open/closed interval is expected.
         XLE = 0.
         XRI = 0.
         IF (INIT.EQ.LEFTP) ITPL = 1
         IF (INIT.EQ.LBRA) ITPL = 2
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPL = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XLE = NUM(NN)
            XLE = ISI*XLE
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
         NT = NT+1
         IF (TOKEN(NT).NE.COMMA) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,274) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         ITPR = 0
         NT = NT+1
         ISI = 1
         IF (TOKEN(NT).EQ.PLUS) THEN
            ISI = 1
            NT = NT+1
         ELSE IF (TOKEN(NT).EQ.MINUS) THEN
            ISI = -1
            NT = NT+1
         END IF
         IF (TOKEN(NT).EQ.IDENT) THEN
            NN = NN+1
            IST = NINT(NUM(NN))
            NN = NN+1
            IEND = NINT(NUM(NN))
            IF (ISCOMP(ALPHA(IST:IEND),'inf').NE.0) THEN
               ITPR = 3*ISI
            ELSE
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,275) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
               RETURN
            END IF
         ELSE IF (TOKEN(NT).EQ.UNSI) THEN
            NN = NN+1
            XRI = NUM(NN)
            XRI = XRI
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,275) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         NT = NT+1
         IF (TOKEN(NT).EQ.RIGHTP) THEN
            IF (ITPR.EQ.0) ITPR = 1
         ELSE IF (TOKEN(NT).EQ.RBRA) THEN
            IF (ITPR.EQ.0) ITPR = 2
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,277) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
C
         IF (NPIR+5.GT.MXPIR)
     &          CALL OVER('pir','PIR','RRANGE',MXPIR,'MXPIR')
         PIR(NPIR+1) = 1
         PIR(NPIR+2) = ITPL
         PIR(NPIR+3) = XLE
         PIR(NPIR+4) = ITPR
         PIR(NPIR+5) = XRI
         NPIR = NPIR+5
         IF (ITPL.EQ. 1 .OR.ITPL.EQ.2) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = XLE.LE.XRI
            ELSE
               OK = ITPR.EQ.3
            END IF
         ELSE IF (ITPL.EQ.-3) THEN
            IF (ITPR.EQ.1 .OR. ITPR.EQ.2) THEN
               OK = .TRUE.
            ELSE
               OK = ITPL.LT.ITPR
            END IF
         ELSE
            OK = .FALSE.
         END IF
         IF (.NOT.OK) THEN
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,278) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
            RETURN
         END IF
      ELSE IF (INIT.EQ.EOL) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,280) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      ELSE
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,279) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
C
270   FORMAT ($(FC)'Reals only should be present in the ',
     &        'list of allowed values'/
     &        $(FC)'for the ',A,' keyword in the ',A,' panel.')
271   FORMAT ($(FC)'The list of real values for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
272   FORMAT ($(FC)'The left limit for the ',A,
     &        ' keyword in the ',A,' panel,'/
     &        $(FC)'is incorrect.')
274   FORMAT ($(FC)'A comma is expected after the left ',
     &        'limit for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel.')
275   FORMAT ($(FC)'The right limit for the ',A,
     &        ' keyword in the ',A,' panel, is incorrect.')
277   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
278   FORMAT ($(FC)'The interval specification for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect. Left limit > right limit.')
280   FORMAT ($(FC)'There is no description of allowed values ',
     &        'for the ',A,' keyword'
     &        /$(FC)'in the ',A,' panel, is incorrect.')
279   FORMAT ($(FC)'The specification of allowed values for the ',A,
     &        ' keyword'
     &        /$(FC)'in the ',A,' panel.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE SRANGE ( TOKEN, NUM, ALPHA, NT, NN, IERR, PIR, NPIR,
     &                    MXPIR, SPOINT, NPSP, MXPPST, PANSTR, NPCH,
     &                    MXPSTR, PDESCF, NPL, COMM, KEY, IUOUT, LEP,
     &                    WEAK )
C  ---------------------------------------------------------------------
C
C  Description:
C    Subroutine SRANGE parses the definition of a range of strings and
C    fills in the appropriate data structures.
C    A set (list) of strings has the form:
C      { S1, S2, S3, ... }
C      where S1, S2,... etc. are strings.
C    or
C      { any }
C    The second form causes any string to be accepted as a keyword
C    value.
C    Examples: {'a','b','c'}  {any} {'Hello','Goodbye'}
C
C    Panel related data structures are described in subroutine PARSP.
C    The lexical analysis tables (TOKEN, NUM, ALPHA) are described in
C    subroutine LEX.
C
C  Input arguments:
C    TOKEN      Tokens in the current line (see subroutine LEX).
C    NUM        Numbers in the current line (see subroutine LEX).
C    ALPHA      Identifiers/strings in the current line (see subroutine
C               LEX).
C    MXPIR      Maximum allowed number of entries in the PIR table.
C    MXPPST     Maximum allowed number of entries in the SPOINT array.
C    MXPSTR     Maximum allowed number of characters in PANSTR.
C    PDESCF     Name of the panel description file.
C    NPL        Number of lines read so far from the panel description
C               file.
C    COMM       The panel currently under processing.
C    KEY        The keyword currently under processing.
C    IUOUT      The Merlin output unit.
C    LEP        Effective length of PDESCF.
C    WEAK       Error messages are printed only of WEAK = .TRUE.
C
C  Output arguments:
C    IERR       Error indicator.
C
C  Input / Output arguments:
C    NT         Number of tokens (TOKEN) used so far.
C    NN         Number of numbers (NUM) used so far.
C    PIR        The PIR table (see subroutine PARSP).
C    NPIR       Number of entries in the PIR array.
C    SPOINT     The SPOINT table (see subroutine PARSP).
C    NPSP       Number of entries in SPOINT.
C    PANSTR     The PANSTR character storage (see subroutine PARSP).
C    NPCH       Number of characters in PANSTR.
C
C  ---------------------------------------------------------------------
C  Arguments:
      INTEGER TOKEN(*), SPOINT(MXPPST)
      CHARACTER*(*) PANSTR, ALPHA, PDESCF, COMM, KEY
      $(TYPE) NUM(*), PIR(MXPIR)
      LOGICAL WEAK
C  ---------------------------------------------------------------------
C  Common blocks:
C  Codes used by the lexical analysis routine LEX.
      INTEGER LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &        LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &        IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &        SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &        RBRACE, QMARK
      COMMON /MCLSCD/ LEFTP, RIGHTP, PERIOD, COMMA, GT, LT, NE, EQ, GE,
     &                LE, PLUS, MINUS, STAR, SLASH, BISTAR, COLON, EOL,
     &                IDENT, UNSI, NOTREC, STRING, QUOTE, LBRA, RBRA,
     &                SEMI, PERCE, AND, OR, XOR, NOT, EQUAL, LBRACE,
     &                RBRACE, QMARK
C  ---------------------------------------------------------------------
C
      IERR = 0
      NT = NT+1
      INIT = TOKEN(NT)
      IF (INIT.NE.LBRACE) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,270) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
C
C  A list of values is expected.
      NT = NT+1
      IAL = 2
      IF (TOKEN(NT).EQ.IDENT) THEN
         NN = NN+1
         IST = NINT(NUM(NN))
         NN = NN+1
         IEND = NINT(NUM(NN))
         IF (ISCOMP(ALPHA(IST:IEND),'any').NE.0) THEN
            NT = NT+1
            IF (TOKEN(NT).NE.RBRACE) THEN
               LE1 = LENGTH(KEY)
               LE2 = LENGTH(COMM)
               CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
               IF (WEAK) WRITE (IUOUT,271) KEY(1:LE1), COMM(1:LE2)
               IERR = 1
            ELSE
C  Any string is accepted.
               IAL = 1
               NPIR = NPIR+1
               IF (NPIR.GT.MXPIR)
     &                  CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
               PIR(NPIR) = IAL
            END IF
         ELSE
            LE1 = LENGTH(KEY)
            LE2 = LENGTH(COMM)
            CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
            IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
            IERR = 1
         END IF
         RETURN
      END IF
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
      PIR(NPIR) = IAL
C  Allocate PIR space for the # of items.
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
      ISTLI = NPIR
10    CONTINUE
      IF (TOKEN(NT).NE.QUOTE) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.STRING) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
      NT = NT+1
      IF (TOKEN(NT).NE.QUOTE) THEN
         LE1 = LENGTH(KEY)
         LE2 = LENGTH(COMM)
         CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
         IF (WEAK) WRITE (IUOUT,272) KEY(1:LE1), COMM(1:LE2)
         IERR = 1
         RETURN
      END IF
      NN = NN+1
      IST = NINT(NUM(NN))
      NN = NN+1
      IEND = NINT(NUM(NN))
      LE1 = IEND-IST+1
      NPIR = NPIR+1
      IF (NPIR.GT.MXPIR)
     &    CALL OVER('pir','PIR','SRANGE',MXPIR,'MXPIR')
      PIR(NPIR) = NPSP+1
      IF (NPSP+2.GT.MXPPST)
     &    CALL OVER('spoint','SPOINT','SRANGE',MXPPST,'MXPPST')
      SPOINT(NPSP+1) = NPCH+1
      SPOINT(NPSP+2) = NPCH+LE1
      IF (NPCH+LE1.GT.MXPSTR)
     &    CALL OVER('panstr','PANSTR','SRANGE',MXPSTR,'MXPSTR')
      PANSTR(NPCH+1:NPCH+LE1) = ALPHA(IST:IEND)
      NPSP = NPSP+2
      NPCH = NPCH+LE1
C
      NT = NT+1
      IF (TOKEN(NT).EQ.COMMA) THEN
         NT = NT+1
         GOTO 10
      END IF
      IF (TOKEN(NT).EQ.RBRACE) THEN
C  List ended ok.
         NLI = NPIR-ISTLI
         PIR(ISTLI) = NLI
         RETURN
      END IF
      LE1 = LENGTH(KEY)
      LE2 = LENGTH(COMM)
      CALL PERRMS(PDESCF,LEP,WEAK,NPL,IUOUT)
      IF (WEAK) WRITE (IUOUT,276) KEY(1:LE1), COMM(1:LE2)
      IERR = 1
C
270   FORMAT ($(FC)'A list of strings is expected as allowed values'
     &        /$(FC)'for the ',A,' keyword in the ',A,' panel.')
271   FORMAT ($(FC)'A right brace is expected after "any"',
     &        ' in the list of allowed'/
     &        $(FC)'values for the ',A,' keyword in the ',A,' panel.')
272   FORMAT ($(FC)'Strings only should be present in the ',
     &        'list of allowed values'/
     &        $(FC)'for the ',A,' keyword in the ',A,' panel.')
276   FORMAT ($(FC)'The list of strings for the ',A,
     &        ' keyword in the ',A,' panel,'
     &        /$(FC)'is incorrect.')
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOINT ( STR, LE, N, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts an integer stored in a string (STR) to its numeric
C    representation. Leading/trailing blanks/tabs in STR are ignored.
C
C  Input arguments:
C    STR        The string to convert.
C    LE         Effective length of STR.
C
C  Output arguments:
C    N          Integer representation of STR.
C    IERR       Error indicator.
C               IERR =  0 -> Conversion succeeded.
C               IERR = -2 -> Conversion failed. STR does not contain
C                            a proper integer.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      LOGICAL NUMBER
      NUMBER(C) = C.GE.'0' .AND. C.LE.'9'
C  ---------------------------------------------------------------------
C
C  Skip leading blanks.
      DO 60,I=1,LE
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            NC = I
            GOTO 70
         END IF
60    CONTINUE
C
C  The string is full of blanks/tabs.
      IERR = -2
      RETURN
C
70    CONTINUE
C
C  Remove trailing blanks/tabs.
C  At least one non-white space character is present.
      DO 80,I=LE,NC,-1
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            IEND = I
            GOTO 90
         END IF
80    CONTINUE
      CALL MABORT('String full of blanks.','TOINT')
C
90    CONTINUE
      N = 0
      IOFF = ICHAR('0')
C
C  Check for a sign.
      C = STR(1:1)
      IF (C.EQ.'+') THEN
         ISI = 1
         IPOS = NC+1
      ELSE IF (C.EQ.'-') THEN
         ISI = -1
         IPOS = NC+1
      ELSE
         ISI = 1
         IPOS = NC
      END IF
C
C  Convert the number.
      IF (IPOS.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      DO 10,I=IPOS,IEND
         C = STR(I:I)
         IF (NUMBER(C)) THEN
            N = 10*N+ICHAR(C)-IOFF
         ELSE
            IERR = -2
            RETURN
         END IF
10    CONTINUE
C
      N = ISI*N
      IERR = 0
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE TOREAL ( STR, LE, X, IERR )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts a real, stored in a string (STR) to its numeric
C    representation. Leading/trailing blanks/tabs in STR are ignored.
C    The number in STR should be of the form:
C      [ + | - ] <INT> . <FRAC> <EXP>
C    where <INT> is the integer part: a sequence of digits,
C          <FRAC> is the fractional part: another sequence of digits,
C          and
C          <EXP> is the exponent in the form: E | D [ + | - ] <INT>
C    The integer part <INT> or the fractional part <FRAC> may be
C    omitted, but not both.
C
C  Input arguments:
C    STR        The string to convert.
C    LE         Effective length of STR.
C
C  Output arguments:
C    N          Integer representation of STR.
C    IERR       Error indicator.
C               IERR =  0 -> Conversion succeeded.
C               IERR = -2 -> Conversion failed. STR does not contain
C                            a proper integer.
C
C  ---------------------------------------------------------------------
      IMPLICIT $(TYPE) (A-H,O-Z)
C  Arguments:
      CHARACTER*(*) STR
C  ---------------------------------------------------------------------
C  Common blocks:
      CHARACTER TAB, ESC
      COMMON /SPECHA/ TAB, ESC
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      LOGICAL IPART, FPART
      LOGICAL NUMBER
      NUMBER(C) = C.GE.'0' .AND. C.LE.'9'
C  ---------------------------------------------------------------------
C
      IZERO = ICHAR('0')
      V = 0.0
      W = 0.0
      IERR = 0
      POT= 1.0
      IEX = 0
      TEN = 10
      IPART = .FALSE.
      FPART = .FALSE.
C
C  Skip leading blanks/tabs.
      DO 60,I=1,LE
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            NC = I
            GOTO 70
         END IF
60    CONTINUE
C
C  The string is full of blanks/tabs.
      IERR = -2
      RETURN
C
70    CONTINUE
C
C  Remove trailing blanks/tabs.
C  At least one non-white space character is present.
      DO 80,I=LE,NC,-1
         C = STR(I:I)
         IF (C.NE.' ' .AND. C.NE.TAB) THEN
            IEND = I
            GOTO 90
         END IF
80    CONTINUE
      CALL MABORT('String full of blanks.','TOREAL')
C
C  Check for a sign.
90    CONTINUE
      C = STR(NC:NC)
      IF (C.EQ.'+') THEN
         IS = 1
      ELSE IF (C.EQ.'-') THEN
         IS = -1
      ELSE
         IS = 1
         GOTO 10
      END IF
C
      NC = NC+1
      IF (NC.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      C = STR(NC:NC)
C
C  Convert the integer part.
10    CONTINUE
      IF (NUMBER(C)) THEN
         V = 10*V + ICHAR(C)-IZERO
         IPART = .TRUE.
         NC = NC+1
         IF (NC.GT.IEND) THEN
            X = IS*V
            RETURN
         END IF
         C = STR(NC:NC)
         GOTO 10
      END IF
C
20    CONTINUE
      IF (C.EQ.'.') GOTO 30
      CALL UPPER(C,1)
      IF (C.EQ.'E' .OR. C.EQ.'D') GOTO 40
      IERR = -2
      RETURN
C
C  Convert the fractional part.
30    CONTINUE
      NC = NC+1
      IF (NC.GT.IEND) THEN
         IF (.NOT.(IPART.OR.FPART)) IERR = -2
         X = IS*(V+W)
         RETURN
      END IF
      C = STR(NC:NC)
      IF (NUMBER(C)) THEN
         POT = 10*POT
         W = W + (ICHAR(C)-IZERO)/POT
         FPART = .TRUE.
         GOTO 30
      END IF
      CALL UPPER(C,1)
      IF (C.EQ.'E' .OR. C.EQ.'D') GOTO 40
      IERR = -2
      RETURN
C
40    NC = NC+1
      IF (NC.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      C = STR(NC:NC)
C
C  An exponent without integer or fractional parts is not allowed.
      IF (.NOT.(IPART.OR.FPART)) THEN
         IERR = -2
         RETURN
      END IF
C
C  Check for a sign in the exponent.
      IF (C.EQ.'+') THEN
         IES = 1
      ELSE IF (C.EQ.'-') THEN
         IES = -1
      ELSE
         IES = 1
         GOTO 50
      END IF
      NC = NC+1
      IF (NC.GT.IEND) THEN
         IERR = -2
         RETURN
      END IF
      C = STR(NC:NC)
C
C  Convert the exponent.
50    CONTINUE
      IF (NUMBER(C)) THEN
         IEX = 10*IEX + ICHAR(C)-IZERO
         NC = NC+1
         IF (NC.GT.IEND) THEN
            X = IS*(V+W) * TEN**(IES*IEX)
            RETURN
         END IF
         C = STR(NC:NC)
         GOTO 50
      END IF
C
      IERR = -2
C
      END
C  ---------------------------------------------------------------------
      CHARACTER FUNCTION UPPER1 ( CHA )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts one character to upper case. It calls the main conversion
C    routine UPPER.
C
C  Input arguments:
C    CHA        The character to be converted.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER CHA
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER CHA1
C  ---------------------------------------------------------------------
C
      CHA1 = CHA
      CALL UPPER(CHA1,1)
      UPPER1 = CHA1
C
      END
C  ---------------------------------------------------------------------
      SUBROUTINE UPPER ( CHA, LL )
C  ---------------------------------------------------------------------
C
C  Description:
C    Converts to upper case a string of characters. Assumes the ASCII
C    character set.
C
C  Input arguments:
C    LL         Effective length of CHA.
C
C  Input / Output arguments:
C    CHA        On input the character to be converted.
C               On output lower case characters are converted to their
C               upper case equivalents.
C
C  ---------------------------------------------------------------------
C  Arguments:
      CHARACTER*(*) CHA
C  ---------------------------------------------------------------------
C  Local variables:
      CHARACTER C
      LOGICAL ISCAP
C  ---------------------------------------------------------------------
C
      DO 10,I=1,LL
         C = CHA(I:I)
         IF (.NOT.ISCAP(C)) THEN
            IC = ICHAR(C)
            IF (IC.GE.97 .AND. IC.LE.122) CHA(I:I) = CHAR(IC-32)
         END IF
10    CONTINUE
C
      END
